(******************************************************************************)
(*     Alt-Ergo: The SMT Solver For Software Verification                     *)
(*     Copyright (C) 2013-2015 --- OCamlPro                                   *)
(*     This file is distributed under the terms of the CeCILL-C licence       *)
(******************************************************************************)

(******************************************************************************)
(*     The Alt-Ergo theorem prover                                            *)
(*     Copyright (C) 2006-2013                                                *)
(*     CNRS - INRIA - Universite Paris Sud                                    *)
(*                                                                            *)
(*     Sylvain Conchon                                                        *)
(*     Evelyne Contejean                                                      *)
(*                                                                            *)
(*     Francois Bobot                                                         *)
(*     Mohamed Iguernelala                                                    *)
(*     Stephane Lescuyer                                                      *)
(*     Alain Mebsout                                                          *)
(*                                                                            *)
(*   This file is distributed under the terms of the CeCILL-C licence         *)
(******************************************************************************)

open Format
open Options
open Sig

module T = Term
module F = Formula
module MF = F.Map
module SF = F.Set
module Ex = Explanation
module MT = T.Map
module ST = T.Set
module SubstT = Term.Subst


(*** Instantiation part of the module ***)

module type S = sig
  type t
  type tbox

  val empty : t
  val add_terms : t -> T.Set.t -> F.gformula -> t
  val add_lemma : t -> F.gformula -> Ex.t -> t
  val add_predicate : t -> F.gformula -> t

  type instances = (F.gformula * Ex.t) list

  val m_lemmas :
    t ->
    tbox ->
    (F.t -> F.t -> bool) ->
    instances * instances (* goal_directed, others *)

  val m_predicates :
    t ->
    tbox ->
    (F.t -> F.t -> bool) ->
    instances * instances (* goal_directed, others *)

  (* returns used axioms/predicates * unused axioms/predicates *)
  val retrieve_used_context : t -> Ex.t -> Formula.t list * Formula.t list

end

module Make(X : Theory.S) : S with type tbox = X.t = struct

  module EM = Matching.EMatch(X)(Formula)

  open EM

  type tbox = X.t
  type instances = (F.gformula * Ex.t) list

  type t = {
    lemmas : (int * Ex.t) MF.t;
    predicates : (int * Ex.t) MF.t;
    matching : EM.t;
  }

  let empty = {
    lemmas = MF.empty ;
    matching = EM.empty;
    predicates = MF.empty;
  }

  let add_terms env s gf =
    let infos = {
      term_age = gf.F.age ;
      term_from_goal    = gf.F.gf ;
      term_from_formula = gf.F.lem ;
      term_from_terms   = gf.F.from_terms
    }
    in
    { env with
      matching = T.Set.fold (EM.add_term infos) s env.matching }

  module SST = Set.Make(String)


  let init_with_replay_used acc f =
    if Sys.command (sprintf "[ -e %s ]" f) <> 0 then
      begin
        fprintf fmt
          "File %s not found! Option -replay-used will be ignored@." f;
        acc
      end
    else
      let cin = open_in f in
      let acc = ref (match acc with None -> SST.empty | Some ss -> ss) in
      begin
        try while true do acc := SST.add (input_line cin) !acc done;
        with End_of_file -> close_in cin
      end;
      Some !acc

  let used =
    if Options.replay_used_context () then
      init_with_replay_used None (Options.get_used_context_file ())
    else if Options.replay_all_used_context () then
      let dir = Filename.dirname (Options.get_used_context_file ()) in
      Array.fold_left
        (fun acc f ->
          let f = sprintf "%s/%s" dir f in
          if (Filename.check_suffix f ".used") then begin
            init_with_replay_used acc f
          end
          else acc
        ) None (Sys.readdir dir)
    else None

  let parent s =
    if String.length s = 0 then s
    else
      match s.[0] with
      | '#' ->
        (match Str.split (Str.regexp "#") s
         with | [a;b] -> a | _ -> assert false)
      | _ -> s

  let unused_context f = match used, F.view f with
    | None  , _ -> false
    | Some s_used, F.Lemma {F.name=s} ->
      not (String.length s = 0 || SST.mem (parent s) s_used)
    | _ -> assert false

  let add_lemma env gf dep =
    let {F.f=f;age=age} = gf in
    if (*not (Ex.is_empty dep) ||*) unused_context f then env
    else
      let age , dep =
        try
          let age' , dep' = MF.find f env.lemmas in
          min age age' , Ex.union dep dep'
        with Not_found -> age , dep
      in
      { env with lemmas=MF.add f (age,dep) env.lemmas }

  let add_predicate env gf =
    let {F.f=f;age=age} = gf in
    if unused_context f then env
    else { env with predicates = MF.add f (age,Ex.empty) env.predicates }

  let mtriggers env formulas =
    MF.fold
      (fun lem (age, dep) env ->
	match F.view lem with
	    F.Lemma {F.triggers = tgs; main = f} ->
	      List.fold_left
		(fun env tr ->
		  let info =
		    { trigger = tr;
                      trigger_age = age ;
		      trigger_orig = lem ;
		      trigger_formula = f ;
		      trigger_dep = dep}
		  in
		  { env with
		    matching =
		      EM.add_trigger info env.matching })
		env tgs
	  | _ -> assert false
      )
      formulas env

  let record_this_instance accepted lorig =
    if Options.profiling() then
      match F.view lorig with
      | F.Lemma {F.name;loc} -> Profiling.new_instance_of name loc accepted
      | _ -> assert false

  let profile_produced_terms env lorig nf s trs =
    let st0 =
      List.fold_left (fun st t -> T.subterms st (T.apply_subst s t))
        T.Set.empty trs
    in
    let name, loc, f = match F.view lorig with
      | F.Lemma {F.name;main;loc} -> name, loc, main
      | _ -> assert false
    in
    let st1 = F.ground_terms_rec nf in
    let diff = Term.Set.diff st1 st0 in
    let info = EM.info env in
    let _new = Term.Set.filter (fun t -> not (MT.mem t info)) diff in
    Profiling.register_produced_terms name loc st0 st1 diff _new

  let new_facts env tbox selector substs =
    List.fold_left
      (fun acc ({trigger_formula=f;
	         trigger_age=age; trigger_dep=dep; trigger_orig=lorig;
                 trigger = tr},
	        subst_list) ->
       let cpt = ref 0 in
       let keeped = ref 0 in
       let res = List.fold_left
	  (fun ((gd, ngd) as acc)
	    {Matching.sbs = sbs;
	     sty = sty;
	     gen = g;
	     goal = b;
	     s_term_orig = torig} ->
              incr cpt;
              let sbs =
                if Options.normalize_instances () then
                  Symbols.Map.fold
                    (fun k t mp ->
                       Symbols.Map.add k (X.term_repr tbox t) mp
                    )sbs Symbols.Map.empty
                else sbs
              in
              let s = sbs, sty in
	      match tr.F.guard with
	        | Some a when
		    X.query (Literal.LT.apply_subst s a) tbox == No ->
                  acc
	        | _ ->
		  let nf = F.apply_subst s f in
                  let accepted = selector nf lorig in
                  record_this_instance accepted lorig;
		  if accepted then
		    let p =
		      { F.f = nf;
                        trigger_depth = tr.F.depth;
                        nb_reductions = 0;
		        age = 1+(max g age);
		        mf = true;
		        gf = b;
		        lem = Some lorig;
		        from_terms = torig
		      }
                    in
                    if Options.profiling() then
                      profile_produced_terms env.matching lorig nf s
                        tr.F.content;
                    let dep =
                      if not (Options.proof() || Options.profiling()) then dep
                      else Ex.union dep (Ex.singleton (Ex.Dep lorig))
                    in
                    incr keeped;
                    if b then (* formula's terms are related to the goal *)
                      (p,dep) :: gd, ngd
                    else
                      gd, (p,dep) :: ngd
                  else acc
	  ) acc subst_list
       in
       (*EM.Debug.new_facts_of_axiom lorig tr.F.content !cpt !keeped;*)
       res
      )
      ([], []) substs


  let sort_facts =
    let rec size f = match F.view f with
      | F.Unit(f1,f2) -> max (size f1) (size f2)
      | _             -> F.size f
    in
    fun lf ->
      List.fast_sort
        (fun (p1,_) (p2,_) ->
          let c = size p1.F.f - size p2.F.f in
          if c <> 0 then c
          else F.compare p2.F.f p1.F.f
        ) lf

  let new_facts env tbox selector substs =
    if Options.timers() then
      try
	Options.exec_timer_start Timers.M_Match Timers.F_new_facts;
	let res = new_facts env tbox selector substs in
	Options.exec_timer_pause Timers.M_Match Timers.F_new_facts;
	res
      with e ->
	Options.exec_timer_pause Timers.M_Match Timers.F_new_facts;
	raise e
    else new_facts env tbox selector substs

  let mround env axs tbox selector =
    (*EM.Debug.new_mround ();*)
    Options.tool_req 2 "TR-Sat-Mround";
    let env = mtriggers env axs in
    let substs = EM.query env.matching tbox in
    let gd, ngd = new_facts env tbox selector substs in
    sort_facts gd, sort_facts ngd

  let m_lemmas env tbox selector = mround env env.lemmas tbox  selector

  let m_predicates env tbox selector = mround env env.predicates tbox selector

  module MI = Map.Make (struct type t = int let compare = compare end)

  let retrieve_used_context env dep =
    let deps = Ex.formulas_of dep in
    let used, unlems, unpreds =
      SF.fold
        (fun f ((used, lems, preds) as acc) ->
          if MF.mem f lems then f :: used, MF.remove f lems, preds
          else if MF.mem f preds then f :: used, lems, MF.remove f preds
          else acc
        ) deps ([], env.lemmas, env.predicates)
    in
    let unused = MF.fold (fun f _ acc -> f::acc) unlems [] in
    let unused = MF.fold (fun f _ acc -> f::acc) unpreds unused in
    used, unused


  (*** add wrappers to profile exported functions ***)

  let add_terms env s gf =
    if Options.timers() then
      try
	Options.exec_timer_start Timers.M_Match Timers.F_add_terms;
	let res = add_terms env s gf in
	Options.exec_timer_pause Timers.M_Match Timers.F_add_terms;
	res
      with e ->
	Options.exec_timer_pause Timers.M_Match Timers.F_add_terms;
	raise e
    else add_terms env s gf

  let add_lemma env gf dep =
    if Options.timers() then
      try
	Options.exec_timer_start Timers.M_Match Timers.F_add_lemma;
	let res = add_lemma env gf dep in
	Options.exec_timer_pause Timers.M_Match Timers.F_add_lemma;
	res
      with e ->
	Options.exec_timer_pause Timers.M_Match Timers.F_add_lemma;
	raise e
    else add_lemma env gf dep

  let add_predicate env gf =
    if Options.timers() then
      try
	Options.exec_timer_start Timers.M_Match Timers.F_add_predicate;
	let res = add_predicate env gf in
	Options.exec_timer_pause Timers.M_Match Timers.F_add_predicate;
	res
      with e ->
	Options.exec_timer_pause Timers.M_Match Timers.F_add_predicate;
	raise e
    else add_predicate env gf

  let m_lemmas env tbox selector =
    if Options.timers() then
      try
	Options.exec_timer_start Timers.M_Match Timers.F_m_lemmas;
	let res = m_lemmas env tbox selector in
	Options.exec_timer_pause Timers.M_Match Timers.F_m_lemmas;
	res
      with e ->
	Options.exec_timer_pause Timers.M_Match Timers.F_m_lemmas;
	raise e
    else m_lemmas env tbox selector

  let m_predicates env tbox selector =
    if Options.timers() then
      try
	Options.exec_timer_start Timers.M_Match Timers.F_m_predicates;
	let res = m_predicates env tbox selector in
	Options.exec_timer_pause Timers.M_Match Timers.F_m_predicates;
	res
      with e ->
	Options.exec_timer_pause Timers.M_Match Timers.F_m_predicates;
	raise e
    else m_predicates env tbox selector

end
