(******************************************************************************)
(*     Alt-Ergo: The SMT Solver For Software Verification                     *)
(*     Copyright (C) 2013-2015 --- OCamlPro                                   *)
(*     This file is distributed under the terms of the CeCILL-C licence       *)
(******************************************************************************)

(******************************************************************************)
(*     The Alt-Ergo theorem prover                                            *)
(*     Copyright (C) 2006-2013                                                *)
(*     CNRS - INRIA - Universite Paris Sud                                    *)
(*                                                                            *)
(*     Sylvain Conchon                                                        *)
(*     Evelyne Contejean                                                      *)
(*                                                                            *)
(*     Francois Bobot                                                         *)
(*     Mohamed Iguernelala                                                    *)
(*     Stephane Lescuyer                                                      *)
(*     Alain Mebsout                                                          *)
(*                                                                            *)
(*   This file is distributed under the terms of the CeCILL-C licence         *)
(******************************************************************************)

open Format
open Options

module T = Term
module F = Formula
module MF = F.Map
module SF = F.Set
module Ex = Explanation
module MT = T.Map
module ST = T.Set
module SubstT = Term.Subst

type gsubst = {
  sbs : T.t SubstT.t;
  sty : Ty.subst;
  gen : int ;     (* l'age d'une substitution est l'age du plus vieux
		     terme qu'elle contient *)
  goal : bool;    (* vrai si la substitution contient un terme ayant un lien
		     avec le but de la PO *)
  s_term_orig : T.t list;
}

module type EM_SIG = sig

  type t
  type xenv
  type form

  type trigger_info = {
    trigger : F.trigger;
    trigger_age : int ;  (* age d'un trigger *)
    trigger_orig : form ; (* lemme d'origine *)
    trigger_formula : form ; (* formule associee au trigger *)
    trigger_dep : Ex.t ;
  }

  type term_info = {
    term_age : int ;        (* age du terme *)
    term_from_goal : bool ;   (* vrai si le terme provient du but de la PO *)
    term_from_formula : form option; (* lemme d'origine du terme *)
    term_from_terms : T.t list;
  }

  type info = {
    age : int ; (* age du terme *)
    lem_orig : form list ; (* lemme d'ou provient eventuellement le terme *)
    t_orig : T.t list;
    but : bool  (* le terme a-t-il un lien avec le but final de la PO *)
  }

  val empty : t

  val add_term : term_info -> Term.t -> t -> t
  val add_trigger : trigger_info -> t -> t
  val info : t -> info Term.Map.t
  val query : t -> xenv -> (trigger_info * gsubst list) list
end

module type TH_SIG = sig
  type t
  val are_equal : t -> Term.t -> Term.t -> Sig.answer
  val term_repr : t -> Term.t -> Term.t
  val class_of : t -> Term.t -> Term.t list
end

module EMatch (X : TH_SIG)(Form : sig type t end) : EM_SIG
  with type xenv = X.t and type form = Form.t = struct

  type xenv = X.t
  type form = Form.t

  type trigger_info = {
    trigger : F.trigger;
    trigger_age : int ;  (* age d'un trigger *)
    trigger_orig : form ; (* lemme d'origine *)
    trigger_formula : form ; (* formule associee au trigger *)
    trigger_dep : Ex.t ;
  }

  type term_info = {
    term_age : int ;        (* age du terme *)
    term_from_goal : bool ;   (* vrai si le terme provient du but de la PO *)
    term_from_formula : form option; (* lemme d'origine du terme *)
    term_from_terms : T.t list;
  }

  type info = {
    age : int ; (* age du terme *)
    lem_orig : form list ; (* lemme d'ou provient eventuellement le terme *)
    t_orig : T.t list;
    but : bool  (* le terme a-t-il un lien avec le but final de la PO *)
  }

  type t = {
    fils : T.t list MT.t SubstT.t ;
    info : info MT.t ;
    pats : trigger_info list
  }

  exception Echec

  let empty = {
    fils = SubstT.empty ;
    info = MT.empty ;
    pats = [ ];
  }

  let info t = t.info

  let age_limite = Options.age_bound
  (* l'age limite des termes, au dela ils ne sont pas consideres par le
     matching *)

  (*BISECT-IGNORE-BEGIN*)
  module Debug = struct

    let add_term t =
      if debug_matching() = 2 then
        fprintf fmt "[matching] add_term:  %a@." T.print t

    let matching tr =
      if debug_matching() = 2 then begin
        fprintf fmt "@.[matching] (multi-)trigger: %a@."
          T.print_list tr.F.content;
        fprintf fmt "========================================================@."
      end

    let match_pats_modulo pat lsubsts =
      if debug_matching() = 2 then begin
        fprintf fmt "@.match_pat_modulo: %a  with accumulated substs@."
          T.print pat;
        List.iter (fun {sbs=sbs; sty=sty} ->
          fprintf fmt ">>> sbs= %a | sty= %a@."
            (SubstT.print Term.print) sbs Ty.print_subst sty
        )lsubsts
      end

    let match_one_pat {sbs=sbs; sty=sty} pat0 =
      if debug_matching() = 2 then
        fprintf fmt "@.match_pat: %a  with subst:  sbs= %a | sty= %a @."
          T.print pat0 (SubstT.print Term.print) sbs Ty.print_subst sty


    let match_one_pat_against {sbs=sbs; sty=sty} pat0 t =
      if debug_matching() = 2 then
        fprintf fmt
          "@.match_pat: %a  against term %a@.with subst:  sbs= %a | sty= %a @."
          T.print pat0
          T.print t
          (SubstT.print Term.print) sbs
          Ty.print_subst sty

    let match_term {sbs=sbs; sty=sty} t pat =
      if debug_matching() = 2 then
        fprintf fmt
          "[match_term] I match %a against %a with subst: sbs=%a | sty= %a@."
          T.print pat T.print t (SubstT.print Term.print) sbs Ty.print_subst sty

    let match_list {sbs=sbs; sty=sty} pats xs =
      if debug_matching() = 2 then
        fprintf fmt
          "@.[match_list] I match %a against %a with subst: sbs=%a | sty= %a@."
          T.print_list pats
          T.print_list xs
          (SubstT.print Term.print) sbs
          Ty.print_subst sty

    let match_class_of t cl =
      if debug_matching() = 2 then
        fprintf fmt "class_of (%a)  = { %a }@."
          T.print t
          (fun fmt -> List.iter (fprintf fmt "%a , " T.print)) cl

    let new_facts_of_axiom trigger_origin trs cpt keeped =
      if debug_matching () = 1 && cpt <> 0 then
        let name = match F.view trigger_origin with
            F.Lemma {F.name=s} -> s | _ -> "!(no-name)"
        in
        fprintf
          fmt
          "%2d / %2d instance for axiom %s with trigger %a@.@."
          keeped cpt name T.print_list trs

    let new_mround () =
      if debug_matching () <> 0 then
        fprintf fmt "# [matching] new matching round ...@."

  end
  (*BISECT-IGNORE-END*)

  let infos op_gen op_but t g b env =
    try
      let i = MT.find t env.info in
      op_gen i.age g , op_but i.but b
    with Not_found -> g , b

  let add_term info t env =
    Debug.add_term t;
    let rec add_rec env t =
      if MT.mem t env.info then env
      else
	let {T.f=f; xs=xs} = T.view t in
	let env =
	  let map_f = try SubstT.find f env.fils with Not_found -> MT.empty in

	  (* - l'age d'un terme est le min entre l'age passe en argument
	     et l'age dans la map
	     - un terme est en lien avec le but de la PO seulement s'il
	     ne peut etre produit autrement (d'ou le &&)
	     - le lemme de provenance est le dernier lemme
	  *)
	  let g, b = infos min (&&) t info.term_age info.term_from_goal env in
	  let from_lems =
	    List.fold_left
	      (fun acc t ->
		try (MT.find t env.info).lem_orig @ acc
		with Not_found -> acc)
	      (match info.term_from_formula with None -> [] | Some a -> [a])
	      info.term_from_terms
	  in
	  { env with
	    fils = SubstT.add f (MT.add t xs map_f) env.fils;
	    info =
	      MT.add t
		{ age=g; lem_orig = from_lems; but=b;
	          t_orig = info.term_from_terms }
		env.info
	  }
	in
	List.fold_left add_rec env xs
    in
    if info.term_age > age_limite () then env else add_rec env t

  let add_trigger p env = { env with pats = p :: env.pats }

  let all_terms
      f ty env
      {sbs=s_t; sty=s_ty; gen=g; goal=b;
       s_term_orig=s_torig} lsbt_acc =
    SubstT.fold
      (fun k s l ->
	MT.fold
	  (fun t _ l ->
	    try
	      let s_ty = Ty.matching s_ty ty (T.view t).T.ty in
	      let ng , but =
		try
		  let {age=ng;lem_orig=lem'; but=bt} = MT.find t env.info in
		  max ng g , bt || b
		with Not_found -> g , b
	      in
	      { sbs = SubstT.add f t s_t;
		sty = s_ty;
		gen = ng;
		goal = but;
		s_term_orig = t :: s_torig;
	      }::l
	    with Ty.TypeClash _ -> l
	  ) s l
      ) env.fils lsbt_acc

  let add_msymb tbox f t ({sbs=s_t; sty=s_ty} as sg)=
    try
      if X.are_equal tbox t (SubstT.find f s_t) == Sig.No then raise Echec;
      sg
    with Not_found ->  {sg with sbs=SubstT.add f t s_t; sty=s_ty}

  let (-@) l1 l2 =
    match l1, l2 with
      | [], _  -> l2
      | _ , [] -> l1
      | _ -> List.fold_left (fun acc e -> e :: acc) l2 (List.rev l1)

  let xs_modulo_records t { Ty.lbs = lbs } =
    List.rev
      (List.rev_map
         (fun (hs, ty) -> T.make (Symbols.Op (Symbols.Access hs)) [t] ty) lbs)

  module SLT = (* sets of lists of terms *)
    Set.Make(struct
      type t = T.t list
      let compare l1 l2 =
        try
          List.iter2
            (fun t1 t2 ->
              let c = T.compare t1 t2 in
              if c <> 0 then raise (Util.Compared c)
            ) l1 l2;
          0
        with Invalid_argument _ ->
          List.length l1 - List.length l2
        | Util.Compared n -> n
    end)

  let filter_classes cl tbox =
    if no_Ematching () then cl
    else
      let mtl =
        List.fold_left
          (fun acc xs ->
            let xs = List.rev (List.rev_map (fun t -> X.term_repr tbox t) xs) in
            SLT.add xs acc
          ) SLT.empty cl
      in
      SLT.elements mtl

  let rec match_term env tbox ( {sbs=s_t; sty=s_ty;gen=g;goal=b} as sg) pat t =
    Options.exec_thread_yield ();
    Debug.match_term sg t pat;
    let {T.f=f_pat;xs=pats;ty=ty_pat} = T.view pat in
    match f_pat with
      |	Symbols.Var _ ->
	let sb =
          (try
	     let s_ty = Ty.matching s_ty ty_pat (T.view t).T.ty in
	     let g',b' = infos max (||) t g b env in
	     add_msymb tbox f_pat t
	       { sg with sbs=s_t; sty=s_ty; gen=g'; goal=b' }
	   with Ty.TypeClash _ -> raise Echec)
        in
        [sb]
      | _ ->
        try
          let s_ty = Ty.matching s_ty ty_pat (T.view t).T.ty in
          let cl = if no_Ematching () then [t] else X.class_of tbox t in
          Debug.match_class_of t cl;
          let cl =
	    List.fold_left
	      (fun l t ->
                let {T.f=f; xs=xs; ty=ty} = T.view t in
	        if Symbols.compare f_pat f = 0 then xs::l
                else
                  begin
                    match f_pat, ty with
                      | Symbols.Op (Symbols.Record), Ty.Trecord record ->
                        (xs_modulo_records t record) :: l
                      | _ -> l
                  end
	      )[] cl
          in
          let cl = filter_classes cl tbox in
          let gsb = { sg with sbs = s_t; sty = s_ty } in
          (**** This optim is very expensive***
                if T.is_ground pat && are_equal tbox pat t <> Sig.No then
                [gsb]
                else*****)
          (* pas sur que ce soit correct ici *)
          List.fold_left
            (fun acc xs ->
              try (match_list env tbox gsb pats xs) -@ acc
              with Echec -> acc
            ) [] cl
        with Ty.TypeClash _ -> raise Echec

  and match_list env tbox sg pats xs =
    Debug.match_list sg pats xs;
    try
      List.fold_left2
        (fun sb_l pat arg ->
          List.fold_left
            (fun acc sg ->
              let aux = match_term env tbox sg pat arg in
              (*match aux with [] -> raise Echec | _  -> BUG !! *)
              List.rev_append aux acc
            ) [] sb_l
        ) [sg] pats xs
    with Invalid_argument _ -> raise Echec

  let match_one_pat env tbox pat0 lsbt_acc sg =
    Debug.match_one_pat sg pat0;
    let pat = T.apply_subst (sg.sbs, sg.sty) pat0 in
    let {T.f=f; xs=pats; ty=ty} = T.view pat in
    match f with
      | Symbols.Var _ -> all_terms f ty env sg lsbt_acc
      | _ ->
        let {sty=sty; gen=g; goal=b} = sg in
        let f_aux t xs lsbt =
          Debug.match_one_pat_against sg pat0 t;
	  try
	    let s_ty = Ty.matching sty ty (T.view t).T.ty in
	    let gen, but = infos max (||) t g b env in
            let sg =
              { sg with
                sty = s_ty; gen = gen; goal = but;
                s_term_orig = t::sg.s_term_orig }
            in
	    let aux = match_list env tbox sg pats xs in
            List.rev_append aux lsbt
	  with Echec | Ty.TypeClash _ -> lsbt
        in
	try MT.fold f_aux (SubstT.find f env.fils) lsbt_acc
	with Not_found -> lsbt_acc

  let match_pats_modulo env tbox lsubsts pat =
    Debug.match_pats_modulo pat lsubsts;
    List.fold_left (match_one_pat env tbox pat) [] lsubsts

  let matching env tbox pat_info =
    let pats = pat_info.trigger in
    Debug.matching pats;
    let egs =
      { sbs = SubstT.empty;
        sty = Ty.esubst;
        gen = 0;
	goal = false;
	s_term_orig = [];
      }
    in
    pat_info, List.fold_left (match_pats_modulo env tbox) [egs] pats.F.content

  let query env tbox =
    List.rev_map (matching env tbox) env.pats

  let query env tbox =
    if Options.timers() then
      try
	Options.exec_timer_start Timers.M_Match Timers.F_query;
	let res = query env tbox in
	Options.exec_timer_pause Timers.M_Match Timers.F_query;
	res
      with e ->
	Options.exec_timer_pause Timers.M_Match Timers.F_query;
	raise e
    else query env tbox

end
