(**************************************************************************)
(*                                                                        *)
(*     The Alt-ergo theorem prover                                        *)
(*     Copyright (C) 2006-2010                                            *)
(*                                                                        *)
(*     Sylvain Conchon                                                    *)
(*     Evelyne Contejean                                                  *)
(*     Stephane Lescuyer                                                  *)
(*     Mohamed Iguernelala                                                *)
(*     Alain Mebsout                                                      *)
(*     Cody Roux                                                          *)
(*                                                                        *)
(*     CNRS - INRIA - Universite Paris Sud                                *)
(*                                                                        *)
(*   This file is distributed under the terms of the CeCILL-C licence     *)
(*                                                                        *)
(**************************************************************************)

open Format
open Options
open Sig

module Q = Numbers.Q
module Z = Numbers.Z
module A = Literal
module T = Term
module Sy = Symbols
module I = Intervals
module Ex = Explanation
module MT = Term.Map
module ST = Term.Set
module TG = TimerGappa
module SA = A.LT.Set
module SubstT = Term.Subst

module Make
  (X : Sig.X)
  (Uf : Sig.UF_SIG with type r = X.r)
  (P : Polynome.EXTENDED_Polynome with type r = X.r)
  : RSig.RELATION with type r = X.r
		  (*      and type p = P.pos P.t*)
                      and type uf = Uf.t

= struct

  type r = X.r
  type uf = Uf.t

  (**************************************************************************)
  (**************************************************************************)

  let new_created_terms = ref Term.Set.empty

  module UfWrap = struct

    let class_of = Uf.class_of

    let mem = Uf.mem

    let empty = Uf.empty

    let add uf t = Util.fst3 (Uf.add uf t)

    let find uf t fonc = Uf.find (if mem uf t then uf else add uf t) t fonc

    let are_equal uf s t =
      let uf = if mem uf t then uf else add uf s in
      let uf = if mem uf s then uf else add uf t in
      let x, ex1 = Uf.find uf s (fun () -> assert false) in
      let y, ex2 = Uf.find uf t (fun () -> assert false) in
      if X.equal x y then Yes (Ex.union ex1 ex2, []) else No

  end

  module F_match : sig

    type pat_token =  TBND |  TFIX |  TFLT

    type fpred = Fbnd of Term.t | Ffix of Term.t | Fflt of Term.t

    module MX : Map.S with type key = X.r

    module MF : Map.S with type key = fpred

    module SF : Set.S with type elt = fpred


    val p_equal : fpred -> fpred -> bool
    val p_class_of : uf -> fpred -> fpred list
    val break_pred : fpred -> pat_token * Term.t
    val term_of_pred : fpred -> Term.t
    val set_of_list : fpred list -> SF.t
    val fpred_type : fpred -> Ty.t


    module type THEOREMS = sig

      type p
      module MPpos : Map.S with type key = p
      type bnd_info = I_info of Intervals.t | Fix_info of Intervals.t
                      | Flt_info of Intervals.t
      type bnds =
        {
          i_map   : Intervals.t MPpos.t;
          fix_map : Intervals.t MX.t;
          flt_map : Intervals.t MX.t;
        }
      type info_mod
      type lem =
        {
          concl : fpred;
          hyps : SF.t;
          name : string;
          apply : info_mod;
        }
      module Lemmas : Set.S with type elt = lem
      type info = {bnds : bnds ; abs_tms : Term.t Term.Map.t}
      type thm

      val run_info : info_mod -> uf -> info -> info

      val is_zero : uf -> Term.t -> info -> bool
      val is_const: uf -> Term.t -> info -> bool
      val find_val : uf -> Term.t -> 'a MX.t -> 'a
      val find_val_nf : uf -> Term.t -> Intervals.t MPpos.t -> Intervals.t
      val mem_val : uf -> Term.t -> 'a MX.t -> bool


      val update_const : uf -> fpred -> Numbers.Q.t -> info -> info
      val update_info : uf -> fpred -> bnd_info -> info -> info

      val instances : uf -> SF.t -> info -> Lemmas.t

    end

    val print_fpred : Format.formatter -> fpred -> unit
    val print_fpred_list : Format.formatter -> fpred list -> unit

    module Flt_thms (P : Polynome.EXTENDED_Polynome with type r = X.r)
      : THEOREMS with type p = P.t

  end = struct

    let (-@) = List.rev_append

    let bnd_name = Symbols.name "@BND"
    let fix_name = Symbols.name "@FIX"
    let flt_name = Symbols.name "@FLT"

    exception F_match_fail
    exception Not_a_value

    module MX = Map.Make(struct type t = X.r let compare t u = X.str_cmp t u end)


    (* ----------------------- matching using the existing module ----------- *)

    (** The type of predicates on terms of the floating-point theory *)
    type fpred = Fbnd of Term.t | Ffix of Term.t | Fflt of Term.t

    (** A syntactic type of patterns on terms *)
    type fpat =
    | Var of string
    | Bop of fpat * Sy.operator * fpat
    | Fixp of string * string * fpat
    | Fltp of string * string * string * fpat
    | Sqrtp of fpat
    | Absp of fpat
    | Real_of_intp of fpat
    | Wildcard
    | Trm of T.t

    (**Patterns on predicates*)
    type ppat = BND of fpat | FIX of fpat | FLT of fpat

    let (+^) x y = Bop (x, Sy.Plus, y)
    let (-^) x y = Bop (x, Sy.Minus, y)
    let ( *^) x y = Bop (x, Sy.Mult, y)
    let (/^) x y = Bop (x, Sy.Div, y)

    let print_fpred fm t = match t with
        Fbnd t -> fprintf fm "BND(%a)" T.print t
      | Ffix t -> fprintf fm "FIX(%a)" T.print t
      | Fflt t -> fprintf fm "FLT(%a)" T.print t

    let rec print_fpred_list fm ts = match ts with
      | [] -> ()
      | [t] -> print_fpred fm t
      | t::l -> Format.fprintf fm "%a,%a" print_fpred t print_fpred_list l

    let pat_to_string () p =
      let term_to_string () t =
        ignore (flush_str_formatter ());
        Term.print str_formatter t;
        flush_str_formatter ()
      in
      let rec pat_to_string () p = match p with
          Var x -> x
        | Bop (p,o,q) ->
          "("^ pat_to_string () p ^ Sy.to_string (Sy.Op o) ^ pat_to_string () q ^")"
        | Fixp (n,d,p) -> "fix("^n^","^d^","^pat_to_string () p^ ")"
        | Fltp (p,e,d,q) -> "float("^p^","^e^","^d^","^pat_to_string () q^ ")"
        | Absp p -> "|" ^ pat_to_string () p ^ "|"
        | Sqrtp p -> "sqrt("^pat_to_string () p^ ")"
        | Real_of_intp p -> "real_of_int("^pat_to_string () p^")"
        | Wildcard -> "_"
        | Trm t -> term_to_string () t
      in
      pat_to_string () p

    let rec fpat_to_string () p = match p with
      | BND p -> "BND(" ^ pat_to_string () p ^ ")"
      | FLT p -> "FLT(" ^ pat_to_string () p ^ ")"
      | FIX p -> "FIX(" ^ pat_to_string () p ^ ")"

    (* we make the following associations:
       BND -> TBND
       FIX -> TFIX
       FLT -> TFLT
    *)
    type pat_token =  TBND |  TFIX |  TFLT

    let make_pat n p = match n with
        TBND -> BND p
      | TFIX -> FIX p
      | TFLT -> FLT p

    type trigger = Eq of fpat * fpat | Not of trigger
                   | Const of fpat | Zero of fpat | True
                   | Conj of trigger list


    (* comparing predicates
       uses structural comparison and not semantic
       comparison *)
    let p_compare t1 t2 =
      if t1 == t2 then 0 else
        match t1, t2 with
          Fbnd t1, Fbnd t2
        | Ffix t1, Ffix t2
        | Fflt t1, Fflt t2 -> T.compare t1 t2
        | Fbnd _, _ -> -1
        | Fflt _, _ -> 1
        | Ffix _ , Fbnd _ -> 1
        | Ffix _ , _ -> -1

    let p_equal t1 t2 = match t1, t2 with
        Fbnd t1, Fbnd t2
      | Ffix t1, Ffix t2
      | Fflt t1, Fflt t2 -> T.equal t1 t2
      | _ -> false

    let make_pred n t = match n with
        TBND -> Fbnd t
      | TFIX -> Ffix t
      | TFLT -> Fflt t

    let make_pred_pat n p = match n with
        TBND -> BND p
      | TFIX -> FIX p
      | TFLT -> FLT p

    let break_pred t = match t with
        Fbnd t -> TBND, t
      | Ffix t -> TFIX, t
      | Fflt t -> TFLT, t

    let term_of_pred t = snd (break_pred t)

    let pred_as_term t = match t with
      | Fbnd t -> T.make bnd_name [t] Ty.Treal
      | Ffix t -> T.make fix_name [t] Ty.Treal
      | Fflt t -> T.make flt_name [t] Ty.Treal

    let fpred_type t =
      let {T.ty=ty} = T.view (term_of_pred t) in
      ty


    module MF = Map.Make
      (struct type t = fpred
	      let compare = p_compare
       end)

    module SF = Set.Make
      (struct type t = fpred
	      let compare = p_compare
       end)

    let rec set_of_list = List.fold_left (fun s t -> SF.add t s) SF.empty


    (** Class of a predicate (simple wrapper around UF.class_of) *)
    let p_class_of uf t = let k, t = break_pred t in
		          let t_class = UfWrap.class_of uf t in
		          List.map (make_pred k) t_class


    module type THEOREMS = sig

      type p
      module MPpos : Map.S with type key = p
      type bnd_info = I_info of Intervals.t | Fix_info of Intervals.t
                      | Flt_info of Intervals.t
      type bnds =
        {
          i_map   : Intervals.t MPpos.t;
          fix_map : Intervals.t MX.t;
          flt_map : Intervals.t MX.t;
        }
      type info_mod
      type lem =
        {
          concl : fpred;
          hyps : SF.t;
          name : string;
          apply : info_mod;
        }
      module Lemmas : Set.S with type elt = lem
      type info = {bnds : bnds ; abs_tms : Term.t MT.t}
      type thm

      val run_info : info_mod -> uf -> info -> info

      val is_zero : uf -> Term.t -> info -> bool
      val is_const: uf -> Term.t -> info -> bool
      val find_val : uf -> Term.t -> 'a MX.t -> 'a
      val find_val_nf : uf -> Term.t -> Intervals.t MPpos.t -> Intervals.t
      val mem_val : uf -> Term.t -> 'a MX.t -> bool


      val update_const : uf -> fpred -> Q.t -> info -> info
      val update_info : uf -> fpred -> bnd_info -> info -> info

      val instances : uf -> SF.t -> info -> Lemmas.t

    end


    module Flt_thms (P : Polynome.EXTENDED_Polynome with type r = X.r) = struct

      let alien_of p = match P.is_monomial p with
        | Some (a,x,b) when Q.equal a Q.one && Q.sign b = 0 -> x
        | _ -> P.embed p

      let poly_of r =
        match P.extract r with
        | Some p -> p
        | None ->
          P.create
            [Numbers.Q.one, r] Numbers.Q.zero (X.type_info r)


      type p = P.t

      module MPpos =
        Map.Make
          (struct
            type t = P.t
            let compare = P.compare
           end)

      (** The type of bounds on terms: a term can be mapped to:
          - an interval of real numbers
          - an upper bound on the number of bits in its fixed point representation
          - a lower bound on the precision required to represent it as a floating-point
          number. *)
      type bnds = {i_map : I.t MPpos.t ; fix_map : I.t MX.t ; flt_map : I.t MX.t}

      (** The information we have on terms:
          -bounds
          -the absolute value of the term appears
      *)
      type info = {bnds : bnds ; abs_tms : Term.t MT.t}


      (** Encapsulates the information encoded in the bnds maps *)
      type bnd_info = I_info of I.t | Fix_info of I.t
                      | Flt_info of I.t

      (* wrapper for the type of lemma applications *)
      type info_mod = Info_mod of (uf -> info -> info)


      let run_info = fun (Info_mod b) -> b
      (* The type of lemmas, to be applied to known intervalls *)
      (* Takes the map of bounds of elements in hyps and updates the
         bound for concl using the apply function*)
      type lem = { concl : fpred ; hyps : SF.t ;  name : string ;
	           apply : info_mod }


      let lem_compare (t,ts,s) (u,us,r) =
        let n = p_compare t u in
        if n = 0 then
          let m = SF.compare ts us in
          if m = 0 then Pervasives.compare s r
          else m
        else n

      (** Sets of lemmas *)
      module Lemmas = Set.Make
        (struct
          type t = lem
          let compare l1 l2 = lem_compare (l1.concl,l1.hyps,l1.name)
            (l2.concl,l2.hyps,l2.name)
         end)


      (** The type of theorems: theorems can be applied to a predicate on terms to
          yield a set of lemmas *)
      type thm = {apply_thm : uf -> Matching.gsubst list -> info -> Lemmas.t;
                  trig_thm : fpred;
                  tag : int}

      module Form = struct
        type t = thm

        let compare t u = Pervasives.compare t.tag u.tag
      end

      module Match = Matching.EMatch(Uf)(Form)


      let trig_info_of_thm t trig =
        let tr =
          {Formula.content = [trig];
           depth = (Term.view trig).Term.depth;
           guard = None}
        in
        { Match.trigger   = tr;
          trigger_age     = 0;
          (* This field is unused *)
          trigger_orig    = Obj.magic ();
          trigger_formula = t;
          trigger_dep     = Ex.empty}


      let empty_info =
        {Match.term_age = 0;
         term_from_goal = false;
         term_from_formula = None;
         term_from_terms = []}

      (* --------------------- A DSL for pattern matching on terms --------------------*)


      (* The type of substitutions which valuates variables represented by
         strings which represent either terms, constants or a rounding
         mode *)
      type subst = Matching.gsubst


      let var_of_string str = Symbols.Var (Hstring.make str)

      let get_int t =
        let {T.f = f} = T.view t in
        match f with
          Sy.Int n -> Q.from_string (Hstring.view n)
        | _ -> raise Not_a_value

      let get_mode t = t

      let int_of_subst sbt s =
        try
          get_int (SubstT.find (var_of_string s) sbt.Matching.sbs)
        with Not_found -> raise F_match_fail

      let mode_of_subst sbt s =
        try
          get_mode (SubstT.find (var_of_string s) (sbt.Matching.sbs))
        with Not_found -> raise F_match_fail


      let make_pat_var str ty = T.make  (var_of_string str) [] ty

      let rec make_term_pat p ty =
        match p with
          Var x -> make_pat_var x ty
        | Bop (x, o, y) ->
          begin
	    match o with
            | (Sy.Plus | Sy.Minus | Sy.Mult | Sy.Div) ->
              let a = make_term_pat x ty in
              let b = make_term_pat y ty in
              T.make (Sy.Op o) [a;b] ty
            | _ -> assert false
          end
        | Fixp (sp, sd, x) ->
          let p = make_pat_var sp Ty.Tint in
          let d = make_pat_var sd Builtin.FPA.ty_mode in
          let a = make_term_pat x ty in
          T.make (Sy.Op Sy.Fixed) [p;d;a] Ty.Treal
        | Fltp (sp, se, sd, x) ->
          let p = make_pat_var sp Ty.Tint in
          let e = make_pat_var se Ty.Tint in
          let d = make_pat_var sd Builtin.FPA.ty_mode in
          let a = make_term_pat x ty in
          T.make (Sy.Op Sy.Float) [p;e;d;a] Ty.Treal
        | Sqrtp x ->
          T.make (Sy.Op Sy.Sqrt) [make_term_pat x ty] Ty.Treal
        | Absp x ->
          T.make (Sy.Op Sy.Abs) [make_term_pat x ty] ty
        | Real_of_intp x ->
          T.make (Sy.Op Sy.Real_of_int) [make_term_pat x Ty.Tint] Ty.Treal
        | Trm t -> t
        | Wildcard -> T.fresh_name ty

      let make_fpred_pat p ty =
        match p with
        | BND p -> Fbnd (make_term_pat p ty)
        | FLT p -> Fflt (make_term_pat p ty)
        | FIX p -> Ffix (make_term_pat p ty)



      (* let make_num n = match n with *)
      (*   | Int n -> T.int (string_of_int n) *)
      (*   | _ -> assert false *)

      (** Returns the term equal to the application of a substitution
          to a pattern. *)
      let rec make_term subst p ty =
        T.apply_subst (subst.Matching.sbs, subst.Matching.sty) (make_term_pat p ty)

      let make_fpred subst p ty =
        match p with
        | BND p -> Fbnd (make_term subst p ty)
        | FLT p -> Fflt (make_term subst p ty)
        | FIX p -> Ffix (make_term subst p ty)


      (* an interval similarity: similar i c d is the interval
         (i + c) * d *)
      let similar i c d ty = I.scale d
        (I.add i (I.point c ty Ex.empty))

      (* an inverse similarity: unsimilar i c d = i / d - c *)
      let unsimilar i c d ty =
        assert (not (Q.equal d Q.zero));
        I.sub (I.scale (Q.inv d) i) (I.point c ty Ex.empty)

      let nf_of_term uf t =
        if false then fprintf fmt "[f_match] nf_of_term %a@." T.print t;
        let x, _ = UfWrap.find uf t (fun () -> assert false) in
        P.normal_form_pos (poly_of x)

      let find_val_nf uf t bnds =
        let p, c, d = nf_of_term uf t in
        similar (MPpos.find p bnds) c d (P.type_info p)

      let find_val uf t bnds =
        let r, _ = UfWrap.find uf t (fun () -> assert false) in
        MX.find r bnds

      let mem_val uf t bnds =
        let r,_ = UfWrap.find uf t (fun () -> assert false) in
        MX.mem r bnds

      let is_const uf t info =
        if false then fprintf fmt "[f_match] is_const %a@." T.print t;
        let p, _, _ = nf_of_term uf t in
        if P.is_empty p then true else
          try
            let i = MPpos.find p info.bnds.i_map in
            match I.is_point i with
            | Some _   -> true
            | None     -> false
          with Not_found -> false

      let is_zero uf t info =
        let p, c, d = nf_of_term uf t in
        if P.is_empty p then
          Q.equal c Q.zero
        else
          try
            let i = MPpos.find p info.bnds.i_map in
            match I.is_point i with
            | Some (x,_) -> Q.equal (Q.add x c) Q.zero
            | None   -> false
          with Not_found -> false

      (* Takes as input a trigger and an "environment" (semantic equality and bounds)
         and returns true if the trigger is satisfied. *)
      let rec make_trig uf subst info tr ty =
        (*fprintf fmt "[f_match] make_trig %a@." T.print tr;*)
        match tr with
        | True -> true
        | Eq (t,u) ->
          let t, u = make_term subst t ty, make_term subst u ty in
          begin
            match UfWrap.are_equal uf t u with
            | Sig.Yes _ -> true
            | Sig.No -> false
          end
        | Not tr -> not (make_trig uf subst info tr ty)
        (* For zero and constant we assume the term has already been
           checked by gappa, but this may not be the case depending
           on the order of application of the theorems *)
        | Const t ->
          let t = make_term subst t ty in
          if false then fprintf fmt "@.[f_match] make_trig %a@." T.print t;
          is_const uf t info
        | Zero t ->
          let t = make_term subst t ty in
          is_zero uf t info
        | Conj trs -> List.for_all (fun tr -> make_trig uf subst info tr ty) trs


      (*--------------------------------------------------------------------------------*)




      (** Add a new interval for t into a map of intervals: takes the
          intersection of the new interval and the old one if it
          exists.*)
      let update_i p i bnds =
        let {i_map = i_map} = bnds in
        try
          let it = MPpos.find p i_map in
          let i_n_it = I.intersect it i in
          {bnds with i_map = MPpos.add p i_n_it i_map}
        with
          Not_found -> {bnds with i_map = MPpos.add p i i_map}
        | I.NotConsistent expl ->
          if Options.debug_fpa() then
            (fprintf fmt "[f_match] interval inconsistent %a@."
	       Ex.print expl);
          raise (Sig.Inconsistent (expl,[]))

      (* Takes the semantic value of a term and updates the bound
         on its normal form. *)
      let update_i_nf r i bnds =
        let p, c, d = P.normal_form_pos (poly_of r) in
        update_i p (unsimilar i c d (P.type_info p)) bnds

      (** updates the value of the fix point representation for a term *)
      let update_fix r e bnds =
        let {fix_map = fix_map} = bnds in
        try
          let e_t = MX.find r fix_map in
          (* let m = max_num e e_t in *)
          let m = I.intersect e e_t in
          {bnds with fix_map = MX.add r m fix_map}
        with Not_found ->
          {bnds with fix_map = MX.add r e fix_map}

      (**updates the value of the floating-point representation for a term *)
      let update_flt r e bnds =
        let {flt_map = flt_map} = bnds in
        try
          let e_t = MX.find r flt_map in
          (* let m = min_num e e_t in *)
          let m = I.intersect e e_t in
          {bnds with flt_map = MX.add r m flt_map}
        with Not_found ->
          {bnds with flt_map = MX.add r e flt_map}


      let update_info_sem r bound info =
        let bnds = match bound with
            I_info i   -> update_i_nf r i info.bnds
          | Fix_info l -> update_fix  r l info.bnds
          | Flt_info p -> update_flt  r p info.bnds
        in {info with bnds = bnds}

      (** Given an uf environment, a syntactic term, a bound and an info value,
          updates the information of the semantic value of the term and
          returns the updated bound *)
      let update_info uf t bound info =
        let t = term_of_pred t in
        let r,_ = UfWrap.find uf t (fun () -> assert false) in
        update_info_sem r bound info

      (** Same as previously, but takes a constant c instead of an info value,
          and give the constant bound [c,c] to the term, as well as the
          fixed-point and floating-point bounds if possible. *)
      let update_const uf t c info =
        let p, u = break_pred t in
        let {T.ty=ty} = T.view u in
        match p with
        | TBND -> let i = I.point c ty Ex.empty in
                  update_info uf t (I_info i) info
        | TFIX ->
          let l = Float_utils.log_2 c in
          let undef = I.undefined Ty.Tint in
          let i = I.new_borne_inf Ex.empty (Q.from_int l) true undef in
          update_info uf t (Fix_info i) info
        | TFLT ->
          let i =
            try
              let m = Float_utils.mantissa c in
              I.mk_closed
                Q.zero (Q.from_int m) true true Ex.empty Ex.empty Ty.Tint
            with Float_utils.Infinity -> I.undefined Ty.Tint
          in update_info uf t (Flt_info i) info

      let get_i uf t bnds =
        match t with
        | Fbnd t -> I_info   (find_val_nf uf t bnds.i_map)
        | Ffix t -> Fix_info (find_val uf t bnds.fix_map)
        | Fflt t -> Flt_info (find_val uf t bnds.flt_map)


      (*Creates a bound update according to a function which
        takes a list of pairs of term*intervall and returns the
        updated interval. *)
      let update_by_f f ts t s = Info_mod (fun uf info ->
        let t_i = List.map
          (fun t ->
            let i = try get_i uf t info.bnds
              (* TODO: have a case for the other constructors *)
              with Not_found ->
                I_info (I.undefined (fpred_type t))
            in
            t, i
          ) ts
        in
        update_info uf t (f t_i s) info)

      (** Takes as input:
          - a pattern p and a list of patterns ps
          - a function f as above
          - a trigger trig
          - a string name
          and returns the theorem:
          p1 : I1 /\.../\ pn : In --> p : f(p1,I1;...;pn,In)
      *)
      (*TODO: understand why 2 instances of uf are needed *)
      let lemmas_by_f p ps f trig name = fun uf substs info ->
        if false then fprintf fmt "@.[lemmas_by_f] name = %s@." name;
        try
          (* let ty = fpred_type t in *)
          let ty = Ty.Treal in
          (* let s = p_match uf p t ty in *)
          List.fold_left
            (fun lems s ->
              if make_trig uf s info trig ty then
                let ts = List.map (fun p -> make_fpred s p ty) ps in
                let concl = make_fpred s p ty in
                Lemmas.add
                  {
                    concl = concl;
                    hyps = set_of_list ts;
                    name = name;
                    apply = update_by_f f ts concl s
                  }
                  lems
              else lems
            ) Lemmas.empty substs
        with F_match_fail -> Lemmas.empty


      (** p <= qs, f , n generates the lemma of name "n" that matches on
          predicate p, depends on the bounds of qs and updates the bounds
          for p using the functon f.*)
      (* We suppose that the type of each qs is the same as for p *)
      let (<:=) p (qs, f, trig, name) =
        {
          apply_thm = lemmas_by_f p qs f trig name;
          trig_thm = make_fpred_pat p Ty.Treal;
          tag = Hashtbl.hash name
        }


      let id_bnd l s = match l with
          [x,i] -> i
        | _ -> assert false

      (*We introduce a Prolog-style notation to define rewrite rules:
        p <:- q signifies that all bounds on term q are also bounds on term p:
        we generate the corresponding lemmas *)
      let (<:-) p (q, tr) =
        let name = pat_to_string () p ^ " <:- " ^ pat_to_string () q in
        let tag = Hashtbl.hash name in
        let thm_bnd =
          { apply_thm = (fun uf substs bnd ->
            lemmas_by_f (BND p) [BND q] id_bnd tr name uf substs bnd);
            trig_thm = make_fpred_pat (BND p) Ty.Treal;
            tag = tag}
        in
        let thm_fix =
          { apply_thm = (fun uf substs bnd ->
            lemmas_by_f (FIX p) [FIX q] id_bnd tr name uf substs bnd);
            trig_thm = make_fpred_pat (FIX p) Ty.Treal;
            tag = tag}
        in
        let thm_flt =
          { apply_thm = (fun uf substs bnd ->
            lemmas_by_f (FLT p) [FLT q] id_bnd tr name uf substs bnd);
            trig_thm = make_fpred_pat (FLT p) Ty.Treal;
            tag = tag}
        in
        [thm_bnd;thm_fix;thm_flt]



      (* ------------------------------ theorems --------------------------------------- *)

      (* Error interval for a floating point approximation:
         given an interval o(x) in I, compute the interval that contains
         o(x) - x
      *)
      let compute_error prec exp mode x =
        try
          let inf_x, e_inf, _ = I.borne_inf x in
          let sup_x, e_sup, _ = I.borne_sup x in
          let bnd_x = Q.max (Q.abs inf_x) (Q.abs sup_x) in
          let error = Float_utils.compute_error prec exp mode bnd_x in
          (* [-error, error] *)
          let u = Ex.union e_inf e_sup in
          I.mk_closed (Q.minus error) error true true u u Ty.Treal

        (* In this case there is no supremum or infimum on x *)
        with I.No_finite_bound -> I.undefined Ty.Treal


      (* bound on the distance between x and float(x) given a bound on
         float(x) *)
      let compute_error_bis prec exp mode o_x =
        (* TODO: rewrite! *)
        compute_error prec exp mode o_x


      (* The theorem which computes simple error bounds on floating point approx *)
      let float_err_lem uf subst bnd =
        (* let t = term_of_pred trm in *)
        let x = Var "x" in
        let ox = Fltp ("p","e","d", x) in
        let error cmpt l s =
          try
            let p, e, d = int_of_subst s "p", int_of_subst s "e", mode_of_subst s "d" in
            match l with
            | [_, I_info i] -> I_info (cmpt p e d i)
            | _ -> assert false
          with Not_a_value -> I_info (I.undefined Ty.Treal)
        in
        Lemmas.union
          (lemmas_by_f (BND (ox -^ x)) [BND x] (error compute_error)
             True  "float_error" uf subst bnd)
          (lemmas_by_f (BND (ox -^ x)) [BND ox] (error compute_error_bis)
             True "float_error_bis" uf subst bnd)

      let float_err_thm =
        let x = Var "x" in
        let ox = Fltp ("p","e","d", x) in
        {
          apply_thm = float_err_lem ;
          trig_thm = make_fpred_pat (BND ox) Ty.Treal;
          tag = Hashtbl.hash "float_err_lem"
        }



      (* could map float_of_num to every interval but here we just take
         the convex closure of i *)
      let float_of_i l prec exp mode =
        match l with
        | [_, I_info i] -> begin
          try
            let (l,el, _) = I.borne_inf i in
            let (h,eh, _) = I.borne_sup i in
            let e = Ex.union el eh in
            let fl = Float_utils.float_of_const prec exp mode l in
            let fh = Float_utils.float_of_const prec exp mode h in
            let j = I.mk_closed fl fh true true e e Ty.Treal in
            I_info j
          with I.No_finite_bound -> I_info (I.undefined Ty.Treal)
        end
        | _ -> assert false

      (* Theorem that computes a bound on float(x) given a bound on x *)
      let float_bnd_thm =
        let ox = Fltp ("p","e","d", Var "x") in
        let x = Var "x" in
        let float_bnd l s =
          try
            (float_of_i l
               (int_of_subst s "p") (int_of_subst s "e") (mode_of_subst s "d"))
          with Not_a_value -> I_info (I.undefined Ty.Treal)
        in
        (BND ox) <:= ([BND x], float_bnd, True, "float_bound")

      let op_i uf o ts ss =
        match ts with
          [_, I_info ia;_, I_info ib] ->
            begin
              match o with
              | Sy.Plus  -> I_info (I.add ia ib)
              | Sy.Minus -> I_info (I.sub ia ib)
              | Sy.Mult  -> I_info (I.mult ia ib)
              | Sy.Div   -> I_info (I.div ia ib)
              | _ -> assert false
            end
        | _ -> assert false


      let op_fix uf o ts ss =
        match ts with
          [_, Fix_info ea;_, Fix_info eb] ->
            begin
              match o with
              | Sy.Plus  -> Fix_info (I.intersect ea eb)
              | Sy.Minus -> Fix_info (I.intersect ea eb)
              | Sy.Mult  -> Fix_info (I.add ea eb)
              | _ -> assert false
            end
        | _ -> assert false


      let op_flt uf o ts ss =
        match ts with
          [_, Flt_info ea;_, Flt_info eb] ->
            begin
              match o with
              | Sy.Mult -> Flt_info (I.add ea eb)
              | _ -> assert false
            end
        | _ -> assert false



      let op_map k = match k with
          TBND -> op_i
        | TFIX -> op_fix
        | TFLT -> op_flt


      let destruct_op_tag op = match op with
        | Sy.Plus | Sy.Minus -> "destruct_op_lin"
        | _ -> "destruct_op"

      let destruct_op pred op =
        let x, y = Var "x", Var "y" in
        let x_op_y = Bop (x,op,y) in
        {apply_thm = (fun uf substs bnds ->
          lemmas_by_f
            (make_pred_pat pred x_op_y)
            [make_pred_pat pred x; make_pred_pat pred y]
            (op_map pred uf op)
            (Conj [Not (Const x); Not (Const y)])
            (destruct_op_tag op)
            uf substs bnds);
         trig_thm = make_fpred_pat (make_pred_pat pred x_op_y) Ty.Treal;
         tag = Hashtbl.hash "destruct_op"}

      (* destruct_op_thms =
         [destruct_op p o | p <- TBND,TFIX,TFLT ; o <- Plus,Minus,Mult,Div] *)
      let destruct_op_thms = List.flatten (
        List.map (fun p ->
          List.map (fun o -> destruct_op p o) [Sy.Plus;Sy.Minus;Sy.Mult;Sy.Div])
          [TBND;TFIX;TFLT])


      (* Theorems for the absolute value of a term *)
      let abs_thm =
        let x = Var "x" in
        let abs_x = Absp x in
        let abs_bnds l s =
          match l with
          | [_, I_info i] -> I_info (I.abs i)
          | _ -> assert false
        in
        (BND abs_x) <:= ([BND x], abs_bnds, Not (Const x), "abs_bnd")



      let sba_thm =
        let x = Var "x" in
        let abs_x = Absp x in
        let sba_bnds l s =
          match l with
          | [_, I_info i] ->
            let minus_i = I.scale Q.m_one i in
            let i = I.merge i minus_i in
            I_info i
          | _ -> assert false
        in
        {
          apply_thm =
            lemmas_by_f (BND x) [BND abs_x] sba_bnds (Not (Const x)) "sba_bnd";
          trig_thm = make_fpred_pat (BND abs_x) Ty.Treal;
          tag = Hashtbl.hash "sba_thm"
        }



      (* theorems for the difference of a product *)
      let prod_diff_l =
        let a,b,c,d = (Var "a", Var "b", Var "c", Var "d") in
        let not_const = Conj
          [Not (Const a); Not (Const b) ; Not (Const c) ; Not (Const d)] in
        (a *^ b -^ c *^ d) <:- ( (a -^ c) *^ b +^ c *^ (b -^ d), not_const)


      let prod_diff_l_mod =
        let a,b,c,d = (Var "a", Var "b", Var "c", Var "d") in
        let not_const = Conj
          [Not (Const a); Not (Const b) ; Not (Const c) ; Not (Const d)] in
        let prod_diff l s = match l with
          | [_, I_info u;
             _, I_info v;
             _, I_info w;
             _, I_info x] -> I_info (I.add (I.mult u v) (I.mult v w))
          | _ -> assert false
        in
        (BND (a *^ b -^ c *^ d)) <:= ([BND (a -^ c); BND b; BND c; BND (b -^ d)],
                                      prod_diff , not_const , "prod_diff")

      let prod_diff_r =
        let a,b,c,d = (Var "a", Var "b", Var "c", Var "d") in
        let not_const = Conj
          [Not (Const a); Not (Const b) ; Not (Const c) ; Not (Const d)] in
        (a *^ b -^ c *^ d) <:- ( (a -^ d) *^ b +^ d *^ (b -^ c), not_const )

      let sum_diff =
        let a,b,c,d = (Var "a", Var "b", Var "c", Var "d") in
        ((a +^ b) -^ (c +^ d)) <:- ( (a -^ c) +^ (b -^ d), True)

      let sum_bnds op l subst =
        match l with
        | [_, I_info i1 ; _, I_info i2] -> I_info (op i1 i2)
        | _ -> assert false

      let rm_float_minus_comp_l =
        let x, y = Var "x", Var "y" in
        let p, e, mode = "p", "e", "d" in
        let x_y = x -^ y in
        let ox_x = Fltp(p,e,mode,x) -^ x in
        let not_const = Conj [Not (Const x); Not (Const y); Not (Const x_y)] in
        BND (Fltp (p,e,mode,x) -^ y) <:=
          ([BND x_y; BND ox_x ],sum_bnds I.add, not_const, "rm_float_minus_l")

      let rm_float_minus_comp_r =
        let x, y = Var "x", Var "y" in
        let p, e, mode = "p", "e", "d" in
        let x_y = x -^ y in
        let oy_y = Fltp(p,e,mode,y) -^ y in
        let not_const = Conj [Not (Const x); Not (Const y)] in
        BND (x -^ Fltp (p,e,mode,y)) <:=
          ([BND x_y; BND oy_y ],sum_bnds I.sub, not_const, "rm_float_minus_r")


      (* type mismatch here: works because x is a variable and
         its type is therefore preserved *)
      let real_of_int_thm =
        let x = Var "x" in
        Real_of_intp x <:- (x, Not (Const x))

      let sqrt_thm =
        let x = Var "x" in
        let sqrt_bnd l s = match l with
          | [_, I_info i] -> I_info (I.root 2 i)
          | _ -> assert false
        in
        BND (Sqrtp x) <:= ([BND x], sqrt_bnd, True, "sqrt_bnd")


      (* The list of theorems *)
      let thms =
        (* theorems for arithmetic operations *)
        destruct_op_thms -@
          (* Theorems for abs, sqrt, real_of_int and floating point operations*)
          [float_err_thm; float_bnd_thm;
           abs_thm; sba_thm;
           sqrt_thm] -@
          (* Structural theorems *)
          [rm_float_minus_comp_l; rm_float_minus_comp_r] -@
          (* Rewrite rules *)
          prod_diff_l -@ real_of_int_thm

      (* --------------------------------------------------------------------------- *)


      let thm_match_env = List.fold_left
        (fun env thm -> let {trig_thm = trig} = thm in
                        let trig = pred_as_term trig in
                        Match.add_trigger (trig_info_of_thm thm trig) env)
        Match.empty thms

      let instances uf ts bnds =
        let match_env =
          SF.fold
            (fun t env ->
              Match.add_term empty_info (pred_as_term t) env)
            ts thm_match_env
        in
        let insts = Match.query match_env uf in
        List.fold_left
          (fun ls (info, substs) ->
            let {Match.trigger_formula = thm} = info in
            Lemmas.union ls (thm.apply_thm uf substs bnds))
          Lemmas.empty insts

    end
  end
  (* end module F_match *)

  (**************************************************************************)
  (**************************************************************************)

  (*= AUXILIARY MODULES AND TYPES ============================================*)

  module FMATCH = F_match
  module MX  = FMATCH.MX (* change this *)
  module SFM = FMATCH.SF
  module MFM = FMATCH.MF

  module FLOATS = FMATCH.Flt_thms(P)
  module LEMMAS = FLOATS.Lemmas
  module MPpos  = FLOATS.MPpos (*change this*)

  module SP = Set.Make(P)

  (*module IC = Fm.Make(X)(UF)(P)*)
  module IC = IntervalCalculus.Make(X)(Uf)(P)
  module MPF = IC.MP(*IC.MPpos*)
  (* module S = Simplex.Make(X)(P) *)
  module SIM = Incremental_simplex.Make(X)(P)

  module LR =
    A.Make(struct
      include X
      type t = X.r
      let compare = X.str_cmp
    end)

  (*= MAIN TYPES =============================================================*)

  (* The type of dependencies: a term may have a bound that is proved by
     a set of lemmas, there are lemmas that can only be applied when the
     bound of a term is known *)
  type deps = {proved_by : LEMMAS.t ; implies : LEMMAS.t}


  (* The state consists of:
     (1)- lemmas from gappa deemed necessary for proving the literals
     (2)- the dependencies on those lemmas
     (3)- the bounds on the atomic elements already deduced
     (4)- the current state of the fourier module
     (5)- the lemmas which -may- be necessary to prove a lemma:
     used to rebuild the set of lemmas if we add new litterals
     (6)- the list of terms which correspond to each normal form:
     given p returns the list of (t,c,d) such that
     (p + c)*d = t
     (7)- the set of terms for which we have bounds.
     (8)- the current state of UF
  *)
  type t =
    {
      (*1*)  lems : LEMMAS.t ;
      (*2*)  deps : deps MFM.t ;
      (*3?*) info : FLOATS.info ;
      (*?*)  maybe_deps : deps MFM.t ;
      (*4*)  fm_env : IC.t ;
      (*5*)  maybe_lems : LEMMAS.t ;
      (*6?*) constr : (P.t * SIM.pred * Q.t) list;
      (*7*)  terms : SFM.t ;
      (*8*)  uf : Uf.t ;
      (*?*)  nf_map : ST.t FLOATS.MPpos.t ;
      (*?*)  lin_class : SP.t MT.t ;
      (*!*)  float_of : (Term.t * Term.t * Term.t * Term.t) MT.t;
      (*!*)  deduced : SA.t;
    }

  (*= HELPERS ================================================================*)

  let debug_gappa = Options.debug_fpa()
  let loop = 100
  let (-@) l1 l2 = List.rev_append l1 l2

  let ale = Hstring.make "<="
  let alt = Hstring.make "<"
  let is_le n = Hstring.compare n ale = 0
  let is_lt n = Hstring.compare n alt = 0

  let alien_of p =
    match P.is_monomial p with
    | Some (a,x,b) when Q.equal a Q.one && Q.sign b = 0 -> x
    | _ -> P.embed p

  let poly_of r =
    match P.extract r with
    | Some p -> p
    | None -> P.create [Numbers.Q.one, r] Numbers.Q.zero (X.type_info r)

  let is_arith r =
    let ty = X.type_info r in
    ty == Ty.Tint || ty == Ty.Treal

  let is_real t = (T.view t).T.ty == Ty.Treal

  (* Gets the bound of a term in an environment *)
  (* TODO: should be in f_match ? *)
  let find_bnd t env =
    FLOATS.find_val_nf env.uf t env.info.FLOATS.bnds.FLOATS.i_map

  let poly_of_term uf t = poly_of (fst (UfWrap.find uf t (fun () -> assert false)))


  (*======================================================================*)

  (* Debuging primitives *)
  module Debug = struct

    let lits fmt la =
      List.iter
        (fun a ->
          fprintf fmt "We assume ";
          match a with
          | a, Some b, _, _ -> fprintf fmt "%a: %a@.@."
            LR.print (LR.make a) A.LT.print b
          | a, None, _, _ -> fprintf fmt "%a: None@.@."
            LR.print (LR.make a)
        )la

    let assume la =
      if debug_gappa then fprintf fmt "\n[gappa] \n%a" lits la

    let hyps fmt l =
      match SFM.elements l with
      | t::ts -> FMATCH.print_fpred fmt t ;
        List.iter (fun x -> fprintf fmt " /\\ "; FMATCH.print_fpred fmt x) ts
      | _ -> ()


    let lemma fmt l =
      fprintf fmt "%s : %a --> %a"
        l.FLOATS.name hyps l.FLOATS.hyps FMATCH.print_fpred l.FLOATS.concl

    let lem_names fmt ls = LEMMAS.iter (fun l -> fprintf fmt "@.   >> %a" lemma l) ls
    (* List.iter (fun l -> fprintf fmt "%s " l.name) ls *)

    let deps fmt deps =
      MFM.iter
        (fun t d ->
          fprintf fmt "%a is proved by:%a@.@."
            FMATCH.print_fpred
            t (* lem_names dep.implies *) lem_names d.proved_by
        ) deps

    let bnds fmt env =
      SFM.iter (fun t ->
        try
          let t = FMATCH.term_of_pred t in
          let i =
            FLOATS.find_val_nf env.uf t env.info.FLOATS.bnds.FLOATS.i_map in
          fprintf fmt "%a : %a@." T.print t I.print i
        with Not_found -> ()
      ) env.terms;
      MX.iter
	(fun p e ->
	  fprintf fmt "exponent(%a) : %a@."
	    X.print p I.print e)
        env.info.FLOATS.bnds.FLOATS.fix_map ;
      MX.iter
        (fun p e ->
          fprintf fmt "|mantissa(%a)| : %a@."
            X.print p I.print e)
        env.info.FLOATS.bnds.FLOATS.flt_map


    let abs fmt abs_tms = MT.iter (fun t _ ->
      fprintf fmt "abs(%a)@." T.print t)
      abs_tms

    let pred fmt p = match p with
        SIM.Eq -> fprintf fmt " = "
      | SIM.Ge -> fprintf fmt " >= "
      | SIM.Le -> fprintf fmt " <= "

    let constr fmt l =
      List.iter
        (fun (p,c,n) ->
          fprintf fmt "%a%a%s@." P.print p pred c
            (Q.to_string n)
        ) l

    let lemmas fmt ls = LEMMAS.iter (fprintf fmt "%a@." lemma) ls

    let polys fmt set = SP.iter (fun p -> fprintf fmt "%a, " P.print p) set

    let lin_class fmt map = MT.iter (fun t ps ->
      fprintf fmt "%a maps to the polys %a @." T.print t polys ps)
      map

    let nf_map fmt map =
      FLOATS.MPpos.iter
        (fun p ts ->
          fprintf fmt "%a maps to the terms %a @."
            P.print p T.print_list (ST.elements ts))
        map

    let rec fpreds fmt ts =
      SFM.iter (fun t -> fprintf fmt "%a " FMATCH.print_fpred t) ts;
      fprintf fmt "@."

    let env env =
      if debug_gappa then
	begin
          fprintf fmt "@.===== [ GAPPA ENV ] =========================";
          fprintf fmt "===============================================@.";
          fprintf fmt "-------------Gappa: dependencies--------@.";
          deps fmt env.deps;
          fprintf fmt "@.-------------Gappa: normal forms--------@.";
          nf_map fmt env.nf_map;
          fprintf fmt "@.-------------Gappa: terms---------------@.";
          fpreds fmt env.terms;
          fprintf fmt "@.-------------Gappa: linear classes------@.";
          lin_class fmt env.lin_class;
          fprintf fmt "@.-------------Gappa: maybe lemmas--------@.";
          lemmas fmt env.maybe_lems;
          fprintf fmt "@.-------------Gappa: lemmas--------------@.";
          lemmas fmt env.lems;
          fprintf fmt "@.-------------Gappa: linear constraints--@.";
          constr fmt env.constr;
          fprintf fmt "@.@.-------------Gappa: absolute values-----@.";
          abs fmt env.info.FLOATS.abs_tms;
          fprintf fmt "@.-------------Gappa: bounds--------------@.";
          bnds fmt env;
          fprintf fmt "===============================================";
          fprintf fmt "===============================================@.";
	end
  end

  (* --------------------------------------------------------------------- *)

  (*= REVIEWED ===============================================================*)

  (*Gets the syntactic terms of type Treal from a list of litterals *)
  let terms_of_assumed =
    let aux env (at, root, _,_) =
      match root, at with
      | Some at, (A.Builtin _ | A.Distinct _ | A.Pred _ | A.Eq _) ->
        A.LT.terms_nonrec at

      | None, (A.Builtin _ | A.Distinct _ | A.Pred _) ->
        ST.empty

      | None, A.Eq (a,b) ->
        (* Why is this case needed ? and not for the others ? *)
        (*TODO: why are both lists needed here? *)
        let l_x = match fst (X.term_extract a), fst (X.term_extract b) with
          | None, None     -> ST.empty
          | Some a, None   -> ST.singleton a
          | None, Some b   -> ST.singleton b
          | Some a, Some b -> ST.add a (ST.singleton b)
        in
        let l_p =
          let pa,_,_ = P.normal_form_pos (poly_of a) in
          let pb,_,_ = P.normal_form_pos (poly_of b) in
          let la = try MPpos.find pa env.nf_map with Not_found -> ST.empty in
          let lb = try MPpos.find pb env.nf_map with Not_found -> ST.empty in
          ST.union la lb
        in
        ST.union l_x l_p
    in
    fun env la ->
      List.fold_left
        (fun acc a ->
          ST.fold
            (fun t acc -> if is_real t then ST.add t acc else acc)
            (aux env a) acc
        ) ST.empty la

  module SC = Set.Make(
    struct
      type t = P.t * SIM.pred * Q.t
      let compare (a,b,c) (e,f,g) =
        let x = P.compare a e in
        if x <> 0 then x
        else
          let y = Pervasives.compare b f in if y <> 0 then y else Q.compare c g
    end)

  (* Checks constraints are well-formed and remove duplicates *)
  (* TODO: Do better: This will not remove trivially implied constrs ? *)
  let clean_constr constr =
    let constr = List.fold_left (fun s c -> SC.add c s) SC.empty constr in
    SC.elements constr

  (*takes a literal and turns it into a constraint for the
    simplex. Also returns the terms involved in the litteral, as
    they need to be linearly related*)
  let constr_of_atom = function
    | A.Builtin(s, n, [a;b]) when is_le n || is_lt n ->
      (*we treat all inequalities as LARGE inequalities *)
      let p = poly_of a in
      let q = poly_of b in
      let sb = if s then P.sub q p else P.sub p q in
      Some (sb,SIM.Ge,Q.zero)
    | A.Eq (a,b) when is_arith a -> (* [] *)
      Some (P.sub (poly_of a) (poly_of b), SIM.Eq, Q.zero)
    | _ -> None

  (* Take a set of terms and merge their linear classes. *)
  let merge_lin_class ts0 env =
    let ts = SFM.fold (fun t acc -> (FMATCH.term_of_pred t)::acc) ts0 [] in
    try
      let ts_class =
        List.fold_left
          (fun ps t ->
            if P.is_empty (poly_of_term env.uf t) then ps
            else SP.union ps (MT.find t env.lin_class)
          )SP.empty ts
      in
      List.fold_left
        (fun lin_class t -> MT.add t ts_class lin_class)
        env.lin_class ts
    with Not_found ->
      eprintf "didn't find %a@." FMATCH.print_fpred_list (SFM.elements ts0);
      assert false

  (* TODO : review it again *)
  let constrs_of_assumed env la =
    let env =
      List.fold_left (fun env (a,root,_,_) ->
        match constr_of_atom a with
        | None -> env
        | Some c ->
          let env = match root with
            | None -> env
            | Some root ->
              match A.LT.view root with
              | A.Builtin(_,_,[t;u]) | A.Eq (t,u) when is_real t ->
                let ts =
                  SFM.add (FMATCH.Fbnd t) (SFM.singleton (FMATCH.Fbnd u)) in
                {env with lin_class = merge_lin_class ts env}
              | _ -> env
          in
          {env with constr = c :: env.constr}
      ) env la
    in
    {env with constr = clean_constr env.constr}

  let make_deductions =
    let make_lit_l t l_bnd e =
      (*create the result that expresses l_bnd <= t *)
      assert ((T.view t).T.ty == Ty.Treal);
      let t_lit = A.LT.mk_builtin true ale [T.real (Q.to_string l_bnd); t] in
      LTerm t_lit, e, Other
    in
    let make_lit_u t u_bnd e =
      (*create the result that expresses t <= u_bnd *)
      assert ((T.view t).T.ty == Ty.Treal);
      let t_lit = A.LT.mk_builtin true ale [t; T.real (Q.to_string u_bnd)] in
      LTerm t_lit, e, Other
    in
    let make_deductions_term env t bnd =
    (* Given an interval of the form "t in [a,b]...[c,d]" create the results
       t <= d and t >= a *)
      if not (P.is_empty
                (poly_of (fst (UfWrap.find env.uf t (fun () -> assert false))))) then
        try
          let lower, e1,_ = I.borne_inf bnd in
	  let upper, e2,_ = I.borne_sup bnd in
          let ll = make_lit_l t lower e1 in
          let lu = make_lit_u t upper e2 in
	  [ll; lu]
        with I.No_finite_bound -> []
      else []
    in
    fun mod_tms env ->
      SFM.fold
        (fun t ls ->
          try
            let t = FMATCH.term_of_pred t in
            (make_deductions_term env t (find_bnd t env)) -@ ls
          with Not_found -> []
        ) mod_tms []

  (*= 6 functions related to add_terms (add_terms <-... build_graph <- assume =*)

  let add_term_set env t = {env with terms = SFM.add t env.terms}

  (* Aie. We add terms to uf copy. Why is this needed *)
  let add_uf env t =
    let t = FMATCH.term_of_pred t in
    if not (UfWrap.mem env.uf t) then
      (* This will allow to add the term globally to uf, cc, and relations *)
      new_created_terms := T.Set.add t !new_created_terms;
    {env with uf = UfWrap.add env.uf t}

  (* Adds the empty dependency of a term to a dependency map *)
  let add_empty_dep =
    let empty_dep = {implies = LEMMAS.empty ; proved_by = LEMMAS.empty} in
    fun env t ->
      if MFM.mem t env.maybe_deps then env
      else  {env with maybe_deps = MFM.add t empty_dep env.maybe_deps}

  (* Adds a new term in the map of linear classes
     a term is always in its own class *)
  let add_lin_class env t =
    let t = FMATCH.term_of_pred t in
    let lin_class = env.lin_class in
    if MT.mem t lin_class then env
    else
      let p,_,_ = P.normal_form_pos (poly_of_term env.uf t) in
      {env with lin_class = MT.add t (SP.singleton p) lin_class}

  (* computes the normal form of a term, and adds it to a map
     of -polynomials- in normal form to a list of terms and their
     normal form coefficients *)
  (* in other words, given a term t such that [[t]] = (p+c)*d
     we add t to the -list- of bindings of p *)
  let add_nf_map ({nf_map=map ; info=info} as env) t =
    let r = FMATCH.term_of_pred t in
    let p, c, d = P.normal_form_pos (poly_of_term env.uf r) in
    let st_p = try FLOATS.MPpos.find p map with Not_found -> ST.empty in
    let map = MPpos.add p (ST.add r st_p) map in
    let info =
      if P.is_empty p then
        FLOATS.update_const env.uf t (Q.mult c d) env.info
      else
        match IC.get_bound p env.fm_env with
        | None -> info
        | Some i ->
          (* Similarity considerations not needed, as fm_map and
             gappa map agree on common terms.*)
          let i_map = MPpos.add p i env.info.FLOATS.bnds.FLOATS.i_map in
          {info with FLOATS.bnds = {info.FLOATS.bnds with FLOATS.i_map = i_map}}
    in
    {env with nf_map = map ; info = info}

  (*adds terms to an environment*)
  let add_terms_env st env =
    (*TODO: why not checking if the term already exists once ?*)
    SFM.fold
      (fun t env ->
        let env = add_term_set env t in
        let env = add_uf env t in
        let env = add_empty_dep env t in
        let env = add_lin_class env t in
        let env = add_nf_map env t in
        env
      )st env

  (*= 1 function related to add_deps (add_deps <-... build_graph <- assume =*)

  (* Add the dependencies of a lemma to the dependencies indexed by a map *)
  let add_deps =
    let f_proved_by lem d = {d with proved_by = LEMMAS.add lem d.proved_by} in
    let f_implies   lem d = {d with implies = LEMMAS.add lem d.implies} in
    let aux func lem t map =
      let deps =
        try MFM.find t map
        with Not_found -> {proved_by=LEMMAS.empty; implies=LEMMAS.empty}
      in
      MFM.add t (func lem deps) map
    in
    fun lem deps_map ->
      let deps_map = aux f_proved_by lem lem.FLOATS.concl deps_map in
      SFM.fold (aux f_implies lem) lem.FLOATS.hyps deps_map


  (*= 1 aux functions on lemmas *)

  (* collects the set of hypotheses (pterms) of a set of lemmas *)
  let collect_hyps_of_lemmas lems =
    LEMMAS.fold
      (fun lem terms -> SFM.union lem.FLOATS.hyps terms) lems SFM.empty

  (* collects the set of conclusions (pterms) of a set of lemmas *)
  let collect_concl_of_lemmas lems =
    LEMMAS.fold
      (fun lem terms -> SFM.add lem.FLOATS.concl terms) lems SFM.empty


  (* aux function of apply_theorems and maybe_lemmas *)

  let info_from_lemmas lems env =
    LEMMAS.fold
      (fun l env ->
        match
          String.compare l.FLOATS.name "destruct_op_lin",
          String.compare l.FLOATS.name "abs_bnd" with
          | 0, 0 -> assert false
          | 0, _ -> (* why this particular case for "destruct_op_lin" ? *)
            let {FLOATS.hyps = hyps ; concl = concl} = l in
            {env with lin_class = merge_lin_class (SFM.add concl hyps) env}

          | _, 0 -> (* why this particular case for "abs_bnd" ? *)
            let {FLOATS.hyps = hyps ; concl = concl} = l in
            let x = SFM.choose hyps in
            let x = FMATCH.term_of_pred x  in
            let abs_x = FMATCH.term_of_pred concl in
            let {FLOATS.abs_tms = abs_tms} as info = env.info in
            {env with info = {info with FLOATS.abs_tms = MT.add x abs_x abs_tms}}

          | _ -> env
      ) lems env

  (* Applies all gappa thms to a set of terms, and returns the modified
     environment with added terms and new lemmas, and the set of
     terms which are hypotheses of the applied theorems.
     If the term is a constant, only adds the term to the environment.
  *)
  let apply_theorems =
    let is_const env t = P.is_empty (poly_of_term env.uf t) in
    let is_zero env t = FLOATS.is_zero env.uf t env.info in
    fun ts env ->
      let env = add_terms_env ts env in
      let ts =
        SFM.filter
          (fun t ->
            let t = FMATCH.term_of_pred t in
            not (is_const env t) && not (is_zero env t)
          )ts
      in
      let thm_lems = FLOATS.instances env.uf ts env.info in
      let maybe_lems = LEMMAS.union thm_lems env.maybe_lems in
      let maybe_deps = LEMMAS.fold add_deps thm_lems env.maybe_deps in
      let env = {env with maybe_lems=maybe_lems; maybe_deps=maybe_deps} in

      let thm_hyps = collect_hyps_of_lemmas thm_lems in
      let concl = collect_concl_of_lemmas thm_lems in
      let hyps = SFM.diff thm_hyps env.terms in
      let env = add_terms_env concl env in
      let env = add_terms_env hyps env in
      let env = info_from_lemmas thm_lems env in
      hyps, env

  let rec maybe_lemmas new_terms env =
    let hyps, env = apply_theorems new_terms env in
    if SFM.is_empty new_terms then env
    else maybe_lemmas hyps env

  (* aux function of build_graph *)

  let gather_terms_of_polys ps {nf_map=nf_map} =
    List.fold_left
      (fun acc p ->
        let z = try MPpos.find p nf_map with Not_found -> ST.empty in
        ST.fold (fun t acc -> SFM.add (FMATCH.Fbnd t) acc) z acc
      ) SFM.empty ps

  (* Takes a set of lemmas and dependencies and removes the
     dependencies on lemmas not in the set *)
  let clean_deps =
    (* takes a dependency map and filters the dependencies according to f *)
    let filter_dep func {proved_by=proved_by; implies=implies} =
      {proved_by = LEMMAS.filter func proved_by;
       implies = LEMMAS.filter func implies}
    in
    fun lems maybe_deps ->
      MFM.map (filter_dep (fun lem -> LEMMAS.mem lem lems)) maybe_deps


  (* Gives the minimal set of lemmas reachable form a given
     starting list of terms
     and dependencies for all considered terms *)
  let prune_graph leaves env =
    let rec aux terms all_terms reached_lems =
      match terms with
      | [] -> reached_lems
      | t::ts ->
	let {implies = implies} =
          try MFM.find t env.maybe_deps with Not_found -> assert false
        in
	let (new_terms, reached_lems) =
          LEMMAS.fold
            (fun lem ((trms, lems) as acc)->
              let {FLOATS.hyps=hyps; concl=concl} = lem in
              if not (LEMMAS.mem lem reached_lems) && SFM.subset hyps all_terms
              then concl :: trms, LEMMAS.add lem lems
	      else acc
            ) implies ([],reached_lems)
	in
        let terms = new_terms -@ ts in
        let all_terms =
          List.fold_left (fun ts t -> SFM.add t ts) all_terms new_terms
        in
        aux terms all_terms reached_lems
    in
    aux leaves (FMATCH.set_of_list leaves) LEMMAS.empty

  (* Takes bounds and returns the list of terms
     wich are bounded *)
  (* TODO: add also the terms which linearly depend on them*)
  (* TODO: This function should be in f_match ??? *)
  (* TODO: change its name ?*)
  let bnd_leaves bnd terms uf =
    SFM.fold
      (fun t ts ->
        let mem =
          match t with
          | FMATCH.Fbnd t ->
            let x, _ = UfWrap.find uf t (fun () -> assert false) in
            let p,_,_ = P.normal_form_pos (poly_of x) in
            MPpos.mem p bnd.FLOATS.i_map
          | FMATCH.Ffix t -> FLOATS.mem_val uf t bnd.FLOATS.fix_map
          | FMATCH.Fflt t -> FLOATS.mem_val uf t bnd.FLOATS.flt_map
        in
        if mem then SFM.add t ts else ts
      ) terms SFM.empty

  (*= ???????? ===============================================================*)






  (* checks to see if there is improvement between two bound maps for
     a given term *)
  let improvement uf t info1 info2 =
    let {FLOATS.bnds = b1},{FLOATS.bnds = b2} = info1, info2 in
    let x, _ = UfWrap.find uf (FMATCH.term_of_pred t) (fun () -> assert false) in
    let p, _, _ = P.normal_form_pos (poly_of x) in
    try
      begin
        match t with
          FMATCH.Fbnd t ->
            let i2, i1 = MPpos.find p b2.FLOATS.i_map, MPpos.find p b1.FLOATS.i_map in
            I.is_strict_smaller i2 i1
        | FMATCH.Ffix t | FMATCH.Fflt t -> failwith "Not implemented!"
      (* | Ffix t -> (find_val uf t b2.fix_map) >/ (find_val uf t b1.fix_map) *)
      (* | Fflt t -> (find_val uf t b2.flt_map) </ (find_val uf t b1.flt_map) *)
      end
    (* In this case the old bounds contained no info on t *)
    with Not_found ->
      match t with
        FMATCH.Fbnd t -> MPpos.mem p b2.FLOATS.i_map
      | FMATCH.Ffix t -> MX.mem x b2.FLOATS.fix_map
      | FMATCH.Fflt t -> MX.mem x b2.FLOATS.flt_map



  (* takes a polynomial p, an interval i, a map
     of term bounds, and an nf_map and:
     updates the bounds for the polynomial,
     returns the set of improved terms *)
  let update_p uf p i info nf_map =
    try
      let j = MPpos.find p info.FLOATS.bnds.FLOATS.i_map in
      let i = I.intersect i j in
      let bnds =
        {info.FLOATS.bnds
         with FLOATS.i_map =
            MPpos.add p i info.FLOATS.bnds.FLOATS.i_map} in
      let new_info = {info with FLOATS.bnds = bnds} in
      if I.is_strict_smaller i j then
        let ts =
          ST.fold (fun t acc -> SFM.add (FMATCH.Fbnd t) acc)
            (MPpos.find p nf_map) SFM.empty
        in
        new_info, ts
      else
        new_info, SFM.empty
    with Not_found -> info, SFM.empty
    | I.NotConsistent expl ->
      if debug_gappa then
        (fprintf fmt "[gappa] interval inconsistent %a@."
	   Ex.print expl);
      raise (Sig.Inconsistent (expl,[]))



  (* applies the lemmas and returns the set of modified terms and the new bounds *)
  let apply_lem_set uf lems info = LEMMAS.fold (
    fun l (tms, info) ->
      let {FLOATS.concl=concl} = l in

      (* applies a lemma to a map of bounds and refreshes the bounds *)
      (* fprintf fmt "applying lemma:%a\nwith bounds\n%a@." *)
      (*   Debug.lemma lem Debug.raw_bnds info.bnds; *)
      let new_info = FLOATS.run_info l.FLOATS.apply uf info in
      (* fprintf fmt "deduced:\n%a@." Debug.raw_bnds info.bnds; *)

      if improvement uf concl info new_info then
        let concl_class = FMATCH.p_class_of uf concl in
        (List.fold_left (fun ts t -> SFM.add t ts) tms concl_class), new_info
      else (tms, info)
  ) lems (SFM.empty,info)

  (* takes a set of terms and applies the lemmas
     that recursively depend on the terms *)
  (* returns the updated bounds and a list of terms with
     improved bounds *)
  (* We conjecture that this terminates thanks to the
     shape of lemmas. *)
  let apply_lems tms env =
    let rec apply_lems_aux tms info modif =
      if SFM.is_empty tms then
        let env = {env with info = info} in
        (env, modif)
      else
        let t = SFM.choose tms in
        let t_lems =
          try let {implies = t_lems} = MFM.find t env.deps in
              t_lems
          with Not_found ->  LEMMAS.empty
        in
        let (new_tms, new_info) = apply_lem_set env.uf t_lems info in
        let new_mod = SFM.union new_tms modif in
        let new_tms = SFM.remove t (SFM.union new_tms tms) in
        apply_lems_aux new_tms new_info new_mod
    in
    apply_lems_aux tms env.info SFM.empty

  (*-------------- Utilities for calling Incremental_simplex -----------------*)


  let add_const p n =
    let ty = P.type_info p in
    let p_n = P.create [] n ty in
    P.add p p_n

  let bnd_to_constr t env =
    let x, _ = UfWrap.find env.uf t (fun () -> assert false) in
    let p, _, _ = P.normal_form_pos (poly_of x) in
    if not (P.is_empty p) then begin
      let ty = P.type_info p in
      try
        let i = MPpos.find p env.info.FLOATS.bnds.FLOATS.i_map in
        let acc, ex =
          try
            let l,e_l,_ = I.borne_inf i in
            let p_l = add_const p (Q.minus l) in
            [(p_l,SIM.Ge, Q.zero)], e_l
          with I.No_finite_bound -> [], Ex.empty
        in
        let acc,ex =
          try
            let u,e_u,_ = I.borne_sup i in
            let u_p = P.sub (P.create [] u ty) p in
            (u_p,SIM.Ge, Q.zero)::acc, Ex.union e_u ex
          with I.No_finite_bound -> acc, ex
        in
        p , acc, ex
      with Not_found -> p, [], Ex.empty
    end
    else p, [], Ex.empty

  (* inline ? *)
  let nf_of_term uf t =
    P.normal_form_pos (poly_of_term uf t)


  (* the linear class of a set of modified terms *)
  let lin_mod env mod_tms =
    SFM.fold
      (* We do not include t in its own linear class, as it's best bound is
         already given by the lemma saturation.*)
      (fun t ps ->
        let t = FMATCH.term_of_pred t in
        let p,_,_ = nf_of_term env.uf t in
        let t_class = try MT.find t env.lin_class with Not_found -> SP.empty in
        let t_class = SP.remove p t_class in
        SP.union t_class ps)
      mod_tms SP.empty

  let bnd_constr t env lin_mod =
    let t = FMATCH.term_of_pred t in
    let p, cl, expl = bnd_to_constr t env in
    (*Excluding leaves causes a possible loss of completeness:
      forall x: real. 0. <= abs(x) + x
    *)
    if (* is_leaf t env.deps && *) not (P.is_empty p) && SP.mem p lin_mod then
      ([p], cl,expl)
    else
      ([], cl, expl)


  (* given terms t and u such that u = abs(t), create the constraint
     t = abs(t) *)
  let abs_bnd_pos uf t abs_t =
    let p, abs_p = poly_of_term uf t, poly_of_term uf abs_t in
    let p_abs_p = P.sub p abs_p in
    (p_abs_p, SIM.Eq, Q.zero)

  (* given terms t and u such that u = abs(t), create the constraint
     t = -abs(t) *)
  let abs_bnd_neg uf t abs_t =
    let p, abs_p = poly_of_term uf t, poly_of_term uf abs_t in
    let p_abs_p = P.add p abs_p in
    (p_abs_p, SIM.Eq, Q.zero)


  (* Gets positive or negative terms that appear as absolute values. *)
  let get_abs_pos_neg env =
    MT.fold
      (fun t _ ts ->
        try let i = find_bnd t env in
            match I.is_positive i, I.is_positive (I.scale Q.m_one i) with
            | Sig.Yes _, _ -> ST.add t ts
            | _, Sig.Yes _ -> ST.add t ts
            | _ -> ts
        with Not_found -> ts) env.info.FLOATS.abs_tms ST.empty

  (* If t >= 0 and abs(t) appears, create the constraint t = abs(t) *)
  (* If t <= 0 and abs(t) appears, create the constraint t = -abs(t) *)
  let abs_constr t env =
    let t = FMATCH.term_of_pred t in
    try
      let abs_t = MT.find t env.info.FLOATS.abs_tms in
      let i = find_bnd t env in
      match I.is_positive i, I.is_positive (I.scale Q.m_one i) with
      | Sig.Yes (e,_), _       -> [abs_bnd_pos env.uf t abs_t], e
      | Sig.No, Sig.Yes (e,_)  -> [abs_bnd_neg env.uf t abs_t], e
      | Sig.No, Sig.No     -> [], Ex.empty
    with Not_found -> [],Ex.empty

  (* Returns the triple of polynomials, simplex constraints and explanations
     to give as input to the simplex algorithm *)
  let make_simplex env mod_tms =
    let lin_mod =  lin_mod env mod_tms in
    let constr =
      SFM.fold (fun t (ps,cs,expls) ->
        let pbnd, cbnd, ebnd = bnd_constr t env lin_mod in
        let cabs, eabs = abs_constr t env in
        let ps = List.fold_right SP.add pbnd ps in
        let expl = Ex.union eabs ebnd in
        (ps, cabs -@ cbnd -@ cs, Ex.union expl expls)
      ) env.terms (SP.empty,[], Ex.empty)
    in
    constr

  (* Calls the simplex algorithm to get the upper bound of p
     on the constraints constr. Updates the bounds of
     p in info using nf_map *)
  let tighten_term_u uf constr p info sim_state expl nf_map =
    TG.SimplexTime.start();
    incr TG.nsmplx;
    let res = SIM.maximize p sim_state in
    TG.SimplexTime.get();
    match res with
    | SIM.Max (n, _) ->
      let ty = P.type_info p in
      let undef = I.undefined ty in
      let ii = I.new_borne_sup expl n true undef in
      update_p uf p ii info nf_map
    | SIM.Unsat (*_*) ->
      TG.SaturateTime.get();
      TG.HammerTime.get();
      raise (Sig.Inconsistent (expl,[]))
    | SIM.Unbound (*_*) -> (info,SFM.empty)

  let tighten_term_l uf constr p info sim_state expl nf_map =
    let m_p = P.mult_const Q.m_one p in
    TG.SimplexTime.start();
    let res = SIM.maximize m_p sim_state in
    TG.SimplexTime.get();
    match res with
    | SIM.Max (n, _) ->
      let ty = P.type_info p in
      let undef = I.undefined ty in
      let ii = I.new_borne_inf expl (Q.minus n) true undef in
      update_p uf p ii info nf_map
    | SIM.Unsat ->
      TG.SaturateTime.get();
      TG.HammerTime.get();
      raise (Sig.Inconsistent (expl,[]))
    | SIM.Unbound  -> (info,SFM.empty)

  let tighten_by_f tighten polys info tms =
    SP.fold
      (fun p (info,ts) ->
        let info, tus = tighten p info in
        info, (SFM.union tus ts)
      ) polys (info,tms)

  (* Take the mutable environment for the incremental simplex,
     the list of linear constraints
     and polynomials and bounds, and apply the simplex algorithm
     to each polynomial, then update the bounds accordingly. *)
  let tighten_simplex env sim_state polys constr expl =
    let {info = info; uf = uf; nf_map = nf_map} = env in
    if SP.is_empty polys then info, SFM.empty
    else begin
      let constr = clean_constr constr in

      SIM.add_constraints constr sim_state;

      let info, impr_u =
        tighten_by_f (fun p info ->
          tighten_term_u uf constr p info sim_state expl nf_map)
          polys info SFM.empty
      in
      tighten_by_f (fun p info ->
        tighten_term_l uf constr p info sim_state expl nf_map)
        polys info impr_u
    end

  let simplex env mod_tms sim_state =
    let polys, simplex_bnds, expl = make_simplex env mod_tms in
    let info, mod_tms =
      tighten_simplex env sim_state polys simplex_bnds expl in
    {env with info = info}, mod_tms

  (* ------------------------------------------------------------------ *)

  (* Takes bounds on terms and bounds on normal forms, a list of
     terms, a list of modified normal forms and a map from normal
     forms to triples and updates the corresponding terms.*)
  let info_from_fm polys {fm_env=fm_env;info=info;uf=uf;nf_map=nf_map} =
    List.fold_left
      (fun new_info p ->
        match IC.get_bound p fm_env with
        | None -> new_info
        | Some i -> fst (update_p uf p i new_info nf_map)
      ) info polys



  (* Given an environment and a list of terms,
     do the following actions:
     - call all lemmas on the terms
     - get the modified bounds
     - call the simplex to tighten the bounds of terms
     - repeat on the improved bounds until there is no
     improvement
  *)
  let saturate env terms =
    TG.SaturateTime.start();
    let sim_state = SIM.empty () in
    (* contraintes issues du probleme de depart *)
    SIM.add_constraints env.constr sim_state;
    let rec sat loop env terms imprvmt =
      incr TG.nsatur;
      let env, gappa_mod = apply_lems terms env in
      (* it suffices to check if there was an improvement
         in the gappa stage of saturation *)
      if SFM.is_empty gappa_mod || loop <= 0 then
        (env, imprvmt)
      else

        (*WARNING: this is terrible for performance. *)
        (* let gappa_mod = env.terms in *)

        let env, simplx_mod = simplex env gappa_mod sim_state in

        (* We check if some new terms with absolute values become
           positive or negative and call the simplex with the new
           linear constraints if necessary *)
        let abs_tms = ST.fold (fun t ts -> SFM.add (FMATCH.Fbnd t) ts)
          (get_abs_pos_neg env) SFM.empty in
        let abs_mod = SFM.inter abs_tms simplx_mod in
        let env, smplx_mod = simplex env abs_mod sim_state in

        let mod_terms = SFM.union simplx_mod gappa_mod in

        (* let mod_terms = simplx_mod in *)
        sat (loop - 1) env mod_terms (SFM.union mod_terms imprvmt)

    in
    try
      let res = sat loop env terms SFM.empty in
      TG.SaturateTime.get();
      res
    with Sig.Inconsistent (expl,[]) ->
      TG.SaturateTime.get();
      TG.HammerTime.get();
      raise (Sig.Inconsistent (expl,[]))

  let build_graph tms env =
    TG.GraphTime.start();
    let env = maybe_lemmas tms env in
    let mod_terms = IC.new_improved env.fm_env in
    let info = info_from_fm mod_terms env in
    let bnd_terms = bnd_leaves info.FLOATS.bnds env.terms env.uf in
    let lemmas = prune_graph (SFM.elements bnd_terms) env in
    let deps = clean_deps lemmas env.maybe_deps in
    let new_lems = LEMMAS.diff lemmas env.lems in
    let lem_terms = collect_hyps_of_lemmas new_lems in
    let mod_terms = gather_terms_of_polys mod_terms env in
    TG.GraphTime.get();
    {	env with lems = lemmas; deps = deps ; info = info},
      (SFM.union lem_terms mod_terms)


  type mantissa_size =
    Unk
  | Exact
  | Zero

  let properties_of env t =
    List.fold_left
      (fun acc t ->
        if acc == Zero then acc
        else
          match T.view t with
          | {T.f = Sy.Op Sy.Float; xs=[u1;v1;m1;a]; ty=ty1} -> Exact
          | {T.f = Sy.Real h} ->
            let q = Q.from_string (Hstring.view h) in
            if Q.is_a_power_of_2 q then Zero
            else acc
          | _ -> acc
      )Unk (UfWrap.class_of env.uf t)

  let deduce_float_of_float t (mant, prec, md, a) (ded, env) =
    let ded, deduced =
      List.fold_left
        (fun ((ded, deduced) as acc) ca ->
          match Term.view ca with
          | {T.f = Sy.Op Sy.Float; xs=[u;v;m;b]; ty=ty} ->
            if mant != u || prec != v then acc (* do better *)
            else
              let eq = Literal.LT.mk_eq t ca in
              if SA.mem eq deduced then acc
              else
                let () =
                  if false then
                    fprintf fmt "R(R(x)) = R(x) :: %a = %a@."
                      Term.print t Term.print ca
                in
                (LTerm eq, Ex.empty, Other) :: ded,
                SA.add eq deduced

          | {T.f = Sy.Op Sy.Mult; xs=[x;y]; ty=ty} ->
            begin match properties_of env x, properties_of env y with
            | Zero, Exact | Exact, Zero ->
              let eq = Literal.LT.mk_eq t ca in
              if SA.mem eq deduced then acc
              else
                let () =
                  if false then
                    fprintf fmt "add term %a = %a@."
                      Term.print t Term.print a
                in
                (LTerm eq, Ex.empty, Other) :: ded,
                SA.add eq deduced
            | _ -> acc
            end

          | _ ->
            match properties_of env ca with
            | Zero ->
              let eq = Literal.LT.mk_eq t ca in
              if SA.mem eq deduced then acc
              else
                let () =
                  if false then
                    fprintf fmt "add term %a = %a@."
                      Term.print t Term.print a
                in
                (LTerm eq, Ex.empty, Other) :: ded,
                SA.add eq deduced
            | _ -> acc
            )(ded, env.deduced) (UfWrap.class_of env.uf a)
      in
    ded, {env with deduced = deduced}

  let mk_direct_deductions env =
    MT.fold
      (fun t args acc ->
        let acc = deduce_float_of_float t args acc in
        (*deduce_float_of_exact t args acc*)
        acc
      )env.float_of ([], env)

  let add_one_float_of t fl_of =
    match Term.view t with
    | {T.f = Sy.Op Sy.Float; xs=[u1;v1;m1;a]; ty=ty} ->
      assert (ty == Ty.Treal);
      MT.add t (u1,v1,m1,a) fl_of
    | _ -> fl_of

  let add_float_of ({float_of=float_of} as env) (a, _, _, _) =
    let lx = match a with
      | A.Eq(r1, r2) -> [r1;r2]
      | A.Builtin(_,_,l) -> l
      | A.Distinct(_,l) -> l
      | A.Pred(r,_) -> [r]
    in
    let lx = List.fold_left (fun acc r -> (X.leaves r) -@ acc)[] lx in
    let sx =
      List.fold_left
        (fun acc x ->
          match X.term_extract x with
          | Some t, true ->
            if ST.mem t acc || MT.mem t float_of then acc
            else T.subterms acc t
          | _ -> acc
        )ST.empty lx
    in
    let fl_of =
      ST.fold
        (fun t fl_of ->
          if MT.mem t fl_of then fl_of else add_one_float_of t fl_of
        )sx float_of
    in
    {env with float_of = fl_of}


  (*
    Assume gets the terms form the litterals that
    are not already in the set of terms, and performs the
    following operations:
    - Call FM on the litterals, grab the bounds and the modified terms
    - Add all lemmas and relevant information
    pertaining to the new terms to the environment.
    - Prune the graph using known bounds, and clean the dependencies
    - Call saturate and return result
  *)
  let assume env uf la =
    new_created_terms := Term.Set.empty;
    Debug.assume la;
    try
      incr TG.nass;
      TG.HammerTime.start();
      let env = {env with uf = uf} in
      let env = List.fold_left add_float_of env la in
      let ded, env = mk_direct_deductions env in
      let terms = terms_of_assumed env la in (*1. get new terms *)
      let new_terms = Lists.rrmap (fun t -> FMATCH.Fbnd t) (ST.elements terms) in
      TG.HammerTime.get();

      let fm_env, r1, ic_new_terms = IC.assume env.fm_env uf la in (* call IC *)
      let r1 = {r1 with assume = ded -@ r1.assume} in
      TG.HammerTime.start();
      let env = {env with fm_env = fm_env} in

      let env, changes = build_graph (FMATCH.set_of_list new_terms) env in
      let env = constrs_of_assumed env la in
      if SFM.is_empty changes then
        begin
          TG.HammerTime.get();
          Debug.env env;
          env, r1, ST.union ic_new_terms !new_created_terms
        end
      else
        begin
          let start = changes in
          let env, imprvmt = saturate env start in
          if SFM.is_empty imprvmt then
            begin
              TG.HammerTime.get();
              TG.nlems := max !TG.nlems (LEMMAS.cardinal env.maybe_lems);
              TG.nterms := max !TG.nterms (SFM.cardinal env.terms);
              Debug.env env;
              env, r1, ST.union ic_new_terms !new_created_terms
            end
          else
            begin
              let result = make_deductions imprvmt env in
              TG.HammerTime.get();
              TG.nlems := max !TG.nlems (LEMMAS.cardinal env.maybe_lems);
              TG.nterms := max !TG.nterms (SFM.cardinal env.terms);
              Debug.env env;
              env, {assume = result -@ r1.assume;
                    remove = r1.remove },
              ST.union ic_new_terms !new_created_terms
            end
        end
    with I.NotConsistent expl ->
      if debug_gappa then
        fprintf fmt "[gappa] interval inconsistent %a@." Ex.print expl;
      raise (Sig.Inconsistent (expl,[]))

  let assume env uf la =
    let env, res, new_created_terms = assume env uf la in
    (*MX.iter (fun x _ -> UfWrap.assert_leaves_are_inited uf x) env.monomes;
    MP.iter (fun p _ -> UfWrap.assert_leaves_are_inited uf (alien_of p)) polys;*)
    env, res, new_created_terms

  (*= SIMPLE FUNCTIONS =======================================================*)

  let query env uf a_ex =
    try ignore(assume env uf [a_ex]); No
    with Sig.Inconsistent (expl,_) -> Yes (expl, [])

  let make_r_lit_l t l_bnd e =
    let ty = (T.view t).T.ty in
    let t_lit = A.LT.mk_builtin true ale [T.real (Q.to_string l_bnd); t] in
    let lit =
      let x,_,_ = X.make t in
      LR.view (LR.mk_builtin true ale [alien_of (P.create [] l_bnd ty); x])
    in
    (lit,Some t_lit,e)

  (* TODO: case split on the following cases:
     -disjoint unions of intervals
     -interval on x: x is either normal float,
     denormal float or underflow*)
  (* Find a suitable variable x on which to split on its sign.
     In this case: abs(x) appears in the problem. Prefer variables
     x such that x is in the same linear-class as abs(x) *)
  let case_split_abs env =
    let {FLOATS.abs_tms = abs_tms} = env.info in
    (* Create the list of pairs (x,b) such that:
       - abs(x) appears in the problem
       - x is not strictly positive or negative
       - b = true if x and abs(x) are in the same linear class,
       false otherwise *)
    let split_list =
      MT.fold (fun t abs_t l ->
        try let i = find_bnd t env in
            match I.is_positive i, I.is_positive (I.scale Q.m_one i) with
            | Sig.No, Sig.No ->
              let p,_,_ = nf_of_term env.uf abs_t in
              let t_lin = MT.find t env.lin_class in
              if SP.mem p t_lin then
                (t,true)::l
              else (t,false)::l
            (* No need to case-split *)
            | _ -> l
        (* t is not in the environment: we don't case-split *)
        with Not_found -> l) abs_tms []
    in
    let l1, l2 = List.partition (fun (_,b) -> b) split_list in
    match l1, l2 with
    | [],[] -> []
    | (x,_)::_, _ | [], (x,_)::_ ->
      (* printf "splitting on %a@." T.print x; *)
      let lit,_,_ = make_r_lit_l x Q.zero Ex.empty in
      [lit, CS (Th_arith, Q.from_int 2)]

  let case_split env uf =
    match case_split_abs env with
    | [] -> (* In this case there is no variable on which to split *)
      IC.case_split env.fm_env uf
    | l -> l

  let add env uf t rt =
    let fm_env, ded = IC.add env.fm_env uf t rt in
    {env with fm_env = fm_env}, ded

  let print_model _ _ _ = assert false

  let new_terms env = Term.Set.empty

  let empty_bnds =
    {FLOATS.i_map   = MPpos.empty;
     fix_map = MX.empty;
     flt_map = MX.empty}

  let empty_info = {FLOATS.bnds = empty_bnds ; abs_tms = MT.empty}

  (*The empty gappa environment *)
  let empty _ =
    {lems = LEMMAS.empty;
     deps = MFM.empty ;
     info = empty_info ;
     fm_env = IC.empty [];
     maybe_lems = LEMMAS.empty ;
     maybe_deps = MFM.empty ;
     nf_map = MPpos.empty ;
     terms = SFM.empty ;
     constr = [] ;
     lin_class = MT.empty;
     uf = UfWrap.empty();
     float_of = MT.empty;
     deduced = SA.empty;
    }

end
