module Q = Numbers.Q
module Z = Numbers.Z

open Format
open Options
(* open Num *)
(* open Gmp *)
module H = Hashtbl


(* let one   = Q.from_int 1 *)
(* let minus_one = Q.from_int (-1) *)


module type SIMPLEX =
  sig
    type r
    type p
    type state
    type pred = Eq | Ge | Le
    type result = Unsat | Unbound | Max of Q.t * (Q.t * r) list

    val empty : unit -> state
    val add_constraints : (p * pred * Q.t) list -> state -> unit
    val maximize : p -> state -> result
  end


module Make (X : Sig.X) (P : Polynome.T with type r = X.r)
  : SIMPLEX with type r = X.r and type p = P.t
= struct

  module SX = Set.Make(struct type t = X.r let compare = X.str_cmp end)
  module MX = Map.Make(struct type t = X.r let compare = X.str_cmp end)

  module SMX = Set.Make(struct
    type t = Q.t MX.t * Q.t
    let compare (m1,c1) (m2,c2) =
      let c = Q.compare c1 c2 in
      if c <> 0 then c
      else MX.compare Q.compare m1 m2
  end)

  let debug = false

  type r = X.r
  type p = P.t
  type pred = Eq | Ge | Le
  type result = Unsat | Unbound | Max of Q.t * (Q.t * r) list

  type state =
      { mutable ictx : SMX.t;
        mutable vdiff  : (X.r * X.r) MX.t;
        mutable base   : X.r array;
        mutable hbase  : X.r array;
        mutable blast  : int;
        mutable hlast  : int;
        mutable matrix : Q.t array array }

  type sbt  = {old_lhs : int; lhs : int; rhs : Q.t array}

  exception E_unsat
  exception E_unbound
  exception E_max of Q.t * (Q.t * r) list

  let ghost_term = Term.fresh_name Ty.Treal
  let const_term = Term.fresh_name Ty.Treal
  let ghost_var () = X.term_embed ghost_term
  let const_var () = X.term_embed const_term
  let slack_var () = X.term_embed (Term.fresh_name Ty.Treal)
  let dummy () = X.term_embed Term.dummy

  (*** module de debbugage **************************************************)
  module D = struct

    let expand s n =
      let rec exrec s n = if n <= -1 then s else exrec (" "^s) (n-1) in
      exrec s (n-(String.length s))

    let print_pred = function Eq -> "=" | Ge -> ">=" | Le -> "<="

    let max_array a =
      Array.fold_left (fun n v -> max n (String.length (Q.to_string v))) 0 a

    let max_matrix m =
      Array.fold_left (fun n a -> max n (max_array a)) 0 m

    let parray hlast sp fmt a =
      for i = 0 to hlast do
        fprintf fmt "%s" (expand (Q.to_string a.(i)) sp)
      done

    let add_one_constraint l c =
      if debug then begin
        fprintf fmt "add one constraint : ";
        List.iter
        (fun (c,x) -> fprintf fmt "%s*%a + "
          (Q.to_string c) X.print x) l;
        fprintf fmt "%s@." (Q.to_string c)
      end

    let pline hlast fmt (i,a) =
      if debug then begin
        let sp = max_array a in
        fprintf fmt "x%d   = %a" i (parray hlast sp) a
      end

    let psbt hlast fmt sbt =
      if debug then begin
        let sp = max_array sbt.rhs in
        fprintf fmt "%d |->  %a / (old %d)@."
          sbt.lhs (parray hlast sp) sbt.rhs sbt.old_lhs
      end

    let print_state s =
      if debug then begin
        fprintf fmt "------ STATE ---------------------------------@.";
        let sp = max_matrix s.matrix in
        fprintf fmt "@.basic vars: ";
        for i = 0 to s.blast do
          fprintf fmt " %d -> %a | " i X.print (s.base.(i));
        done;
        fprintf fmt "@.@.non basic vars: ";
        for i = 0 to s.hlast do
          fprintf fmt " %d -> %a | " i X.print (s.hbase.(i));
        done;
        fprintf fmt "@.@.vdiff:@.";
        MX.iter
          (fun x (yp, yn) ->
            fprintf fmt "%a -> %a - %a@." X.print x X.print yp X.print yn
          )s.vdiff;
        fprintf fmt "@.@.matrix:@.";
        for i = 0 to s.blast do
          fprintf fmt "%d  = %a@." i (parray s.hlast sp) s.matrix.(i)
        done;
        fprintf fmt "---------------------------------------------@.@."
      end

    let print_given_pb cost s =
      if debug then begin
        let sp = max (max_array cost) (max_matrix s.matrix) in
        print_state s;
        fprintf fmt "cost:%a@." (parray s.hlast sp) cost;
        fprintf fmt "---------------------------------------------@.@."
      end

    let print_init_result e gcost state =
      if debug then begin
        fprintf fmt "INIT's result: (%s)@."
          (match e with
            | E_unbound -> "Unbound"
            | E_unsat   -> "Unsat"
            | E_max _   -> "Max"
            | _ -> assert false
          );
        print_given_pb gcost state
      end


  end
  (*** / module de debbugage ***********************************************)

  let empty () =
    if debug then fprintf fmt "EMPTY:@.";
    if debug then fprintf fmt "ghost_var = %a@." X.print (ghost_var());
    let base  = Array.make 20 (dummy()) in
    let hbase = Array.make 20 (dummy()) in
    hbase.(0) <- const_var();
    hbase.(1) <- ghost_var();
    { base   = base;
      hbase  = hbase;
      blast  = -1;
      hlast  = 1;
      vdiff  = MX.empty;
      matrix = Array.make 20 [||];
      ictx   = SMX.empty }


  (* let to_gmp_list p = *)
  (*   let l, c = P.to_list p in *)
  (*   List.rev_map (fun (c,x) -> c, x) l, c *)

  let normalize_direction acc (p,pred,cst) =
    let l, c = P.to_list p in
    let c = Q.sub c cst in
    let minus_list l = List.rev_map (fun (c,x) -> Q.minus c, x) l in
    match pred with
      | Ge -> (l,c) :: acc
      | Le -> (minus_list l, Q.minus c) :: acc
      | Eq -> (l,c) :: (minus_list l, Q.minus c) :: acc

  let apply_vdiff l vdiff =
    List.fold_left
      (fun (mp, new_vars, vdiff) (c,x) ->
        try
          let y_p, y_n = MX.find x vdiff in
          let mp = MX.add y_p c mp in
          let mp = MX.add y_n (Q.minus c) mp in
          mp, new_vars, vdiff
        with Not_found ->
          let y_p = slack_var () in
          let y_n = slack_var () in
          let vdiff = MX.add x (y_p,y_n) vdiff in
          let mp = MX.add y_p c mp in
          let mp = MX.add y_n (Q.minus c) mp in
          let new_vars = y_p :: y_n :: new_vars in
          mp, new_vars, vdiff
      )(MX.empty, [], vdiff) l

  let subst_by hlast a v rhs =
    if not (Q.equal v Q.zero) then
      for j = 0 to hlast do
        a.(j) <- Q.add (a.(j)) (Q.mult v rhs.(j))
      done

  let normalized_array m c s =
    let {hbase=hbase;base=base;hlast=hlast;blast=blast;matrix=matrix} = s in
    let a = Array.make (Array.length hbase) Q.zero in
    a.(0) <- c;
    for i = 1 to hlast do
      assert (Q.equal Q.zero a.(i));
      try a.(i) <- MX.find (hbase.(i)) m
      with
        | Not_found -> ()
        | Invalid_argument("index out of bounds") -> assert false
    done;
    for i = 0 to blast do
      try
        let v = MX.find (base.(i)) m in
        try subst_by s.hlast a v (matrix.(i))
        with Not_found -> assert false
      with Not_found -> ()
    done;
    a

  let grow_to new_sz alast dummy a =
    let b = Array.make new_sz dummy in
    for i = 0 to alast do b.(i) <- a.(i) done;
    for i = alast + 1 to Array.length a - 1 do assert (a.(i) == dummy) done;
    b

  let add_non_basic_vars new_vars state =
    let {matrix=matrix; hlast=hlast; blast=blast; hbase=hbase} = state in
    match new_vars with
      | [] -> ()
      | _  ->
        let hsize = Array.length hbase in
        let nb_new_vars = List.length new_vars in
        let new_hlast = hlast + nb_new_vars in
        let hbase =
          if new_hlast < hsize then hbase
          else
            let new_hsize = max (2 * hsize) (new_hlast + 1) in
            let hbase = grow_to new_hsize hlast (dummy()) hbase in
            for i = 0 to blast do
              matrix.(i) <- grow_to new_hsize hlast Q.zero matrix.(i)
            done;
            hbase
        in
        let cpt = ref hlast in
        List.iter (fun v -> incr cpt; hbase.(!cpt) <- v) new_vars;
        assert (!cpt=new_hlast);
        state.hbase <- hbase;
        state.hlast <- new_hlast

  let add_basic_var a s =
    let {matrix=matrix;blast=blast;base=base} = s in
    let slk  = slack_var () in
    let new_blast = blast + 1 in
    let bsize = Array.length base in
    let matrix,base =
      if new_blast < bsize then matrix, base
      else
        let new_bsize = max (2 * bsize) (new_blast + 1) in
        let base   = grow_to new_bsize blast (dummy()) base in
        let matrix = grow_to new_bsize blast [||] matrix in
        matrix, base
    in
    base.(new_blast) <- slk;
    matrix.(new_blast) <- a;
    s.matrix <- matrix;
    s.base <- base;
    s.blast <- new_blast


  let add_one_constraint state (l, c) =
    D.print_state state;
    D.add_one_constraint l c;
    let m, new_vars, vdiff = apply_vdiff l state.vdiff in
    if not (SMX.mem (m,c) state.ictx) then begin
      state.vdiff <- vdiff;
      state.ictx  <- SMX.add (m,c) state.ictx;
      add_non_basic_vars new_vars state;
      let a = normalized_array m c state in
      add_basic_var a state;
      D.print_state state;
    end

  let add_constraints ctx state =
    let ctx = List.fold_left normalize_direction [] ctx in
    List.iter (add_one_constraint state) ctx


  (**********************************************************************)

  (*** structures de donnees et fonctions utiles ***************************)

  exception Out of int * bool

  let index_of_ghost state =
    try
      if debug then begin
        fprintf fmt "--------------------@.ghost=%a@." X.print (ghost_var ());
        fprintf fmt "@.basic vars: ";
        for i = 0 to state.blast do
          fprintf fmt " %d -> %a | " i X.print (state.base.(i));
        done;
        fprintf fmt "@.@.non basic vars: ";
        for i = 0 to state.hlast do
          fprintf fmt " %d -> %a | " i X.print (state.hbase.(i));
        done;
        fprintf fmt "@.@.";
      end;
      let ghost = ghost_var () in
      let hbase = state.hbase in
      for i = 1 to state.hlast do
        if X.equal hbase.(i) ghost then raise (Out(i, true))
      done;
      let base = state.base in
      for i = 0 to state.blast do
        if X.equal base.(i) ghost then raise (Out(i, false))
      done;
      assert false
    with Out (i, b) -> i, b

  let mk_max state co =
    (* XXX : TODO, faire mieux pour la liste *)
    E_max (co.(0), [])

  let line_with_min_const state =
    let mx = state.matrix in
    let res = ref 0 in
    for i = 1 to state.blast do
      if Q.compare mx.(i).(0) mx.(!res).(0) < 0 then res := i
    done;
    !res

  let mult_const p c =
    let f i cs = p.(i) <- Q.zero in
    let g i cs = () in
    let h i cs = p.(i) <- Q.mult c cs in
    let k =
      if Q.equal c Q.zero then f
      else if Q.equal c Q.one then g else h in
    Array.iteri k p

  let subst hlast a lhs rhs =
    let v = a.(lhs) in
    if not (Q.equal v Q.zero) then begin
      a.(lhs) <- Q.zero; subst_by hlast a v rhs
    end

  let subst_line hlast {old_lhs=old_lhs; lhs=lhs; rhs=rhs} i a =
    if i <> old_lhs then subst hlast a lhs rhs

  let subst_matrix state sbt =
    let mx = state.matrix in
    let hlast = state.hlast in
    for i = 0 to state.blast do subst_line hlast sbt i mx.(i) done


  (*** / structures de donnees et fonctions utiles **************************)



  (*** coeur du simplex ****************************************************)

  exception Choose_index of int


  let choose_var state co order =
    try
      for j = 1 to state.hlast do
        let i = Queue.pop order in
        Queue.push i order;
        if Q.compare co.(i) Q.zero > 0 then raise (Choose_index i)
      done;
      raise (mk_max state co)
        (*failwith "raise (E_max (ctx,co))"*)
    with Choose_index ind -> ind

  let choose_eq state co ch_vr =
    let acc = ref None in
    let mx = state.matrix in
    for j = 0 to state.blast do
      let a = mx.(j) in
      let v_ch_vr = a.(ch_vr) in
      if Q.compare v_ch_vr Q.zero < 0 then
        let rap = Q.minus (Q.div a.(0) v_ch_vr) in
        match !acc with
          | None -> acc := Some (v_ch_vr, rap, j)
          | Some (v_r, r, jj) ->
            let delta = Q.compare rap r in
            let change = delta < 0 || (delta = 0 &&  j < jj) in
            if change then acc := Some (v_ch_vr, rap, j)
    done;
    match !acc with
      | None            -> raise E_unbound
      | Some (_, _, j) -> j, mx.(j)


  let change_pivot ch_vr (old_vr, old_p) state order =
    if debug then fprintf fmt "ch_vr: %d et old_vr: %d@." ch_vr old_vr;
    if debug then fprintf fmt "swap %a and %a@."
      X.print state.hbase.(ch_vr) X.print state.base.(old_vr);
    let tmp = state.hbase.(ch_vr) in
    state.hbase.(ch_vr) <- state.base.(old_vr);
    state.base.(old_vr) <- tmp;
    D.print_state state;
    let v = old_p.(ch_vr) in
    old_p.(ch_vr) <- Q.m_one;
    mult_const old_p (Q.div Q.m_one v);
    {old_lhs=old_vr; lhs=ch_vr; rhs=old_p}


  let rec simplex state co co_opt order =
    if debug then fprintf fmt "IN rec simplex@.";
    D.print_given_pb co state;
    let ch_vr = choose_var state co order in
    if debug then fprintf fmt "chosen var's index: %d@." ch_vr;
    let ch_eq = choose_eq state co ch_vr in
    if debug then fprintf fmt "chosen eq:     %a@."(D.pline state.hlast) ch_eq;
    let sbt = change_pivot ch_vr ch_eq state order in
    if debug then fprintf fmt "pivot's result: %a@." (D.psbt state.hlast) sbt;
    begin match co_opt with
        None -> () | Some co_bis -> subst state.hlast co_bis sbt.lhs sbt.rhs;
    end;
    subst state.hlast co sbt.lhs sbt.rhs;
    subst_matrix state sbt;
    simplex state co co_opt order


  (*** / coeur du simplex ***************************************************)

  exception Fin of int

  let deleta_ghost ighost aghost state order =
    D.print_state state;
    try
      for i = 1 to state.hlast do
        if Q.compare aghost.(i) Q.zero <> 0 then raise (Fin i)
      done;
      failwith "Pas possible"
    with Fin ch_vr ->
      let sbt = change_pivot ch_vr (ighost, aghost) state order in
      if debug then
        fprintf fmt "pivot's result (Ghost): %a@." (D.psbt state.hlast) sbt;
      subst_matrix state sbt


  let analyze_init_result e state orig_cost gcost order =
    D.print_init_result e gcost state;
    match e with
      | E_unbound ->
        if debug then fprintf fmt "IOG 1@.";
        let ighost, not_basic = index_of_ghost state in
        if not not_basic then begin
          let aghost = state.matrix.(ighost) in
          deleta_ghost ighost aghost state order
        end;
        if debug then fprintf fmt "IOG 2@.";
        let ighost, not_basic = index_of_ghost state in
        assert (not_basic);
        for i = 0 to state.blast do state.matrix.(i).(ighost) <- Q.zero done;
        orig_cost.(ighost) <- Q.zero;
        raise e

      | E_unsat ->
        if debug then fprintf fmt "IOG 3@.";
        let ighost, not_basic = index_of_ghost state in
        if not not_basic then begin
          let aghost = state.matrix.(ighost) in
          deleta_ghost ighost aghost state order
        end;
        if debug then fprintf fmt "IOG 4@.";
        let ighost, not_basic = index_of_ghost state in
        assert (not_basic);
        for i = 0 to state.blast do state.matrix.(i).(ighost) <- Q.zero done;
        orig_cost.(ighost) <- Q.zero;
        raise e

      | E_max _ ->
        let unsat = ref false in
        if debug then fprintf fmt "IOG 5@.";
        let ighost, not_basic = index_of_ghost state in
        if not not_basic then begin
          let aghost = state.matrix.(ighost) in
          deleta_ghost ighost aghost state order;
          unsat := not (Q.equal state.matrix.(ighost).(0) Q.zero)
        end;
        if debug then fprintf fmt "IOG 6@.";
        let ighost, not_basic = index_of_ghost state in
        assert (not_basic);
        for i = 0 to state.blast do state.matrix.(i).(ighost) <- Q.zero done;
        orig_cost.(ighost) <- Q.zero;
        if !unsat then raise E_unsat

      | _ -> assert false

  (* 0 n'est pas une solution *)
  let init_solution state orig_cost co_opt index order =
    (* Le point 4 du INITIALIZE-SIMPLEX *)
    if debug then fprintf fmt "IOG 7@.";
    let ighost, not_basic = index_of_ghost state in
    assert (not_basic);
    for i = 0 to state.blast do state.matrix.(i).(ighost) <- Q.one done;
    let gcost = Array.make (state.hlast + 1) Q.zero in
    gcost.(ighost) <- Q.m_one;
    let line = index, state.matrix.(index) in

    if debug then fprintf fmt "------------------------------------------@.";
    if debug then fprintf fmt "INIT SOLUTION ..........................@.";
    D.print_given_pb gcost state;
    if debug then fprintf fmt "INIT chosen var's index: %d = ghost@." ighost;
    if debug then
      fprintf fmt "INIT chosen eq:     %a@." (D.pline state.hlast) line;
    (* choix du premier pivot sur la variable fantome *)

    let sbt = change_pivot ighost line state order in
    if debug then fprintf fmt "pivot's result: %a@." (D.psbt state.hlast) sbt;
    subst_matrix state sbt;
    subst state.hlast gcost sbt.lhs sbt.rhs;
    try simplex state gcost co_opt order
    with e -> analyze_init_result e state orig_cost gcost order


  let solve state cost =
    D.print_given_pb cost state;
    let order = Queue.create () in
    for i = 1 to state.hlast do Queue.push i order done;
    let index = line_with_min_const state in
    if debug then fprintf fmt "index_min=%d@." index;
    try
      if Q.compare state.matrix.(index).(0) Q.zero >= 0 then
        simplex state cost None order
      else
        begin
          let co_opt = Some cost in
          (* cost est mis a jour par effet de bord dans co_opt *)
          init_solution state cost co_opt index order;
          simplex state cost None order
        end
    with
      | E_unbound   -> Unbound
      | E_unsat     -> Unsat
      | E_max (c,l) -> Max (c,l)

  let cmp_line len a1 a2 =
    try
      for i = 1 to len do
        let c = Q.compare a1.(i) a2.(i) in
        if c <> 0 then raise (Fin c)
      done;
      0
    with Fin c -> c

  (*
  let test_doublons state =
    let m = Array.init (state.blast+1) (fun i -> Array.copy state.matrix.(i)) in
    Array.fast_sort (cmp_line state.hlast) m;
    let sp = D.max_matrix m in
    (*fprintf fmt "@.@.matrix:@.";
    for i = 0 to state.blast do
      fprintf fmt "%d  = %a@." i (D.parray state.hlast sp) m.(i)
    done;*)
    let cpt = ref 0 in
    for i = 0 to state.blast - 1 do
      if cmp_line state.hlast m.(i) m.(i+1) = 0 then
        incr cpt;
    done;
    fprintf fmt "%d doublons trouves@." !cpt
    *)



  let maximize cost state =
    (*test_doublons state;*)
    if debug then fprintf fmt "Maximize: %a@." P.print cost;
    let l, c = P.to_list cost in
    let m, new_vars, _ = apply_vdiff l state.vdiff in
    match new_vars with
      | _::_ -> Unbound
      | [] ->
        let cost = normalized_array m c state in
        if state.blast <> 0 then solve state cost
        else if l == [] then Max (c, []) else Unbound

end

