(******************************************************************************)
(*     Alt-Ergo: The SMT Solver For Software Verification                     *)
(*     Copyright (C) 2013-2015 --- OCamlPro                                   *)
(*     This file is distributed under the terms of the CeCILL-C licence       *)
(******************************************************************************)

(******************************************************************************)
(*     The Alt-Ergo theorem prover                                            *)
(*     Copyright (C) 2006-2013                                                *)
(*     CNRS - INRIA - Universite Paris Sud                                    *)
(*                                                                            *)
(*     Sylvain Conchon                                                        *)
(*     Evelyne Contejean                                                      *)
(*                                                                            *)
(*     Francois Bobot                                                         *)
(*     Mohamed Iguernelala                                                    *)
(*     Stephane Lescuyer                                                      *)
(*     Alain Mebsout                                                          *)
(*                                                                            *)
(*   This file is distributed under the terms of the CeCILL-C licence         *)
(******************************************************************************)

open Format
open Options
open Sig

module Z = Numbers.Z
module Q = Numbers.Q

let ale = Hstring.make "<="
let alt = Hstring.make "<"
let eval = Hstring.make "@eval"
let is_le n = Hstring.compare n ale = 0
let is_lt n = Hstring.compare n alt = 0
let is_eval n = Hstring.compare n eval = 0

let (-@) l1 l2 = List.rev_append l1 l2

module L = Literal
module Sy = Symbols
module I = Intervals
module T = Term
module MT = Term.Map
module ST = Term.Set

exception NotConsistent of Literal.LT.Set.t

module OracleContainer =
  (val (Inequalities.get_current ()) : Inequalities.Container_SIG)


module type EXTENDED_RELATION = sig
  include RSig.RELATION
  module MP : Map.S
  val get_bounds : t -> Intervals.t MP.t
  val get_bound : MP.key -> t -> Intervals.t option
  val new_improved : t -> MP.key list
end

module Make
  (X : Sig.X)
  (Uf : Sig.UF_SIG with type r = X.r)
  (P : Polynome.EXTENDED_Polynome with type r = X.r) = struct

    module MP = Map.Make(P)
    module SP = Set.Make(P)
    module SX = Set.Make(struct type t = X.r let compare = X.hash_cmp end)
    module MX = Map.Make(struct type t = X.r let compare = X.hash_cmp end)
    module MPL = Literal.LT.Map
    module SPL = Literal.LT.Set


    module Oracle = OracleContainer.Make(X)(Uf)(P)

    type r = P.r
    type uf = Uf.t
    module LR =
      Literal.Make(struct type t = X.r let compare = X.hash_cmp include X end)

    module MR = Map.Make(
      struct
        type t = r L.view
        let compare a b = LR.compare (LR.make a) (LR.make b)
      end)

    module MSR = Set.Make(
      struct
        type t = r L.view
        let compare a b = LR.compare (LR.make a) (LR.make b)
      end)

    let alien_of p = match P.is_monomial p with
      | Some (a,x,b) when Q.equal a Q.one && Q.sign b = 0 -> x
      | _ -> P.embed p

    let poly_of r =
      match P.extract r with
      | Some p -> p
      | None ->
        P.create
          [Numbers.Q.one, r] Numbers.Q.zero (X.type_info r)


    type tau =
      (I.t * SX.t) MX.t ->
      I.t MP.t ->
      I.t

    type t = {
      inequations : Oracle.t MPL.t;
      monomes: (I.t * SX.t) MX.t;
      polynomes : I.t MP.t;
      known_eqs : SX.t;
      improved : SP.t;
      classes : Term.Set.t list;
      size_splits : Q.t;
      equiv : tau option MP.t;
      all_terms : r MT.t;
      roi : Term.t MX.t;
    }

    (*BISECT-IGNORE-BEGIN*)
    module Debug = struct

      let assume a expl =
        if debug_fm () then begin
	  fprintf fmt "[fm] We assume: %a@." LR.print (LR.make a);
	  fprintf fmt "explanations: %a@." Explanation.print expl
        end

      let print_use fmt use =
        SX.iter (fprintf fmt "%a, " X.print) use

      let env env =
        if debug_fm () then begin
	  fprintf fmt "------------ FM: inequations-------------------------@.";
          MPL.iter
            (fun a {Oracle.ple0=p; is_le=is_le} ->
              fprintf fmt "%a%s0  |  %a@."
                P.print p (if is_le then "<=" else "<") L.LT.print a
            )env.inequations;
          fprintf fmt "------------ FM: monomes ----------------------------@.";
	  MX.iter
	    (fun x (i, use) ->
	      fprintf fmt "%a : %a |-use-> {%a}@."
		X.print x I.print i print_use use)
	    env.monomes;
	  fprintf fmt "------------ FM: polynomes---------------------------@.";
	  MP.iter
	    (fun p i ->
	      fprintf fmt "%a : %a@."
		P.print p I.print i)
	    env.polynomes;
	  fprintf fmt "-----------------------------------------------------@."
        end

      let implied_equalities l =
        if debug_fm () then
          begin
            fprintf fmt "[fm] %d implied equalities@." (List.length l);
            List.iter
	      (fun (ra, ex, _) ->
                match ra with
                | LSem ra ->
                  fprintf fmt "  %a %a@."
		    LR.print (LR.make ra) Explanation.print ex
                | LTerm a ->
                  fprintf fmt "  %a %a@."
		    Literal.LT.print a Explanation.print ex
              ) l
          end

      let case_split r1 r2 =
	if debug_fm () then
	  fprintf fmt "[case-split] %a = %a@." X.print r1 X.print r2

      let no_case_split s =
	if debug_fm () then fprintf fmt "[case-split] %s : nothing@." s


      let inconsistent_interval id expl =
	if debug_fm () then
	  fprintf fmt "[IntervalCalculus] interval inconsistent(%d) %a@."
            id Explanation.print expl

      let added_inequation kind ineq =
        if debug_fm () then begin
          fprintf fmt "[fm] I derived the (%s) inequality: %a %s 0@."
            kind P.print ineq.Oracle.ple0 (if ineq.Oracle.is_le then "<=" else "<");
          fprintf fmt "from the following combination:@.";
          Util.MI.iter
            (fun a (coef, ple0, is_le) ->
              fprintf fmt "\t%a * (%a %s 0) + @."
                Q.print coef P.print ple0 (if is_le then "<=" else "<")
            )ineq.Oracle.dep;
          fprintf fmt "\t0@.@."
        end

    end
    (*BISECT-IGNORE-END*)

    let empty classes = {
      inequations = MPL.empty;
      monomes = MX.empty ;
      polynomes = MP.empty ;
      known_eqs = SX.empty ;
      improved = SP.empty ;
      classes = classes;
      size_splits = Q.one;
      equiv = MP.empty;
      all_terms = MT.empty;
      roi = MX.empty
    }

    (*let up_improved env p oldi newi =
      if I.is_strict_smaller newi oldi then
      { env with improved = SP.add p env.improved }
      else env*)

    (** computes an interval for vars = x_1^n_1 * ..... * x_i^n_i
        (1) if some var is not in monomes, then return undefined
        (2) check that all vars are in monomes before doing interval ops **)
    let mult_bornes_vars vars monomes ty =
      try
        let l = List.rev_map (fun (y,n) -> fst (MX.find y monomes), n) vars in
        List.fold_left
          (fun ui (yi,n) -> I.mult ui (I.power n yi))
          (I.point Q.one ty Explanation.empty) l
      with Not_found -> I.undefined ty

    (** computes the interval of a polynome from those of its monomes.
        The monomes are supposed to be already added in env **)
    let intervals_from_monomes ?(monomes_inited=true) env p =
      let pl, v = P.to_list p in
      List.fold_left
        (fun i (a, x) ->
	  let i_x, _  =
            try MX.find x env.monomes
            with Not_found ->
              if monomes_inited then assert false;
              I.undefined (X.type_info x), SX.empty
          in
	  I.add (I.scale a i_x) i
        ) (I.point v (P.type_info p) Explanation.empty) pl


    let rec init_monomes_of_poly are_eq env p use_p expl =
      List.fold_left
        (fun env (_, x) ->
          try
            let u, old_use_x = MX.find x env.monomes in
            { env
              with monomes =
                MX.add x (u, SX.union old_use_x use_p) env.monomes }
          with Not_found ->
            update_monome are_eq expl use_p env x
        ) env (fst (P.to_list p))

    and init_alien are_eq expl p (normal_p, c, d) ty use_x env =
      let env = init_monomes_of_poly are_eq env p use_x expl in
      let i = intervals_from_monomes env p in
      let i =
        try
	  let old_i = MP.find normal_p env.polynomes in
	  let old_i = I.scale d
	    (I.add old_i (I.point c ty Explanation.empty)) in
	  I.intersect i old_i
        with Not_found -> i
      in
      env, i



    and update_monome are_eq expl use_x env x =
      let ty = X.type_info x in
      let ui, env = match  X.ac_extract x with
        | Some {h=h; l=l }
	    when Symbols.equal h (Symbols.Op Symbols.Mult) ->
	  let use_x = SX.singleton x in
	  let env =
	    List.fold_left
	      (fun env (r,_) ->
	        update_monome are_eq expl use_x env r) env l in
	  let m = mult_bornes_vars l env.monomes ty in
	  m, env
        | _ ->
	  match X.term_extract x with
	  | Some t, _ ->
	    let use_x = SX.singleton x in
	    begin
	      match Term.view t with
	      | {Term.f = (Sy.Op Sy.Div); xs = [a; b]} ->
                if false then fprintf fmt "Ici %a@." Term.print t;
		let pa = poly_of (Util.fst3 (X.make a)) in
		let pb = poly_of (Util.fst3 (X.make b)) in
		let (pa', ca, da) as npa = P.normal_form_pos pa in
		let (pb', cb, db) as npb = P.normal_form_pos pb in
		let env, ia =
		  init_alien are_eq expl pa npa ty use_x env in
		let env, ib =
		  init_alien are_eq expl pb npb ty use_x env in
		let ia, ib = match I.doesnt_contain_0 ib with
		  | Yes (ex, _) when Q.equal ca cb
		      && P.compare pa' pb' = 0 ->
		    let expl = Explanation.union ex expl in
		    I.point da ty expl, I.point db ty expl
		  | Yes (ex, _) ->
		    begin
		      match are_eq a b with
		      | Yes (ex_eq, _) ->
			let expl = Explanation.union ex expl in
			let expl = Explanation.union ex_eq expl in
			I.point Q.one ty expl,
			I.point Q.one ty expl
		      | No -> ia, ib
		    end
		  | No -> ia, ib
		in
                let new_i = I.div ia ib in
                let sub = P.sub pa pb in
                if false then fprintf fmt "sub = %a@." P.print sub;
                let ppp, ccc, ddd = P.normal_form_pos sub in
                if false then fprintf fmt "sub = %a = (%a + %a) * %a@."
                  P.print sub P.print ppp Q.print ccc Q.print ddd;
                let new_i =
                  try
                    let pppi = MP.find ppp env.polynomes in
                    if false then
                      fprintf fmt "ppp has interval %a@." I.print pppi;
                    let ccci =
                      I.point ccc (Term.type_info t) Explanation.empty
                    in
                    let pppj = I.add pppi ccci in
                    if false then
                      fprintf fmt "ppp+ccc has interval %a@." I.print pppj;
                    let res = I.scale ddd pppj in
                    if false then
                      fprintf fmt "interval of %a is %a@."
                        P.print sub I.print res;
                    match I.is_positive ib with
                      No -> new_i
                    | Yes (ex,_) ->
                      let rres = I.div res ib in
                      if false then
                        fprintf fmt "final result is %a@." I.print rres;
                      I.intersect new_i rres
                  with Not_found ->
                    new_i
                in
                if false then begin
                  fprintf fmt "%a has interval %a@." Term.print a I.print ia;
                  fprintf fmt "%a has interval %a@." Term.print a I.print ib;
                  fprintf fmt "I.div yields %a@." I.print (I.div ia ib);
                end;
		new_i, env
	      | _ -> I.undefined ty, env
	    end
	  | _ -> I.undefined ty, env
      in
      let u, use_x' =
        try MX.find x env.monomes
        with Not_found -> I.undefined (X.type_info x), use_x in
      let ui = I.intersect ui u in
      { env with monomes = MX.add x (ui, (SX.union use_x use_x')) env.monomes }

    let rec tighten_ac are_eq x env expl =
      let ty = X.type_info x in
      let u, use_x =
        try MX.find x env.monomes
        with Not_found -> I.undefined ty, SX.empty in
      try
        match X.ac_extract x with
	| Some {h=h;t=t;l=[x,n]}
            when Symbols.equal h (Symbols.Op Symbols.Mult) && n mod 2 = 0  ->
	  let u = I.root n u in
	  let u, use_x =
	    let (pu, use_px) =
	      try MX.find x env.monomes
	      with Not_found -> I.undefined ty, SX.empty
	    in
	    I.intersect u pu, use_px in
	  let env =
	      (* let u =  I.new_borne_inf expl Q.zero true u in *)
	    let env =
              { env with monomes = MX.add x (u, use_x) env.monomes } in
	    tighten_non_lin are_eq x use_x env expl
	  in env
	| Some {h=h;t=t;l=[x,n]} when
            Symbols.equal h (Symbols.Op Symbols.Mult) && n > 2 ->
	  let u = I.root n u in
	  let u, use_x =
	    let pu, use_px =
	      try MX.find x env.monomes
	      with Not_found -> I.undefined ty, SX.empty
	    in
	    I.intersect u pu, use_px in
	  let env = { env with monomes = MX.add x (u, use_x) env.monomes } in
	  tighten_non_lin are_eq x use_x env expl
	| _ -> env
      with Q.Not_a_float -> env


    and tighten_div x env expl =
      env

    and tighten_non_lin are_eq x use_x env expl =
      let env' = tighten_ac are_eq x env expl in
      let env' = tighten_div x env' expl in
      (*let use_x = SX.union use1_x use2_x in*)
      (* let i, _ = MX.find x env.monomes in *)
      SX.fold
        (fun x acc ->
	  let _, use = MX.find x acc.monomes in
	  (*if I.is_strict_smaller new_i i then*)
	  update_monome are_eq expl use acc x
       (*else acc*))
        use_x env'

    let update_monomes_from_poly p i polynomes monomes =
      let lp, _ = P.to_list p in
      let ty = P.type_info p in
      List.fold_left (fun monomes (a,x) ->
        let np = P.remove x p in
        let (np,c,d) = P.normal_form_pos np in
        try
	  let inp = MP.find np polynomes in
	  let new_ix =
	    I.scale
	      (Q.div Q.one a)
	      (I.add i
	         (I.scale (Q.minus d)
		    (I.add inp
		       (I.point c ty Explanation.empty)))) in
	  let old_ix, ux = MX.find x monomes in
	  let ix = I.intersect old_ix new_ix in
	  MX.add x (ix, ux) monomes
        with Not_found -> monomes)
        monomes lp

    let update_polynomes_intervals env =
      let polynomes, monomes, improved = MP.fold
        (fun p ip (polynomes, monomes, improved) ->
	  let new_i = intervals_from_monomes env p in
	  let i = I.intersect new_i ip in
	  if I.is_strict_smaller i ip then
	    let monomes = update_monomes_from_poly p i polynomes monomes in
	    MP.add p i polynomes, monomes, SP.add p improved
	  else polynomes, monomes, improved
        ) env.polynomes (env.polynomes, env.monomes, env.improved) in
      {env with polynomes = polynomes; monomes = monomes ; improved = improved}


    let find_one_eq x u =
      match I.is_point u with
      | Some (v, ex) when X.type_info x != Ty.Tint || Q.is_int v ->
        let eq =
	  LR.mkv_eq x (alien_of (P.create [] v (X.type_info x))) in
	Some (LSem eq, ex, Sig.Other)
      | _ -> None

    let find_eq eqs x u env =
      match find_one_eq x u with
      | None -> eqs
      | Some eq1 ->
	begin
          match X.ac_extract x with
	  | Some {h = h; l = [y,n]}
	      when Symbols.equal h (Symbols.Op Symbols.Mult) ->
	    let neweqs = try
		           let u, _ = MX.find y env.monomes in
		           match find_one_eq y u with
		           | None -> eq1::eqs
		           | Some eq2 -> eq1::eq2::eqs
	      with Not_found -> eq1::eqs
	    in neweqs
	  | _ -> eq1::eqs
	end

    type ineq_status =
    | Trivial_eq
    | Trivial_ineq of Q.t
    | Bottom
    | Monome of Q.t * P.r * Q.t
    | Other

    let ineq_status {Oracle.ple0 = p ; is_le = is_le} =
      match P.is_monomial p with
	Some (a, x, v) -> Monome (a, x, v)
      | None ->
	if P.is_empty p then
	  let _, v = P.separate_constant p in
	  let c = Q.sign v (* equiv. to compare v Q.zero *) in
	  if c > 0 || (c >=0 && not is_le) then Bottom
	  else
	    if c = 0 && is_le then Trivial_eq
	    else Trivial_ineq v
	else Other

    (*let ineqs_from_dep dep borne_inf is_le =
      List.map
      (fun {poly_orig = p; coef = c} ->
      let (m,v,ty) = P.mult_const minusone p in
      (* quelle valeur pour le ?????? *)
      { ple0 = {poly = (m, v +/ (Q.div borne_inf c), ty); le = is_le} ;
      dep = []}
      )dep*)

    let mk_equality p =
      let r1 = alien_of p in
      let r2 = alien_of (P.create [] Q.zero (P.type_info p)) in
      LSem (LR.mkv_eq r1 r2)

    let fm_equalities eqs { Oracle.ple0 = p; dep = dep; expl = ex } =
      Util.MI.fold
	(fun _ (_, p, _) eqs ->
         (mk_equality p, ex, Sig.Other) :: eqs
	) dep eqs


    let update_intervals are_eq env eqs expl (a, x, v) is_le =
      let uints, use_x =
        match X.ac_extract x with
	| Some {h=h; l=l}
	    when Symbols.equal h (Symbols.Op Symbols.Mult) ->
	  let u', use_x' = MX.find x env.monomes in
	  let m = mult_bornes_vars l env.monomes (X.type_info x) in
	  I.intersect m u', use_x'
	| _ -> MX.find x env.monomes
      in
      let b = Q.div (Q.mult Q.m_one v) a in
      let u =
        if Q.sign a > 0 then
	  I.new_borne_sup expl b is_le uints
        else
	  I.new_borne_inf expl b is_le uints in
      let env = { env with monomes = MX.add x (u, use_x) env.monomes } in
      let env =  tighten_non_lin are_eq x use_x env expl in
      env, (find_eq eqs x u env)

    let update_ple0 are_eq env p0 is_le expl =
      if P.is_empty p0 then env
      else
        let ty = P.type_info p0 in
        let a, _ = P.choose p0 in
        let p, change =
	  if Q.sign a < 0 then
	    P.mult_const Q.m_one p0, true
	  else p0, false in
        let p, c, _ = P.normal_form p in
        let c = Q.minus c in
        let u =
	  if change then
            I.new_borne_inf expl c is_le (I.undefined ty)
	  else
	    I.new_borne_sup expl c is_le (I.undefined ty) in
        let u, pu =
	  try
	    let pu = MP.find p env.polynomes in
	    let i = I.intersect u pu in
	    i, pu
	  with Not_found -> u, I.undefined ty
        in
        let env =
	  if I.is_strict_smaller u pu then
	    let polynomes = MP.add p u env.polynomes in
	    let monomes = update_monomes_from_poly p u polynomes env.monomes in
	    let improved = SP.add p env.improved in
	    { env with
	      polynomes = polynomes;
	      monomes = monomes;
	      improved = improved }
	  else env
        in
        match P.to_list p0 with
        | [a,x], v ->
	  fst(update_intervals are_eq env [] expl (a, x, v) is_le)
        | _ -> env

    let add_inequations are_eq acc lin =
      List.fold_left
        (fun ((env, eqs) as acc) ineq ->
	  let expl = ineq.Oracle.expl in
	  match ineq_status ineq with
	  | Bottom           ->
            Debug.added_inequation "Bottom" ineq;
	    raise (Sig.Inconsistent (expl, env.classes))

	  | Trivial_eq       ->
            Debug.added_inequation "Trivial_eq" ineq;
	    env, fm_equalities eqs ineq

	  | Trivial_ineq  c  ->
            Debug.added_inequation "Trivial_ineq"  ineq;
	    let n, pp =
	      Util.MI.fold
		(fun _ (_, p, is_le) ((n, pp) as acc) ->
		  if is_le then acc else
		    match pp with
		    | Some _ -> n+1, None
		    | None when n=0 -> 1, Some p
		    | _ -> n+1, None) ineq.Oracle.dep (0,None)
	    in
	    let env =
	      Util.MI.fold
		(fun _ (coef, p, is_le) env ->
		  let ty = P.type_info p in
		  let is_le =
		    match pp with
		      Some x -> P.compare x p = 0 | _ -> is_le && n=0
		  in
		  let p' = P.sub (P.create [] (Q.div c coef) ty) p in
		  update_ple0 are_eq env p' is_le expl
		) ineq.Oracle.dep env
	    in
	    env, eqs

	  | Monome (a, x, v) ->
            Debug.added_inequation "Monome" ineq;
	    let env, eqs =
	      update_intervals
                are_eq env eqs expl (a, x, v) ineq.Oracle.is_le
	    in
	    env, eqs

	  | Other            ->
	    acc


        ) acc lin

    let split_problem env ineqs aliens =
      let current_age = Oracle.current_age () in
      let l, all_lvs =
        List.fold_left
          (fun (acc, all_lvs) ({Oracle.ple0=p} as ineq) ->


	    match ineq_status ineq with
	    | Trivial_eq | Trivial_ineq _ -> (acc, all_lvs)
	    | Bottom ->
              raise (Sig.Inconsistent (ineq.Oracle.expl, env.classes))
	    | _ ->
              let lvs =
                List.fold_left (fun acc e -> SX.add e acc) SX.empty (aliens p)
              in
              ([ineq], lvs) :: acc , SX.union lvs all_lvs
          )([], SX.empty) ineqs
      in
      let ll =
        SX.fold
          (fun x l ->
            let lx, l_nx = List.partition (fun (_,s) -> SX.mem x s) l in
            match lx with
            | [] -> assert false
            | e:: lx ->
              let elx =
                List.fold_left
                  (fun (l, lvs) (l', lvs') ->
                    List.rev_append l l', SX.union lvs lvs') e lx in
              elx :: l_nx
          ) all_lvs l
      in
      let ll =
        List.filter
          (fun (ineqs, _) ->
            List.exists
              (fun ineq -> Z.equal current_age ineq.Oracle.age) ineqs
          )ll
      in
      List.fast_sort (fun (a,_) (b,_) -> List.length a - List.length b) ll

    let is_normalized_poly uf p =
      let p = alien_of p in
      let rp, _  = Uf.find_r uf p in
      if X.equal p rp then true
      else begin
        fprintf fmt "%a <= 0 NOT normalized@." X.print p;
        fprintf fmt "It is equal to %a@." X.print rp;
        false
      end

    let better_upper_bound_from_intervals env p =
      try I.borne_sup (MP.find p env.polynomes)
      with
      | I.No_finite_bound ->
         assert false (*env.polynomes is up to date w.r.t. ineqs*)

      | Not_found ->
         try
           let q = P.mult_const Q.m_one p in
           let bi, ex, is_large = I.borne_inf (MP.find q env.polynomes) in
           Q.minus bi, ex, is_large
         with
         | I.No_finite_bound ->
            assert false (*env.polynomes is up to date w.r.t. ineqs*)

         | Not_found ->
            assert false

    let better_bound_from_intervals env ({Oracle.ple0; is_le; dep} as v) =
      let p, c = P.separate_constant ple0 in
      assert (not (P.is_empty p));
      assert (let _,_,d = P.normal_form p in Q.equal d Q.one);
      let cur_up_bnd = Q.minus c in
      let i_up_bnd, expl, is_large = better_upper_bound_from_intervals env p in
      let new_p = P.add_const (Q.minus i_up_bnd) p in
      let a = match Util.MI.bindings dep with [a,_] -> a | _ -> assert false in
      let cmp = Q.compare i_up_bnd cur_up_bnd in
      assert (cmp <= 0);
      if cmp = 0 then
        match is_le, is_large with
        | false, true -> assert false (* intervals are normalized wrt ineqs *)
        | false, false | true, true -> v (* no change *)
        | true , false -> (* better bound, Large ineq becomes Strict *)
          {v with
            Oracle.ple0 = new_p;
            expl = expl;
            is_le = false;
            dep = Util.MI.singleton a (Q.one, new_p, false)}
      else (* new bound is better. i.e. i_up_bnd < cur_up_bnd *)
        {v with
          Oracle.ple0 = new_p;
          expl = expl;
          is_le = is_large;
          dep = Util.MI.singleton a (Q.one, new_p, is_large)}

    let args_of p = List.rev_map snd (fst (P.to_list p))

    let fm uf are_eq env eqs =
      if debug_fm () then fprintf fmt "[fm] in fm/fm-simplex@.";
      Options.tool_req 4 "TR-Arith-Fm";
      let ineqs =
        MPL.fold (fun k v acc ->
          assert (is_normalized_poly uf v.Oracle.ple0);
          (better_bound_from_intervals env v) :: acc
        ) env.inequations []
      in
      (*let pbs = split_problem env ineqs (fun p -> P.leaves p) in*)
      let pbs = split_problem env ineqs args_of in
      let res = List.fold_left
        (fun acc (ineqs, _) ->
          let mp = Oracle.MINEQS.add_to_map Oracle.MINEQS.empty ineqs in
          Oracle.available add_inequations are_eq acc mp)
        (env, eqs) pbs
      in
      if debug_fm () then fprintf fmt "[fm] out fm/fm-simplex@.";
      res

    let is_num r =
      let ty = X.type_info r in ty == Ty.Tint || ty == Ty.Treal

    let add_disequality are_eq env eqs p expl =
      let ty = P.type_info p in
      match P.to_list p with
      | ([], v) ->
        if Q.sign v = 0 then
          raise (Sig.Inconsistent (expl, env.classes));
	env, eqs
      | ([a, x], v) ->
	let b = Q.div (Q.minus v) a in
	let i1 = I.point b ty expl in
	let i2, use2 =
	  try
	    MX.find x env.monomes
	  with Not_found -> I.undefined ty, SX.empty
	in
	let i = I.exclude i1 i2 in
	let env ={ env with monomes = MX.add x (i,use2) env.monomes } in
	let env = tighten_non_lin are_eq x use2 env expl in
	env, find_eq eqs x i env
      | _ ->
	let a, _ = P.choose p in
	let p = if Q.sign a >= 0 then p
	  else P.mult_const Q.m_one p in
	let p, c, _ = P.normal_form p in
	let i1 = I.point (Q.minus c) ty expl in
	let i2 =
	  try
	    MP.find p env.polynomes
	  with Not_found -> I.undefined ty
	in
	let i = I.exclude i1 i2 in
	let env =
	  if I.is_strict_smaller i i2 then
	    let polynomes = MP.add p i env.polynomes in
	    let monomes = update_monomes_from_poly p i polynomes env.monomes
	    in
	    let improved = SP.add p env.improved in
	    { env with
	      polynomes = polynomes;
	      monomes = monomes;
	      improved = improved}
	  else env
	in
	env, eqs

    let add_equality are_eq env eqs p expl =
      let ty = P.type_info p in
      match P.to_list p with
      | ([], v) ->
        if Q.sign v <> 0 then
	  raise (Sig.Inconsistent (expl, env.classes));
        env, eqs

      | ([a, x], v) ->
	let b = Q.div (Q.minus v) a in
	let i = I.point b ty expl in
	let i, use =
	  try
	    let i', use' = MX.find x env.monomes in
	    I.intersect i i', use'
	  with Not_found -> i, SX.empty
	in
	let env = { env with monomes = MX.add x (i, use) env.monomes} in
	let env = tighten_non_lin are_eq x use env expl in
	env, find_eq eqs x i env
      | _ ->
	let a, _ = P.choose p in
	let p = if Q.sign a >= 0 then p
	  else P.mult_const Q.m_one p in
	let p, c, _ = P.normal_form p in
	let i = I.point (Q.minus c) ty expl in
	let i, ip =
	  try
	    let ip =  MP.find p env.polynomes in
	    I.intersect i ip, ip
	  with Not_found -> i, I.undefined ty
	in
	let env =
	  if I.is_strict_smaller i ip then
	    let polynomes = MP.add p i env.polynomes in
	    let monomes = update_monomes_from_poly p i polynomes env.monomes
	    in
	    let improved = SP.add p env.improved in
	    { env with
	      polynomes = polynomes;
	      monomes = monomes;
	      improved = improved }
	  else env
	in
	let env =
	  { env with
	    known_eqs = SX.add (alien_of p) env.known_eqs
          } in
	env, eqs

    let normal_form a = match a with
      | L.Builtin (false, n, [r1; r2])
          when is_le n && X.type_info r1 == Ty.Tint ->
        let pred_r1 = P.sub (poly_of r1) (P.create [] Q.one Ty.Tint) in
	LR.mkv_builtin true  n  [r2; alien_of pred_r1]

      | L.Builtin (true, n, [r1; r2]) when
	  not (is_le n) && X.type_info r1 == Ty.Tint ->
        let pred_r2 = P.sub (poly_of r2) (P.create [] Q.one Ty.Tint) in
	LR.mkv_builtin true ale [r1; alien_of pred_r2]

      | L.Builtin (false, n, [r1; r2]) when is_le n ->
	LR.mkv_builtin true alt [r2; r1]

      | L.Builtin (false, n, [r1; r2]) when is_lt n ->
	LR.mkv_builtin true ale [r2; r1]

      | _ -> a

    let remove_trivial_eqs eqs la =
      let la, lb =
        List.fold_left
          (fun (s, s') (a, root, _, _) ->
            let s = MSR.add a s in
            match root with
            | None -> s, s'
            | Some at -> s, SPL.add at s'
          )(MSR.empty, SPL.empty) la
      in
      let eqs, _, _ =
        List.fold_left
          (fun ((eqs, s, s') as acc) ((sa, ex, orig) as e) ->
            match sa with
            | LSem sa ->
              if MSR.mem sa s then acc else e :: eqs, MSR.add sa s, s'
            | LTerm a ->
              if SPL.mem a s' then acc else e :: eqs, s, SPL.add a s'
          )([], la, lb) eqs
      in
      eqs

    let equalities_from_polynomes env eqs =
      let known, eqs =
        MP.fold
          (fun p i (knw, eqs) ->
            let xp = alien_of p in
            if SX.mem xp knw then knw, eqs
            else
              match I.is_point i with
              | Some (num, ex) ->
                let r2 = alien_of (P.create [] num (P.type_info p)) in
                SX.add xp knw,
                (LSem (LR.mkv_eq xp r2), ex, Sig.Other) :: eqs
              | None -> knw, eqs
          ) env.polynomes  (env.known_eqs, eqs)
      in {env with known_eqs= known}, eqs



    let equalities_from_monomes env eqs =
      let known, eqs =
        MX.fold
          (fun x (i,_) (knw, eqs) ->
            if SX.mem x knw then knw, eqs
            else
              match I.is_point i with
              | Some (num, ex) ->
                let r2 = alien_of (P.create [] num (X.type_info x)) in
                SX.add x knw,
                (LSem (LR.mkv_eq x r2), ex, Sig.Other) :: eqs
              | None -> knw, eqs
          ) env.monomes  (env.known_eqs, eqs)
      in {env with known_eqs= known}, eqs

    let equalities_from_intervals env eqs =
      let env, eqs = equalities_from_polynomes env eqs in
      equalities_from_monomes env eqs

    let count_splits env la =
      let nb =
        List.fold_left
          (fun nb (_,_,_,i) ->
            match i with
            | CS (Th_arith, n) -> Numbers.Q.mult nb n
            | _ -> nb
          )env.size_splits la
      in
      {env with size_splits = nb}

    let remove_ineq a ineqs =
      match a with None -> ineqs | Some a -> MPL.remove a ineqs

    let add_ineq a v ineqs =
      match a with None -> ineqs | Some a -> MPL.add a v ineqs

    (* begin experimental section *)

    let is_mult h = Sy.equal (Sy.Op Sy.Mult) h

    let interval_of p mx mp =
      let pl, v = P.to_list p in
      (* interval_from_monomes *)
      let res =
        List.fold_left
          (fun i (a, x) ->
	    let i_x, _  =
              try MX.find x mx
              with Not_found -> I.undefined (X.type_info x), SX.empty
            in
	    I.add (I.scale a i_x) i
          ) (I.point v (P.type_info p) Explanation.empty) pl
      in
      (* should have something to compute it from polynomes *)
      res

    let poly_in_nonlinear_form ty l =
      List.fold_left
        (fun (acc, cpt) (c, x) ->
          match X.ac_extract x with
          | Some ac when is_mult ac.h ->
            assert (ac.t == ty);
            let mx =
              List.fold_left (fun mx (x,n) -> MX.add x n mx)MX.empty ac.l
            in
            (c, mx) :: acc, cpt + 1
          | _ ->
            (c, MX.singleton x 1) :: acc, cpt
        )([], 0) l

    let count_occ_of_nonlinear_leaves l =
      List.fold_left
        (fun mx (c, mlt) ->
          MX.fold
            (fun x rdd m ->
              try
                let cpt, rdd0 = MX.find x m in
                MX.add x (cpt+1, max rdd rdd0) m
              with Not_found -> MX.add x (1, rdd) m
            )mlt mx
        )MX.empty l

    let poly_of_expanded l0 ty =
      let l, c, p =
        List.fold_left
          (fun (l_acc, c_acc, p_acc) (coef, mx) ->
            match MX.bindings mx with
              [] -> assert (c_acc == Q.zero); (l_acc, coef, p_acc)
            | lx ->
              assert (List.for_all (fun (_,rdd) -> rdd > 0) lx);
              let ac = {l = lx; t=ty; h = Sy.Op Sy.Mult; distribute=true} in
              let r = X.ac_embed ac in
              match P.extract r with
              | None -> (coef, r) :: l_acc, c_acc, p_acc
              | Some p -> l_acc, c_acc, P.add p_acc (P.mult_const coef p)
          )([], Q.zero, P.create [] Q.zero ty) l0
      in
      P.add p (P.create l c ty)

    let add_factorization_by x rdd p l equiv ty =
      if debug_fm() then
        fprintf fmt "@.Factorization of %a by (%a ^ %d) is possible@."
          P.print p X.print x rdd;
      let dv, rem =
        List.fold_left
          (fun (dv, rem) ((c, (mx : int MX.t)) as e) ->
            try
              let rdd' = MX.find x mx in
              if rdd > rdd' then dv, e :: rem
              else
                let d = rdd' - rdd in
                let mx = if d = 0 then MX.remove x mx else MX.add x d mx in
                (c, mx) :: dv, rem
            with Not_found -> dv, e :: rem
          )([], []) l
      in
      let p_dv = poly_of_expanded dv ty in
      match P.to_list p_dv with
      | [], c when Q.equal c Q.one ->
        if debug_fm () then fprintf fmt "trivial factorization@.";
        equiv

      (*| case ([Q.one, _], Q.zero) -> kept *)

      | l_dv, c_dv ->
        let p_rem = poly_of_expanded rem ty in
        let fact = {l = [x,rdd]; t=ty; h = Sy.Op Sy.Mult; distribute=true} in
        let fact = X.ac_embed fact in
        if debug_fm() then
          fprintf fmt "(%a) =@.(%a) mult (%a) + %a@.@."
            P.print p P.print p_dv X.print fact  P.print p_rem;
        (*** verification ***)
        assert
          (
            let exct =
              X.ac_embed {l=[alien_of p_dv, 1 ; fact, 1];
                          t=ty;
                          h=Sy.Op Sy.Mult;
                          distribute= true}
            in
            let q = P.add (poly_of exct) p_rem in
            let xp = alien_of p in
            let xq = alien_of q in
            if debug_fm () then begin
              fprintf fmt "problem is:@.%a = %a@.@."
                X.print xp X.print xq;
            end;
            let pfact = poly_of fact in
            let acc = P.mult_const Q.m_one p in
            let acc = P.add acc (P.mult_const c_dv pfact) in
            let acc = P.add acc p_rem in
            let res =
              List.fold_left
                (fun acc (c, x) ->
                  P.add acc (P.mult_const c (P.mult (poly_of x) pfact))
                )acc l_dv
            in
            match P.to_list res with
              [], c -> Q.is_zero c
            | _ ->
              try
                match X.solve (alien_of res) (alien_of (P.create [] Q.zero ty))
                with
                | [] ->  true
                | _ ->
                  if debug_fm () then
                    fprintf fmt "res = %a has a solution@." P.print res;
                  false
              with _ ->
                if debug_fm () then
                  fprintf fmt "res = %a is inconsistent@." P.print res;
                false
          );
        try
          let new_fun mx mp =
            let idv = interval_of p_dv mx mp in
            let ifact = interval_of (poly_of fact) mx mp in
            let irem = interval_of p_rem mx mp in
            I.add (I.mult idv ifact) irem
          in
          match MP.find p equiv with
          | None -> MP.add p (Some new_fun) equiv
          | Some f ->
            let comb mx mp = I.intersect (f mx mp) (new_fun mx mp) in
            MP.add p (Some comb) equiv
        with Not_found -> assert false

    let init_equivalences_of_poly p equiv =
      let l, c = P.to_list p in
      assert (c == Q.zero);
      let ty = P.type_info p in
      let l, nb_orig_ac = poly_in_nonlinear_form ty l in
      if nb_orig_ac = 0 then equiv
      else
        let mx = count_occ_of_nonlinear_leaves l in
        let equiv = ref equiv in
        MX.iter (fun x (cpt, rdd) ->
          if cpt > 1 then
            begin
              assert (rdd >= 1);
              for i = 1 to rdd do
                equiv := add_factorization_by x i p l !equiv ty
              done;
            end
        )mx;
        !equiv

    let add_new_equivalences env =
      let equiv =
        MP.fold
          (fun p _ equiv ->
            if MP.mem p equiv then equiv
            else init_equivalences_of_poly p (MP.add p None equiv)
          )env.polynomes env.equiv
      in
      {env with equiv = equiv}


    let better_poly_bounds_modulo_equivalences env =
      let polynomes =
        MP.fold
          (fun p fopt polynomes ->
            match fopt with
            | None -> polynomes
            | Some f ->
              try
                let i = MP.find p polynomes in
                let j = f env.monomes polynomes in
                MP.add p (I.intersect i j) polynomes
              with Not_found -> polynomes
          )env.equiv env.polynomes
      in
      {env with polynomes = polynomes}

    let update_non_linear_equivalences env =
      let env = add_new_equivalences env in
      better_poly_bounds_modulo_equivalences env

    (*/ begin experimental section *)

    let mk_roi_of t =
      T.make (Sy.Op Sy.Real_of_int) [t] Ty.Treal

    let extra_deductions_aux env uf r1 r2 expl acc =
      match P.is_const (poly_of r2) with
      | None ->
        begin
          try
            let roi = MX.find r2 env.roi in
            match P.extract r1 with
            | Some p ->
              assert (P.type_info p == Ty.Tint);
              let l, c = P.to_list p in
              assert (Q.is_int c);
              let t = mk_roi_of (T.int (Q.to_string c)) in
              let t =
                List.fold_left
                  (fun t (c, x) ->
                    assert (Q.is_int c);
                    let s1 = mk_roi_of (T.int (Q.to_string c)) in
                    let s2 =
                      match X.term_extract x with
                      | None, _ -> assert false (*TODO*)
                      | Some e, _ ->
                        assert (T.type_info e == Ty.Tint);
                        mk_roi_of e
                    in
                    let s = T.make (Sy.Op Sy.Mult) [s1;s2] Ty.Treal in
                    T.make (Sy.Op Sy.Plus) [t; s] Ty.Treal
                  )t l
              in
              let eq = Literal.LT.mk_eq roi t in
              if debug_fm() then fprintf fmt "deduce %a@." Literal.LT.print eq;
              (* eq on terms because we have new terms to add to TH' env*)
              (LTerm eq, expl, Sig.Other) :: acc
            | None ->
              match X.ac_extract r1 with
              | Some ac when Pervasives.(=) ac.h (Sy.Op Sy.Mult) ->
                assert (ac.t == Ty.Tint);
                let t_opt =
                  List.fold_left
                    (fun t_opt (x, n) ->
                      assert (X.type_info x == Ty.Tint);
                      match X.term_extract x with
                        None, _ -> assert false
                      | Some t, _ ->
                        assert (n >= 1);
                        let roit = mk_roi_of t in
                        let acc = ref roit in
                        for i = 2 to n do
                          acc := T.make (Sy.Op Sy.Mult) [!acc; roit] Ty.Treal
                        done;
                        match t_opt with
                        | None -> Some !acc
                        | Some z ->
                          Some (T.make (Sy.Op Sy.Mult) [!acc; z] Ty.Treal)
                    )None ac.l
                in
                begin
                  match t_opt with
                    None -> assert false
                  | Some t ->
                    let eq = Literal.LT.mk_eq roi t in
                    if debug_fm() then
                      fprintf fmt "AC deduce %a@." Literal.LT.print eq;
                  (* eq on terms because we have new terms to add to TH' env*)
                    (LTerm eq, expl, Sig.Other) :: acc
                end
              | _ -> acc
          with Not_found -> acc
        end

      | Some c ->
        let acc =
          match X.term_extract r1 with
          | None, _ | Some _, false -> acc
          | Some t, true ->
            match Term.view t with
            | {T.f=Sy.Op Sy.Real_of_int; xs=[a]; ty=Ty.Treal} ->
              assert (T.type_info a == Ty.Tint);
              if not (Q.is_int c) then
                raise (Sig.Inconsistent (expl, env.classes));
              let tc = T.int (Q.to_string c) in
              let eq = Literal.LT.mk_eq a tc in
              if debug_fm() then
                fprintf fmt "[fm.extra_deductions_aux 1] deduce %a@."
                  Literal.LT.print eq;
              (LTerm eq, expl, Sig.Other) :: acc
            | _ -> acc
        in
        (try
           let roi = MX.find r1 env.roi in
           let tc = T.real (Q.to_string c) in
           let eq = Literal.LT.mk_eq roi tc in
           if debug_fm() then
             fprintf fmt "[fm.extra_deductions_aux 2] deduce %a@."
               Literal.LT.print eq;
           (LTerm eq, expl, Sig.Other) :: acc
         with Not_found -> acc)

    let extra_deductions env uf r1 r2 expl eqs orig =
      match orig with
      | Sig.Subst ->
          let eqs = extra_deductions_aux env uf r1 r2 expl eqs in
          extra_deductions_aux env uf r2 r1 expl eqs
      | _ -> eqs

    let assume env uf la =
      Oracle.incr_age ();
      let env = count_splits env la in
      let are_eq = Uf.are_equal uf in
      let classes = Uf.cl_extract uf in
      let env = {env with improved = SP.empty; classes = classes} in
      Debug.env env;
      let nb_num = ref 0 in
      let env, eqs, new_ineqs, to_remove =
        List.fold_left
	  (fun ((env, eqs, new_ineqs, rm) as acc) (a, root, expl, orig) ->
	    let a = normal_form a in
	    Debug.assume a expl;
	    try
              match a with
	      | L.Builtin(_, n, [r1;r2]) when is_le n || is_lt n ->
                incr nb_num;
		let p1 = poly_of r1 in
		let p2 = poly_of r2 in
		let ineq =
                  Oracle.create_ineq p1 p2 (is_le n) root expl in
                begin
                  match ineq_status ineq with
                  | Bottom -> raise (Sig.Inconsistent (expl, env.classes))
                  | Trivial_eq | Trivial_ineq _ ->
                     {env with inequations=remove_ineq root env.inequations},
                     eqs, new_ineqs,
                    (match root with None -> rm | Some a -> a:: rm)

                  | Monome _
                  | Other ->
		     let env =
		       init_monomes_of_poly
                         are_eq env ineq.Oracle.ple0 SX.empty
                         Explanation.empty
                     in
		     let env =
		       update_ple0
                         are_eq env ineq.Oracle.ple0 (is_le n) expl
                     in
		     {env with inequations=add_ineq root ineq env.inequations},
                     eqs, true, rm
                end

	      | L.Builtin(_, n, [r]) when is_eval n && is_num r ->
                let p = poly_of r in
                let env =
                  init_monomes_of_poly are_eq env p SX.empty
                    Explanation.empty
                in
                let p, _, _ = P.normal_form_pos p in
                let ty = X.type_info r in
                let env =
                  if MP.mem p env.polynomes then env
                  else
                    let i = I.undefined ty in
                    {env with polynomes = MP.add p i env.polynomes}
                in
                env, eqs, true, rm



	      | L.Distinct (false, [r1; r2]) when is_num r1 && is_num r2 ->
                 incr nb_num;
		 let p = P.sub (poly_of r1) (poly_of r2) in
                 begin
                   match P.is_const p with
                   | Some c ->
                      if Q.is_zero c then (* bottom *)
                        raise (Sig.Inconsistent (expl, env.classes))
                      else (* trivial *)
                        let rm = match root with Some a -> a::rm | None -> rm in
                        env, eqs, new_ineqs, rm
                   | None ->
		      let env = init_monomes_of_poly are_eq env p SX.empty
                                                     Explanation.empty
                      in
		      let env, eqs = add_disequality are_eq env eqs p expl in
                      env, eqs, new_ineqs, rm
                 end

	      | L.Eq(r1, r2) when is_num r1 && is_num r2 ->
                incr nb_num;
                let eqs = extra_deductions env uf r1 r2 expl eqs orig in
		let p = P.sub (poly_of r1) (poly_of r2) in
		let env = init_monomes_of_poly are_eq env p SX.empty
                  Explanation.empty
                in
		let env, eqs = add_equality are_eq env eqs p expl in
                env, eqs, new_ineqs, rm

	      | _ -> acc

	    with I.NotConsistent expl ->
              Debug.inconsistent_interval 1 expl ;
	      raise (Sig.Inconsistent (expl, env.classes))
	  )
	  (env, [], false, []) la

      in
      if !nb_num = 0 then
        env, {assume=[]; remove = to_remove}, Term.Set.empty
      else
        try
          (* we only call fm when new ineqs are assumed *)
          let env, eqs =
            if new_ineqs then fm uf are_eq env eqs
            else env, eqs
          in

          let env = update_non_linear_equivalences env in
          let env = update_polynomes_intervals env in
          let env, eqs = equalities_from_intervals env eqs in
          Debug.env env;
          let eqs = remove_trivial_eqs eqs la in
          Debug.implied_equalities eqs;
          let to_assume = (List.rev eqs) in (* List.rev for compatibility *)
          env, {assume = to_assume; remove = to_remove}, Term.Set.empty
        with I.NotConsistent expl ->
          Debug.inconsistent_interval 2 expl ;
          raise (Sig.Inconsistent (expl, env.classes))


    let assume env uf la =
      let env, res, terms = assume env uf la in
      let polys =
        MP.fold
          (fun p _ mp ->
            if Uf.is_normalized uf (alien_of p) then mp else MP.remove p mp)
          env.polynomes env.polynomes
      in
      MX.iter (fun x _ -> Uf.assert_leaves_are_inited uf x) env.monomes;
      MP.iter (fun p _ -> Uf.assert_leaves_are_inited uf (alien_of p)) polys;
      {env with polynomes = polys}, res, terms

    let query env uf a_ex =
      try
        ignore(assume env uf [a_ex]);
        No
      with Sig.Inconsistent (expl, classes) -> Yes (expl, classes)


    let assume env uf la =
      if Options.timers() then
        try
	  Options.exec_timer_start Timers.M_Arith Timers.F_assume;
	  let res = assume env uf la in
	  Options.exec_timer_pause Timers.M_Arith Timers.F_assume;
	  res
        with e ->
	  Options.exec_timer_pause Timers.M_Arith Timers.F_assume;
	  raise e
      else assume env uf la

    let query env uf la =
      if Options.timers() then
        try
	  Options.exec_timer_start Timers.M_Arith Timers.F_query;
	  let res = query env uf la in
	  Options.exec_timer_pause Timers.M_Arith Timers.F_query;
	  res
        with e ->
	  Options.exec_timer_pause Timers.M_Arith Timers.F_query;
	  raise e
      else query env uf la

    let case_split_polynomes env =
      let o = MP.fold
        (fun p i o ->
	  match I.finite_size i with
	  | Some s when Q.compare s Q.one > 0 ->
	    begin
	      match o with
	      | Some (s', p', _) when Q.compare s' s < 0 -> o
	      | _ ->
	        let n, ex, is_large = I.borne_inf i in
                assert (is_large);
		Some (s, p, n)
	    end
	  | _ -> o
        ) env.polynomes None in
      match o with
      | Some (s, p, n) ->
        let r1 = alien_of p in
	let r2 = alien_of (P.create [] n  (P.type_info p)) in
        Debug.case_split r1 r2;
	[LR.mkv_eq r1 r2, CS (Th_arith, s)], s
      | None ->
        Debug.no_case_split "polynomes";
	[], Q.zero

    let case_split_monomes env =
      let o = MX.fold
        (fun x (i,_) o ->
	  match I.finite_size i with
	  | Some s when Q.compare s Q.one > 0 ->
	    begin
	      match o with
	      | Some (s', _, _) when Q.compare s' s < 0 -> o
	      | _ ->
                let n, ex, is_large = I.borne_inf i in
                assert (is_large);
		Some (s, x, n)
	    end
	  | _ -> o
        ) env.monomes None in
      match o with
      | Some (s,x,n) ->
        let ty = X.type_info x in
        let r1 = x in
	let r2 = alien_of (P.create [] n  ty) in
        Debug.case_split r1 r2;
	[LR.mkv_eq r1 r2, CS (Th_arith, s)], s
      | None ->
        Debug.no_case_split "monomes";
	[], Q.zero

    let check_size env res = match res with
      | [] -> res
      | [_, CS (Th_arith, s)] ->
        if Numbers.Q.compare (Q.mult s env.size_splits) (max_split ()) <= 0  ||
	  Numbers.Q.sign  (max_split ()) < 0 then res
        else []
      | _ -> assert false

    let case_split env uf =
      Options.tool_req 4 "TR-Arith-CaseSplit";
      let cs1, sz1 = case_split_polynomes env in
      let cs2, sz2 =  case_split_monomes env in
      match check_size env cs1, check_size env cs2 with
      | [], cs | cs, []  -> cs
      | cs1, cs2 -> if Q.compare sz1 sz2 < 0 then cs1 else cs2

    let deduce_from_real_of_int ?(add=true) env uf roi a =
      let ra, ex =
        try Uf.find uf a (fun () -> assert false)
        with Not_found -> assert false in
      match P.is_const (poly_of ra) with
      | None ->
        let mk_a = try Uf.make uf a with Not_found -> assert false in
        let env = if add then {env with roi=MX.add mk_a roi env.roi} else env in
        env, []
      | Some c ->
        let i = T.real (Q.to_string c) in
        let eq = L.LT.mk_eq roi i in
        if debug_fm() then
          fprintf fmt "[fm.add 1] deduce %a@." L.LT.print eq;
        env, [LTerm eq, ex, Sig.Other]



    let deduce_from_roi_arith_op env eqs uf t a =
      (* should be done modulo eq ??? *)
      (* LIMITATION. Not working on make (a) will not allow to
         simplify (x+y) / 1 to (x+y), and the deduction is not
         performed
      *)
      match T.view a with
      | {T.f= Sy.Op (Sy.Plus | Sy.Minus | Sy.Mult) as op; xs=[x;y]} ->
        let roi_x = Term.make (Sy.Op Sy.Real_of_int) [x] Ty.Treal in
        let roi_y = Term.make (Sy.Op Sy.Real_of_int) [y] Ty.Treal in
        let eq = L.LT.mk_eq t (Term.make op [roi_x; roi_y] Ty.Treal) in
        if debug_fm() then
          fprintf fmt "[fm.add 2] deduce %a@." L.LT.print eq;
        env, (LTerm eq, Explanation.empty, Sig.Other) :: eqs
      | _ -> env, eqs

    let add env uf t rt =
      if MT.mem t env.all_terms then env, []
      else
        let env = {env with all_terms = MT.add t rt env.all_terms} in
        match T.view t with
          {T.f=Sy.Op Sy.Real_of_int; xs=[a]; ty=Ty.Treal} ->
            assert (T.type_info a == Ty.Tint);
            let env, eqs = deduce_from_real_of_int env uf t a in
            deduce_from_roi_arith_op env eqs uf t a
        | _ -> env, []

    let extract_improved env =
      SP.fold
        (fun p acc ->
	  MP.add p (MP.find p env.polynomes) acc)
        env.improved MP.empty


    let print_model fmt env rs = match rs with
      | [] -> ()
      | _ ->
	fprintf fmt "Relation:";
        List.iter (fun (t, r) ->
          let p = poly_of r in
          let ty = P.type_info p in
          if ty == Ty.Tint || ty == Ty.Treal then
	    let p', c, d = P.normal_form_pos p in
	    let pu' =
	      try MP.find p' env.polynomes
	      with Not_found -> I.undefined ty
	    in
	    let pm' =
	      try intervals_from_monomes ~monomes_inited:false env p'
	      with Not_found -> I.undefined ty
	    in
	    let u' = I.intersect pu' pm' in
	    if I.is_point u' == None && I.is_undefined u' then
	      let u =
	        I.scale d
	          (I.add u'
		     (I.point c ty Explanation.empty)) in
	      fprintf fmt "\n %a ∈ %a" Term.print t I.pretty_print u
        ) rs;
        fprintf fmt "\n@."

    let new_terms env = Term.Set.empty

      (*- -*)
    let assert_pos_poly p =
      let _, c, d = P.normal_form_pos p in
      if Q.equal c Q.zero && Q.equal d Q.one then Some p
      else None

    let get_bounds env =
      MX.fold (fun x (ix,_) acc ->
        let px = P.create [Q.one, x] Q.zero (X.type_info x) in
        let px = match assert_pos_poly px with
          | Some p -> p | None -> assert false in
        let ix =
	  try
            let ip = MP.find px acc in
	    Intervals.intersect ix ip
	  with Not_found -> ix
        in
        MP.add px ix acc
      ) env.monomes env.polynomes

  let get_bound p env =
    try Some (MP.find p env.polynomes)
    with Not_found ->
      try Some (fst (MX.find (alien_of p) env.monomes))
      with Not_found -> None

    (*- -*)
    let new_improved env =
      SP.elements env.improved

  end
