(******************************************************************************)
(*     Alt-Ergo: The SMT Solver For Software Verification                     *)
(*     Copyright (C) 2013-2015 --- OCamlPro                                   *)
(*     This file is distributed under the terms of the CeCILL-C licence       *)
(******************************************************************************)

(******************************************************************************)
(*     The Alt-Ergo theorem prover                                            *)
(*     Copyright (C) 2006-2013                                                *)
(*     CNRS - INRIA - Universite Paris Sud                                    *)
(*                                                                            *)
(*     Sylvain Conchon                                                        *)
(*     Evelyne Contejean                                                      *)
(*                                                                            *)
(*     Francois Bobot                                                         *)
(*     Mohamed Iguernelala                                                    *)
(*     Stephane Lescuyer                                                      *)
(*     Alain Mebsout                                                          *)
(*                                                                            *)
(*   This file is distributed under the terms of the CeCILL-C licence         *)
(******************************************************************************)

open Format
open Options

module Z = Numbers.Z
module Q = Numbers.Q

module Ex = Explanation

type borne =
  | Strict of (Q.t * Ex.t)
  | Large of (Q.t * Ex.t)
  | Pinfty | Minfty

type t = {
  ints : (borne * borne) list;
  is_int : bool;
  expl: Ex.t
}

exception EmptyInterval of Ex.t
exception NotConsistent of Ex.t
exception No_finite_bound

(*BISECT-IGNORE-BEGIN*)
module Debug = struct
  let print_borne pretty fmt = function
    | Minfty -> fprintf fmt "%s" (if pretty then "-∞" else "-inf")
    | Pinfty -> fprintf fmt "%s" (if pretty then "+∞" else "+inf")
    | Strict (v, e) | Large (v, e) ->
      if verbose () || proof () then
        fprintf fmt "%s %a" (Q.to_string v) Ex.print e
      else fprintf fmt "%s" (Q.to_string v)

  let print_interval pretty fmt (b1,b2) =
    let c1, c2 = match b1, b2 with
      | Large _, Large _ -> '[', ']'
      | Large _, _ -> '[', '['
      | _, Large _ -> ']', ']'
      | _, _ -> ']', '['
    in
    fprintf fmt "%c%a;%a%c"
      c1 (print_borne pretty) b1 (print_borne pretty) b2 c2

  let rec print_list pretty fmt = function
    | [] -> ()
    | [t] -> print_interval pretty fmt t
    | t::l -> fprintf fmt "%a%s%a"
      (print_interval pretty) t
      (if pretty then " ∪ " else " ") (print_list pretty) l

  let pretty_print fmt {ints = ints; is_int = b; expl = e } =
    print_list true fmt ints;
    if verbose () || proof () then fprintf fmt " %a" Ex.print e

  let print fmt {ints = ints; is_int = b; expl = e } =
    print_list false fmt ints;
    if verbose () || proof () then fprintf fmt " %a" Ex.print e

end
(*BISECT-IGNORE-END*)

let print = Debug.print
let pretty_print = Debug.pretty_print


let large_borne_of  n e = Large  (n, e)
let strict_borne_of n e = Strict (n, e)

let undefined_int  = {ints = [Minfty, Pinfty]; is_int = true ; expl = Ex.empty}
let undefined_real = {ints = [Minfty, Pinfty]; is_int = false; expl = Ex.empty}

let undefined ty = match ty with
  | Ty.Tint  -> undefined_int
  | Ty.Treal -> undefined_real
  | _ -> assert false

let is_undefined t = match t.ints with
  | [Minfty, Pinfty] -> true
  | _ -> false

let point b ty e = {
  ints = [Large (b, e), Large (b, e)];
  is_int = ty == Ty.Tint;
  expl = Ex.empty
}

let is_point { ints = l; expl = e } =
  match l with
    | [Large (v1, e1) , Large (v2, e2)] when Q.equal v1 v2 ->
      Some (v1, Ex.union e (Ex.union e1 e2))
    | _ -> None

let finite_size {ints = l; is_int = is_int} =
  if not is_int then None
  else
    try
      let acc = ref [] in
      List.iter
        (fun (b1, b2) ->
          match b1, b2 with
          | Minfty, _ | _, Pinfty -> raise Exit
          | Large (v1, _) , Large (v2, _) -> acc := (v1, v2) :: !acc
          | _, _ -> assert false
        )l;
      let res =
        List.fold_left
          (fun n (v1, v2) -> Q.add n (Q.add (Q.sub v2 v1) Q.one)) Q.zero !acc
      in
      Some res
    with Exit -> None

let borne_inf = function
  | {ints = (Large (v, ex), _)::_} -> v, ex, true
  | {ints = (Strict (v, ex), _)::_} -> v, ex, false
  | _ -> raise No_finite_bound

let borne_sup {ints=ints} =
  match List.rev ints with
  | (_, Large (v, ex))::_ -> v, ex, true
  | (_, Strict (v, ex))::_ -> v, ex, false
  | _ -> raise No_finite_bound

let explain_borne = function
  | Large (_, e) | Strict (_, e) -> e
  | _ -> Ex.empty

let add_expl_to_borne b e =
  if Ex.is_empty e then b
  else
    match b with
    | Large (n, e') -> Large (n, Ex.union e e')
    | Strict (n, e') -> Strict (n, Ex.union e e')
    | Pinfty | Minfty -> b

let add_expl_zero i expl =
  if Ex.is_empty expl then i
  else
    let res =
      List.rev_map (fun x ->
        match x with
        | Large (c1, e1), Large (c2, e2) when Q.sign c1 = 0 && Q.sign c2 = 0 ->
          Large (Q.zero, Ex.union e1 expl), Large (Q.zero, Ex.union e2 expl)
        | _ -> x
      ) i.ints
    in
    { i with ints = List.rev res }

let int_of_borne_inf b =
  match b with
    | Minfty | Pinfty -> b
    | Large (v, e) -> Large (Q.ceiling v, e)
    | Strict (v, e) ->
      let v' = Q.ceiling v in
      if Q.compare v' v > 0 then Large (v', e) else Large (Q.add v Q.one, e)

let int_of_borne_sup b =
  match b with
    | Minfty | Pinfty -> b
    | Large (v, e) -> Large (Q.floor v, e)
    | Strict (v, e) ->
      let v' = Q.floor v in
      if Q.compare v' v < 0 then Large (v', e) else Large (Q.sub v Q.one, e)

let int_bornes (l, u) = int_of_borne_inf l, int_of_borne_sup u

let compare_bounds b1 ~is_low1 b2 ~is_low2 =
  match b1, b2 with
  | Minfty, Minfty | Pinfty, Pinfty -> 0
  | Minfty, _ | _, Pinfty -> -1
  | _, Minfty | Pinfty, _ -> 1

  | Large (v1, ex1), Large (v2, ex2) -> Q.compare v1 v2

  | Strict (v1, ex1), Strict (v2, ex2) ->
    let c = Q.compare v1 v2 in
    if c <> 0 then c
    else if is_low1 == is_low2 then 0  (* bl_bl or bu_bu *)
    else if is_low1 then 1  (* implies not is_low2 *)
    else -1  (* implies not is_low1 and is_low2 *)

  | Strict (v1, ex1), Large (v2, ex2) ->
    let c = Q.compare v1 v2 in
    if c <> 0 then c else if is_low1 then 1 else -1

  | Large (v1, ex1), Strict (v2, ex2) ->
    let c = Q.compare v1 v2 in
    if c <> 0 then c else if is_low2 then -1 else 1

let zero_endpoint b =
  match b with
  | Large (v, _) -> Q.is_zero v
  | _ -> false

let min_of_lower_bounds b1 b2 =
  if compare_bounds b1 ~is_low1:true b2 ~is_low2:true <= 0 then b1 else b2

let max_of_upper_bounds b1 b2 =
  if compare_bounds b1 ~is_low1:false b2 ~is_low2:false >= 0 then b1 else b2

let pos_borne_strict b = match b with
  | Pinfty -> true
  | Minfty -> false
  | Strict (v, _) -> Q.sign v >= 0
  | Large  (v, _) -> Q.sign v > 0

let pos_borne b = match b with
  | Pinfty -> true
  | Minfty -> false
  | Strict (v, _) | Large (v, _) -> Q.sign v >= 0

let neg_borne_strict b = match b with
  | Pinfty -> false
  | Minfty -> true
  | Strict (v, _) -> Q.sign v <= 0
  | Large  (v, _) -> Q.sign v < 0

let neg_borne b = match b with
  | Pinfty -> false
  | Minfty -> true
  | Strict (v, _) | Large (v, _) -> Q.sign v <= 0

let intersect =
  let rec step is_int l1 l2 acc expl =
    match l1, l2 with
    | [], _ | _, [] ->  List.rev acc, expl
    | (lo1, up1)::r1, (lo2, up2)::r2 ->
      let cul = compare_bounds up1 ~is_low1:false lo2 ~is_low2:true in
      if cul < 0 then
	let nexpl  = Ex.union (explain_borne up1) (explain_borne lo2) in
	match r1 with
	| [] -> List.rev acc, Ex.union nexpl expl
	| (lor1,upr1)::rr1 ->
	  let lor1 = add_expl_to_borne lor1 nexpl in
	  step is_int ((lor1,upr1)::rr1) l2 acc expl
      else
        let clu = compare_bounds lo1 ~is_low1:true up2 ~is_low2:false in
        if clu > 0 then
	  let nexpl  = Ex.union (explain_borne up2) (explain_borne lo1) in
	  match r2 with
	  | [] -> List.rev acc, Ex.union nexpl expl
	  | (lor2,upr2)::rr2 ->
	    let lor2 = add_expl_to_borne lor2 nexpl in
	    step is_int l1 ((lor2,upr2)::rr2) acc expl
        else
          let cll = compare_bounds lo1 ~is_low1:true lo2 ~is_low2:true in
          let cuu = compare_bounds up1 ~is_low1:false up2 ~is_low2:false in
          if cll <= 0 && cuu >= 0 then
	    step is_int l1 r2 ((lo2,up2)::acc) expl
          else if cll >= 0 && cuu <= 0 then
	    step is_int r1 l2 ((lo1,up1)::acc) expl
          else if cll <= 0 && cuu <= 0 && cul >= 0 then
	    step is_int r1 l2 ((lo2,up1)::acc) expl
          else if cll >= 0 && cuu >= 0 && clu <= 0 then
	    step is_int l1 r2 ((lo1,up2)::acc) expl
          else assert false
  in
  fun ({ints=l1; expl=e1; is_int=is_int} as uints1) {ints=l2; expl=e2} ->
    let l1, l2 =
      if is_int then
        List.rev (List.rev_map int_bornes l1),
        List.rev (List.rev_map int_bornes l2)
      else l1, l2
    in
    let l,expl = step is_int l1 l2 [] (Ex.union e1 e2) in
    match l with
    | [] -> raise (NotConsistent expl);
    | _  -> { uints1 with ints = l; expl = expl }

let new_borne_sup expl b ~is_le uints =
  let aux = if is_le then large_borne_of else strict_borne_of in
  intersect
    { ints = [Minfty, aux b expl];
      is_int = uints.is_int;
      expl = Ex.empty } uints

let new_borne_inf expl b ~is_le uints =
  let aux = if is_le then large_borne_of else strict_borne_of in
  intersect
    { ints = [aux b expl, Pinfty];
      is_int = uints.is_int;
      expl = Ex.empty } uints

let exclude =
  let rec complement l prev acc =
    match l with
    | (b1,b2)::r ->
      let bu = match b1 with
	| Strict v -> Large v
	| Large v -> Strict v
	| _ -> b1
      in
      let bl = match b2 with
	| Strict v -> Large v
	| Large v -> Strict v
	| _ -> b2
      in
      if bu == Minfty then complement r bl acc
      else complement r bl ((prev, bu)::acc)
    | [] ->
      List.rev (if prev == Pinfty then acc else (prev, Pinfty)::acc)
  in
  fun uints1 uints2 ->
    let uints1_c = {uints1 with ints = complement uints1.ints Minfty []} in
    intersect uints1_c uints2

type interval_class =
  | P of Ex.t
  | M of Ex.t
  | N of Ex.t
  | Z

let class_of l u =
  if zero_endpoint l && zero_endpoint u then Z
  else if pos_borne l && pos_borne u then begin
    assert (pos_borne_strict u);
    P (explain_borne l)
  end
  else if neg_borne l && neg_borne u then begin
    assert (neg_borne_strict l);
    N (explain_borne u)
  end
  else begin
    assert (neg_borne_strict l);
    assert (pos_borne_strict u);
    M (Ex.union (explain_borne l) (explain_borne u))
  end

let rec union_bornes l =
  match l with
  | [] | [_] -> l
  | (l1, u1)::((l2, u2)::r as r2) ->
    if compare_bounds u1 ~is_low1:false l2 ~is_low2:true < 0 then
      (l1, u1)::(union_bornes r2)
    else if
        compare_bounds u1 ~is_low1:false u2 ~is_low2:false > 0 then
      union_bornes ((l1, u1)::r)
    else
      union_bornes ((l1, u2)::r)

let union_intervals uints =
    let l = List.fast_sort (fun (l1, _) (l2, _) ->
      compare_bounds l1 ~is_low1:true l2 ~is_low2:true) uints.ints in
    {uints with ints = union_bornes l}

let minus_borne = function
  | Minfty -> Pinfty
  | Pinfty -> Minfty
  | Large (v, e) -> Large (Q.minus v, e)
  | Strict (v, e) -> Strict (Q.minus v, e)

let rev_normalize_int_bounds rl ex =
  (* XXX: can we do better than modifying the global explanation ? *)
  let l, ex =
    List.fold_left
      (fun (acc, ex) b ->
        let b = int_bornes b in
        match b with
        | Large (v, ex1), Large (w, ex2) when Q.compare w v < 0 ->
          acc, Ex.union (Ex.union ex1 ex2) ex
        | Strict (v, ex1), Large (w, ex2)
        | Large (v, ex1) , Strict (w, ex2)
        | Strict (v, ex1), Strict (w, ex2) when Q.compare w v <= 0 ->
          acc, Ex.union (Ex.union ex1 ex2) ex
        | _ -> b :: acc, ex
      )([], ex) rl
  in
  match l with
  | [] -> raise (NotConsistent ex)
  | _ -> l, ex

let scale_interval_zero n (b1, b2) =
  assert (Q.sign n = 0);
  Large (Q.zero, explain_borne b1), Large (Q.zero, explain_borne b2)

let scale_borne_non_zero n b =
  assert (Q.sign n > 0);
  match b with
  | Pinfty | Minfty -> b
  | Large (v, e) -> Large (Q.mult n v, e)
  | Strict (v, e) -> Strict (Q.mult n v, e)

let scale_interval_pos n (b1, b2) =
  scale_borne_non_zero n b1, scale_borne_non_zero n b2

let scale_interval_neg n (b1, b2) =
  minus_borne (scale_borne_non_zero (Q.minus n) b2),
  minus_borne (scale_borne_non_zero (Q.minus n) b1)

let scale n uints =
  Options.tool_req 4 "TR-Arith-Axiomes scale";
  if Q.equal n Q.one then uints
  else
    let sgn = Q.sign n in
    let aux =
      if sgn = 0 then scale_interval_zero
      else if sgn > 0 then scale_interval_pos
      else scale_interval_neg
    in
    let rl = List.rev_map (aux n) uints.ints in
    let l, ex =
      if uints.is_int then rev_normalize_int_bounds rl uints.expl
      else List.rev rl, uints.expl
    in
    union_intervals { uints with ints = l; expl = ex }

let add_borne b1 b2 =
  match b1,b2 with
    | Minfty, Pinfty | Pinfty, Minfty -> assert false
    | Minfty, _ | _, Minfty -> Minfty
    | Pinfty, _ | _, Pinfty -> Pinfty
    | Large (v1, e1), Large (v2, e2) ->
      Large (Q.add v1 v2, Ex.union e1 e2)
    | (Large (v1, e1) | Strict (v1, e1)), (Large (v2, e2) | Strict (v2, e2)) ->
      Strict (Q.add v1 v2, Ex.union e1 e2)

let add_interval l (b1,b2) =
  List.fold_right
    (fun (b1', b2') l ->
      let l1 = ((add_borne b1 b1'),(add_borne b2 b2'))::l in
      union_bornes (l1)
    ) l []

let add {ints = l1; is_int = is_int; expl = e1} {ints = l2; expl = e2}=
  let l =
    List.fold_left
      (fun l bs -> let i = add_interval l1 bs in i@l) [] l2
  in
  union_intervals { ints = l ; is_int = is_int; expl = Ex.union e1 e2 }

let sub i1 i2 = add i1 (scale Q.m_one i2)

let merge i1 i2 =
  union_intervals
    {ints = List.rev_append i1.ints i2.ints; is_int = i1.is_int;
     expl = Explanation.union i1.expl i2.expl}

exception Found of Sig.answer

let doesnt_contain_0 {ints=l} =
  try
    let max = List.fold_left
      (fun old_u (l, u) ->
	if neg_borne l && pos_borne u then raise (Found Sig.No);
	if neg_borne_strict old_u && pos_borne_strict l then
	  raise (Found
		   (Sig.Yes
		      (Ex.union
			 (explain_borne old_u) (explain_borne l), [])));
	u
      )Minfty l
    in
    if neg_borne_strict max then Sig.Yes (explain_borne max, [])
    else Sig.No
  with Found ans -> ans

let is_positive {ints=ints; expl=expl} =
  match ints with
  | [] -> assert false
  | (lb,_)::_ -> if pos_borne lb then Sig.Yes (expl, []) else Sig.No

let root_num a n =
  let sgn = Q.sign a in
  assert (sgn >= 0);
  if sgn = 0 then Q.zero
  else
    let v = Q.to_float a in
    let w =
      if Pervasives.(<) v min_float then min_float
      else if Pervasives.(>) v max_float then max_float
      else v
    in
    Q.from_float (if n = 2 then sqrt w else w ** (1. /. float n))

let root_default_num a n =
  let s = root_num a n in
  let d = Q.sub a (Q.power s n) in
  if Q.sign d >= 0 then s else Q.div a (Q.power s (n - 1))

let root_exces_num a n =
  let s = root_num a n in
  let d = Q.sub a (Q.power s n) in
  if Q.sign d <= 0 then s else Q.div a (Q.power s (n - 1))



let compare_bornes b1 b2 =
  match b1, b2 with
    | Minfty, Minfty | Pinfty, Pinfty -> 0
    | Minfty, _ | _, Pinfty -> -1
    | Pinfty, _ | _, Minfty -> 1
    | Strict (v1, _), Strict (v2, _) | Large (v1, _), Large (v2, _)
    | Strict (v1, _), Large (v2, _) | Large (v1, _), Strict (v2, _) ->
      Q.compare v1 v2

let is_strict_smaller {ints=i1} {ints=i2} =
  match i1, i2 with
    | _, [] -> false
    | [], _ -> true
    | _ ->
      try
	List.iter2 (fun (l1, u1) (l2, u2) ->
	  if compare_bounds l1 ~is_low1:true l2 ~is_low2:true > 0
            || compare_bounds u1 ~is_low1:false u2 ~is_low2:false < 0
	  then raise Exit
	) i1 i2;
	false
      with
	| Exit -> true
	| Invalid_argument _ -> true

let mult_borne b1 b2 =
  match b1,b2 with
    | Minfty, Pinfty | Pinfty, Minfty -> assert false
    | Minfty, b | b, Minfty ->
      if compare_bornes b (large_borne_of Q.zero Ex.empty) = 0
      then b
      else if pos_borne b then Minfty
      else Pinfty
    | Pinfty, b | b, Pinfty ->
      if compare_bornes b (large_borne_of Q.zero Ex.empty) = 0
      then b
      else if pos_borne b then Pinfty
      else Minfty
    | Strict (_, e1), Large (v, e2)
    | Large (v, e1), Strict (_, e2) when Q.is_zero v ->
      Large (Q.zero, Ex.union e1 e2)

    | Strict (v1, e1), Strict (v2, e2) | Strict (v1, e1), Large (v2, e2)
    | Large (v1, e1), Strict (v2, e2) ->
      Strict (Q.mult v1 v2, Ex.union e1 e2)
    | Large (v1, e1), Large (v2, e2) ->
      Large (Q.mult v1 v2, Ex.union e1 e2)

let mult_borne_inf b1 b2 =
  match b1,b2 with
    | Minfty, Pinfty | Pinfty, Minfty -> Minfty
    | _, _ -> mult_borne b1 b2

let mult_borne_sup b1 b2 =
  match b1,b2 with
    | Minfty, Pinfty | Pinfty, Minfty -> Pinfty
    | _, _ -> mult_borne b1 b2


let mult_bornes (a,b) (c,d) =
  (* see ../extra/intervals_mult.png *)
  match class_of a b, class_of c d with
    | P e1, P e2 ->
      mult_borne_inf a c, mult_borne_sup b d, Ex.union e1 e2
    | P e1, M e2 ->
      mult_borne_inf b c, mult_borne_sup b d, Ex.union e1 e2
    | P e1, N e2 ->
      mult_borne_inf b c, mult_borne_sup a d, Ex.union e1 e2
    | M e1, P e2 ->
      mult_borne_inf a d, mult_borne_sup b d, Ex.union e1 e2
    | M e1, M e2 ->
      min_of_lower_bounds (mult_borne_inf a d) (mult_borne_inf b c),
      max_of_upper_bounds (mult_borne_sup a c) (mult_borne_sup b d),
      Ex.union e1 e2
    | M e1, N e2 ->
      mult_borne_inf b c, mult_borne_sup a c, Ex.union e1 e2
    | N e1, P e2 ->
      mult_borne_inf a d, mult_borne_sup b c, Ex.union e1 e2
    | N e1, M e2 ->
      mult_borne_inf a d, mult_borne_sup a c, Ex.union e1 e2
    | N e1, N e2 ->
      mult_borne_inf b d, mult_borne_sup a c, Ex.union e1 e2
    | Z, (P _ | M _ | N _ | Z) -> (a, b, Ex.empty)
    | (P _ | M _ | N _ ), Z -> (c, d, Ex.empty)

let rec power_borne_inf p b =
  match p with
    | 1 -> b
    | p -> mult_borne_inf b (power_borne_inf (p-1) b)

let rec power_borne_sup p b =
  match p with
    | 1 -> b
    | p -> mult_borne_sup b (power_borne_sup (p-1) b)

let max_merge b1 b2 =
  let ex = Ex.union (explain_borne b1) (explain_borne b2) in
  let max = max_of_upper_bounds b1 b2 in
  match max with
    | Minfty | Pinfty -> max
    | Large (v, _) -> Large (v, ex)
    | Strict (v, _) -> Strict (v, ex)

let power_bornes p (b1,b2) =
  if neg_borne b1 && pos_borne b2 then
    match p with
      | 0 -> assert false
      | p when p mod 2 = 0 ->
	(* max_merge to have explanations !!! *)
	let m = max_merge (power_borne_sup p b1) (power_borne_sup p b2) in
	(Large (Q.zero, Ex.empty), m)
      | _ -> (power_borne_inf p b1, power_borne_sup p b2)
  else if pos_borne b1 && pos_borne b2 then
    (power_borne_inf p b1, power_borne_sup p b2)
  else if neg_borne b1 && neg_borne b2 then
    match p with
      | 0 -> assert false
      | p when p mod 2 = 0 -> (power_borne_inf p b2, power_borne_sup p b1)
      | _ -> (power_borne_inf p b1, power_borne_sup p b2)
  else assert false

let int_div_of_borne_inf b =
  match b with
    | Minfty | Pinfty -> b
    | Large (v, e) -> Large (Q.floor v, e)
    | Strict (v, e) ->
      let v' = Q.floor v in
      if Q.compare v' v > 0 then Large (v', e) else Large (Q.add v Q.one, e)

let int_div_of_borne_sup b =
  match b with
    | Minfty | Pinfty -> b
    | Large (v, e) -> Large (Q.floor v, e)
    | Strict (v, e) ->
      let v' = Q.floor v in
      if Q.compare v' v < 0 then Large (v', e) else Large (Q.sub v Q.one, e)

let int_div_bornes l u =
  int_div_of_borne_inf l, int_div_of_borne_sup u

let mult u1 u2 =
  Options.tool_req 4 "TR-Arith-Axiomes mult";
  let resl, expl =
    List.fold_left
      (fun (l', expl) b1 ->
	List.fold_left
	  (fun (l, ex) b2 ->
	    let bl, bu, ex' = mult_bornes b1 b2 in
	    (bl, bu)::l, Ex.union ex ex') (l', expl) u2.ints)
      ([], Ex.empty) u1.ints
  in
  union_intervals { ints= resl; is_int = u1.is_int;
	  expl = Ex.union expl
      (Ex.union u1.expl u2.expl) }

let power n u =
  Options.tool_req 4 "TR-Arith-Axiomes power";
  let l = List.map (power_bornes n) u.ints in
  union_intervals { u with ints = l }

let root_default_borne is_int x n =
  match x with
    | Pinfty -> Pinfty
    | Minfty -> Minfty
    | Large (v, e) | Strict (v, e) ->
      let s = if Q.sign v >= 0 then root_default_num v n
	else (Q.minus (root_exces_num (Q.minus v) n)) in
      if is_int then
	let cs = Q.ceiling s in
	let cs2 = Q.power cs n in
	if Q.compare v cs2 <= 0 then Large (cs, e)
	else Large (Q.add cs Q.one, e)
      else Large (s, e)

let root_exces_borne is_int x n =
  match x with
    | Pinfty -> Pinfty
    | Minfty -> Minfty
    | Large (v, e) | Strict (v, e) ->
      let s = if Q.sign v >= 0 then root_exces_num v n
	else (Q.minus (root_default_num (Q.minus v) n)) in
      if is_int then
	let cs = Q.floor s in
	let cs2 = Q.power cs n in
	if Q.compare v cs2 >= 0 then Large (cs, e)
	else Large (Q.sub cs Q.one, e)
      else Large (s, e)

let sqrt_interval is_int (b1,b2) =
  let l1 = minus_borne (root_exces_borne is_int b2 2) in
  let u1 = minus_borne (root_default_borne is_int b1 2) in
  let l2 = root_default_borne is_int b1 2 in
  let u2 = root_exces_borne is_int b2 2 in
  let c1 = compare_bornes l1 u1 in
  let c2 = compare_bornes l2 u2 in
  if c1 > 0 then if c2 > 0 then [] else [l2,u2]
  else if c2 > 0 then [l1, u1] else union_bornes [(l1,u1); (l2, u2)]

let root_interval is_int (b1,b2) n =
  let u,l = (root_default_borne is_int b1 n, root_exces_borne is_int b2 n) in
  if compare_bornes u l > 0 then [] else [u,l]

let sqrt {ints = l; is_int = is_int; expl = e } =
  Options.tool_req 4 "TR-Arith-Axiomes sqrt";
  let l =
    List.fold_left
      (fun l' bs ->
	(sqrt_interval is_int bs)@l'
      ) [] l in
  union_intervals { ints = l; is_int = is_int; expl = e }

let rec root n ({ints = l; is_int = is_int; expl = e} as u) =
  Options.tool_req 4"TR-Arith-Axiomes root";
  if n mod 2 = 0 then root (n/2) (sqrt u)
  else
    let l =
      List.fold_left
	(fun l' bs ->
	  (root_interval is_int bs n)@l'
	) [] l
    in
    union_intervals { ints = l; is_int = is_int; expl = e }

let inv_borne_inf b is_int ~other =
  match b with
    | Pinfty -> assert false
    | Minfty ->
      if is_int then Large (Q.zero,  explain_borne other)
      else Strict (Q.zero, explain_borne other)
    | Strict (c, e) | Large (c, e) when Q.sign c = 0 -> Pinfty
    | Strict (v, e) -> Strict (Q.div Q.one v, e)
    | Large (v, e) -> Large (Q.div Q.one v, e)

let inv_borne_sup b is_int ~other =
  match b with
    | Minfty -> assert false
    | Pinfty ->
      if is_int then Large (Q.zero, explain_borne other)
      else Strict (Q.zero, explain_borne other)
    | Strict (c, e) | Large (c, e) when Q.sign c = 0 -> Minfty
    | Strict (v, e) -> Strict (Q.div Q.one v, e)
    | Large (v, e) -> Large (Q.div Q.one v, e)

let inv_bornes (l, u) is_int =
  inv_borne_sup u is_int ~other:l, inv_borne_inf l is_int ~other:u


let inv ({ints=l; is_int=is_int} as u) =
  try
    let l' = List.fold_left
      (fun acc (l,u) ->
	if (pos_borne_strict l && pos_borne_strict u)
	  || (neg_borne_strict l && neg_borne_strict u) then
	  (inv_bornes (l, u) is_int) :: acc
	else raise Exit
      ) [] l in
    union_intervals { u with ints = l' }
  with Exit -> { u with ints = [Minfty, Pinfty]  }

let div i1 i2 =
  Options.tool_req 4 "TR-Arith-Axiomes div";
  let inv_i2 = inv i2 in
  if is_undefined inv_i2 then inv_i2
  else
    let i1 = match doesnt_contain_0 i2 with
      | Sig.Yes (ex, _) -> add_expl_zero i1 ex
      | Sig.No -> i1
    in
    let ({ints=l; is_int=is_int} as i) = mult i1 inv_i2 in
    let l =
      if is_int then
	List.map (fun (l,u) -> int_div_bornes l u) l
      else l in
    { i with ints = l }

let abs =
  let zero_inf_r = new_borne_inf Ex.empty Q.zero true (undefined Ty.Treal) in
  let zero_inf_i = new_borne_inf Ex.empty Q.zero true (undefined Ty.Tint) in
  fun i ->
    let xx = if i.is_int then zero_inf_i else zero_inf_r in
    intersect (merge i (scale Q.m_one i)) xx

let mk_closed l u llarge ularge lexp uexp ty =
  let lb = if llarge then Large(l, lexp) else Strict (l, lexp) in
  let ub = if ularge then Large(u, uexp) else Strict (u, uexp) in
  { ints = [lb, ub]; is_int = ty == Ty.Tint; expl = Ex.union lexp uexp }
