(******************************************************************************)
(*     Alt-Ergo: The SMT Solver For Software Verification                     *)
(*     Copyright (C) 2013-2015 --- OCamlPro                                   *)
(*     This file is distributed under the terms of the CeCILL-C licence       *)
(******************************************************************************)

(******************************************************************************)
(*     The Alt-Ergo theorem prover                                            *)
(*     Copyright (C) 2006-2013                                                *)
(*     CNRS - INRIA - Universite Paris Sud                                    *)
(*                                                                            *)
(*     Sylvain Conchon                                                        *)
(*     Evelyne Contejean                                                      *)
(*                                                                            *)
(*     Francois Bobot                                                         *)
(*     Mohamed Iguernelala                                                    *)
(*     Stephane Lescuyer                                                      *)
(*     Alain Mebsout                                                          *)
(*                                                                            *)
(*   This file is distributed under the terms of the CeCILL-C licence         *)
(******************************************************************************)

open Sig

(*** Combination module of Relations ***)

module CX = Combine.Shostak

module Rel1 : RSig.RELATION
  with type r = CX.r and type uf = Combine.Uf.t =
         RArith.Relation.Make (CX)(Combine.Uf)
    (struct
      include Combine.TX1
      let extract = CX.extract1
      let embed =  CX.embed1
     end)

module Rel2 : RSig.RELATION
  with type r = CX.r and type uf = Combine.Uf.t =
         RRecords.Relation
           (struct
             include CX
             let extract = extract2
             let embed = embed2
            end)(Combine.Uf)


module Rel3 : RSig.RELATION
  with type r = CX.r and type uf = Combine.Uf.t =
         RBitv.Relation
           (struct
             include CX
             let extract = extract3
             let embed = embed3
            end)(Combine.Uf)


module Rel4 : RSig.RELATION
  with type r = CX.r and type uf = Combine.Uf.t =
         RArrays.Relation
           (struct
             include CX
             let extract = extract4
             let embed = embed4
            end)(Combine.Uf)


module Rel5 : RSig.RELATION
  with type r = CX.r and type uf = Combine.Uf.t =
         RSum.Relation
           (struct
             include CX
             let extract = extract5
             let embed = embed5
            end)(Combine.Uf)

module Relation : RSig.RELATION with type r = CX.r and type uf = Combine.Uf.t =
struct
  type r = CX.r
  type uf = Combine.Uf.t

  type t = {
    r1: Rel1.t;
    r2: Rel2.t;
    r3: Rel3.t;
    r4: Rel4.t;
    r5: Rel5.t;
  }

  let empty classes = {
    r1=Rel1.empty classes;
    r2=Rel2.empty classes;
    r3=Rel3.empty classes;
    r4=Rel4.empty classes;
    r5=Rel5.empty classes;
  }

  let (|@|) l1 l2 =
    if l1 == [] then l2
    else if l2 == [] then l1
    else List.rev_append l1 l2


  let ($) s1 s2 = Term.Set.union s1 s2

  let assume env uf sa =
    Options.exec_thread_yield ();
    let env1, { assume = a1; remove = rm1}, nterms1 =
      Rel1.assume env.r1 uf sa in
    let env2, { assume = a2; remove = rm2}, nterms2 =
      Rel2.assume env.r2 uf sa in
    let env3, { assume = a3; remove = rm3}, nterms3 =
      Rel3.assume env.r3 uf sa in
    let env4, { assume = a4; remove = rm4}, nterms4 =
      Rel4.assume env.r4 uf sa in
    let env5, { assume = a5; remove = rm5}, nterms5 =
      Rel5.assume env.r5 uf sa in
    {r1=env1; r2=env2; r3=env3; r4=env4; r5=env5},
    { assume = a1 |@| a2 |@| a3 |@| a4 |@| a5;
      remove = rm1 |@| rm2 |@| rm3 |@| rm4 |@| rm5;},
    nterms1 $ nterms2 $ nterms3 $ nterms4 $ nterms5

  let query env uf a =
    Options.exec_thread_yield ();
    match Rel1.query env.r1 uf a with
      | Yes _ as ans -> ans
      | No ->
	match Rel2.query env.r2 uf a with
	  | Yes _ as ans -> ans
	  | No ->
	    match
              Rel3.query env.r3 uf a with
		| Yes _ as ans -> ans
		| No ->
		  match
                    Rel4.query
                      env.r4 uf a with
		        | Yes _ as ans -> ans
		        | No ->
                          Rel5.query
                            env.r5 uf a

  let case_split env uf =
    Options.exec_thread_yield ();
    let seq1 = Rel1.case_split env.r1 uf in
    let seq2 = Rel2.case_split env.r2 uf in
    let seq3 = Rel3.case_split env.r3 uf in
    let seq4 = Rel4.case_split env.r4 uf in
    let seq5 = Rel5.case_split env.r5 uf in
    let l = seq1 |@| seq2 |@| seq3 |@| seq4 |@| seq5 in
    List.sort
      (fun (_,sz1) (_,sz2) ->
        match sz1, sz2 with
        | CS(_,sz1), CS(_,sz2) -> Numbers.Q.compare sz1 sz2
        | _ -> assert false
      )l


  let add env uf t r =
    Options.exec_thread_yield ();
    let r1,l1 = Rel1.add env.r1 uf t r in
    let r2,l2 = Rel2.add env.r2 uf t r in
    let r3,l3 = Rel3.add env.r3 uf t r in
    let r4,l4 = Rel4.add env.r4 uf t r in
    let r5,l5 = Rel5.add env.r5 uf t r in
    {r1=r1;r2=r2;r3=r3;r4=r4;r5=r5},
    l5 |@| l4 |@| l3 |@| l2 |@| l1

  let print_model fmt env rs =
    Rel1.print_model fmt env.r1 rs;
    Rel2.print_model fmt env.r2 rs;
    Rel3.print_model fmt env.r3 rs;
    Rel4.print_model fmt env.r4 rs;
    Rel5.print_model fmt env.r5 rs


  let new_terms env =
    let t1 = Rel1.new_terms env.r1 in
    let t2 = Rel2.new_terms env.r2 in
    let t3 = Rel3.new_terms env.r3 in
    let t4 = Rel4.new_terms env.r4 in
    let t5 = Rel5.new_terms env.r5 in
    Term.Set.union t1
      (Term.Set.union t2
         (Term.Set.union t3
            (Term.Set.union t4 t5)))

end

