(******************************************************************************)
(*     Alt-Ergo: The SMT Solver For Software Verification                     *)
(*     Copyright (C) 2013-2015 --- OCamlPro                                   *)
(*     This file is distributed under the terms of the CeCILL-C licence       *)
(******************************************************************************)

module Sy = Symbols
module Hs = Hstring
module T = Term

module FPA = struct

  let fpa_rounding_mode_syntax =
    "type fpa_rounding_mode =
  (* five standard/why3 fpa rounding modes *)
    NearestTiesToEven (*ne in Gappa: to nearest, tie breaking to even mantissas*)
  | ToZero (* zr in Gappa: toward zero *)
  | Up (* up in Gappa: toward plus infinity *)
  | Down (* dn in Gappa: toward minus infinity *)
  | NearestTiesToAway (* na : to nearest, tie breaking away from zero *)

  (* additional Gappa rounding modes *)
  | Aw (* aw in Gappa: away from zero **)
  | Od (* od in Gappa: to odd mantissas *)
  | No (* no in Gappa: to nearest, tie breaking to odd mantissas *)
  | Nz (* nz in Gappa: to nearest, tie breaking toward zero *)
  | Nd (* nd in Gappa: to nearest, tie breaking toward minus infinity *)
  | Nu (* nu in Gappa: to nearest, tie breaking toward plus infinity *)
"

  let fpa_symbols_syntax =
    "
(* the first argument is mantissas' size (including the implicit bit),
   the second one is the exp of the min representable normalized number,
   the third one is the rounding mode, and the last one is the real to
   be rounded *)
logic float: int, int, fpa_rounding_mode, real -> real

(* syntactic sugar for simple precision floats:
   mantissas size = 24, min exp = 149 *)
logic float32: fpa_rounding_mode, real -> real

(* syntactic sugar for simple precision floats with default rounding mode
(i.e. NearestTiesToEven)*)
logic float32d: real -> real

(* syntactic sugar for double precision floats:
   mantissas size = 53, min exp = 1074 *)
logic float64: fpa_rounding_mode, real -> real

(* syntactic sugar for double precision floats with default rounding mode
(i.e. NearestTiesToEven) *)
logic float64d: real -> real


(* type cast: from int to real *)
logic real_of_int : int -> real

(* abs value of a real *)
logic abs_real : real -> real

(* sqrt value of a real *)
logic sqrt_real : real -> real

(* abs value of an int *)
logic abs_int : int -> int
"

  let fpa_axioms =
"(* there are FPA axioms that are not inlined in the reasoning
   part. They are thus managed by the generic matching mechanism
   of Alt-Ergo *)
(* Remarque: should add semantic trigger 'x <= y'
   or maybe also 'float(m,p,md,x) > float(m,p,md,y)' in future
   version *)
axiom float_is_monotonic:
  forall m, p : int.
  forall md : fpa_rounding_mode.
  forall x, y : real [float(m,p,md,x), float(m,p,md,y)].
  x <= y -> float(m,p,md,x) <= float(m,p,md,y)
"

  let headers =
    Format.sprintf "%s\n%s\n%s\n"
      fpa_rounding_mode_syntax
      fpa_symbols_syntax
      fpa_axioms

  let prelude () =
    if Options.use_fpa() then begin
      if Options.debug_fpa() then
        Format.eprintf "[loading fpa prelude]\n%s\n@." headers;
      headers
    end
    else ""

  let is_rounding_mode t =
    Options.use_fpa() &&
      match (T.view t).T.ty with
      | Ty.Tsum (hs, _) -> String.compare (Hs.view hs) "fpa_rounding_mode" = 0
      | _ -> false

  let ty_mode =
    let mode_ty = Hs.make "fpa_rounding_mode" in
    let mode_constrs =
      [ (* standards *)
        Hs.make "NearestTiesToEven";
        Hs.make "NearestTiesToAway";
        Hs.make "ToZero";
        Hs.make "Up";
        Hs.make "Down";
        (* non standards *)
        Hs.make "Aw";
        Hs.make "Od";
        Hs.make "Nodd";
        Hs.make "Nz";
        Hs.make "Nd";
        Hs.make "Nu" ]
    in
    Ty.Tsum(mode_ty, mode_constrs)

  (*  why3/standard rounding modes*)

  let _NearestTiesToEven__rounding_mode =
    T.make (Sy.Name(Hs.make "NearestTiesToEven", Sy.Constructor)) [] ty_mode
  (** ne in Gappa: to nearest, tie breaking to even mantissas*)

  let _ToZero__rounding_mode =
    T.make (Sy.Name(Hs.make "ToZero", Sy.Constructor)) [] ty_mode
  (** zr in Gappa: toward zero *)

  let _Up__rounding_mode =
    T.make (Sy.Name(Hs.make "Up", Sy.Constructor)) [] ty_mode
  (** up in Gappa: toward plus infinity *)

  let _Down__rounding_mode =
    T.make (Sy.Name(Hs.make "Down", Sy.Constructor)) [] ty_mode
  (** dn in Gappa: toward minus infinity *)

  let _NearestTiesToAway__rounding_mode =
    T.make (Sy.Name(Hs.make "NearestTiesToAway", Sy.Constructor)) [] ty_mode
  (** na : to nearest, tie breaking away from zero *)

  (* additional Gappa rounding modes *)

  let _Aw__rounding_mode =
    T.make (Sy.Name(Hs.make "Aw", Sy.Constructor)) [] ty_mode
  (** aw in Gappa: away from zero **)

  let _Od__rounding_mode =
    T.make (Sy.Name(Hs.make "Od", Sy.Constructor)) [] ty_mode
  (** od in Gappa: to odd mantissas *)

  let _No__rounding_mode =
    T.make (Sy.Name(Hs.make "No", Sy.Constructor)) [] ty_mode
  (** no in Gappa: to nearest, tie breaking to odd mantissas *)

  let _Nz__rounding_mode =
    T.make (Sy.Name(Hs.make "Nz", Sy.Constructor)) [] ty_mode
  (** nz in Gappa: to nearest, tie breaking toward zero *)

  let _Nd__rounding_mode =
    T.make (Sy.Name(Hs.make "Nd", Sy.Constructor)) [] ty_mode
  (** nd in Gappa: to nearest, tie breaking toward minus infinity *)

  let _Nu__rounding_mode =
    T.make (Sy.Name(Hs.make "Nu", Sy.Constructor)) [] ty_mode
  (** nu in Gappa: to nearest, tie breaking toward plus infinity *)

end
