(**************************************************************************)
(*                                                                        *)
(*     The Alt-ergo theorem prover                                        *)
(*     Copyright (C) 2006-2010                                            *)
(*                                                                        *)
(*     Sylvain Conchon                                                    *)
(*     Evelyne Contejean                                                  *)
(*     Stephane Lescuyer                                                  *)
(*     Mohamed Iguernelala                                                *)
(*     Alain Mebsout                                                      *)
(*     Cody Roux                                                          *)
(*                                                                        *)
(*     CNRS - INRIA - Universite Paris Sud                                *)
(*                                                                        *)
(*   This file is distributed under the terms of the CeCILL-C licence     *)
(*                                                                        *)
(**************************************************************************)

open Format
open Options

module Q = Numbers.Q
module Z = Numbers.Z

exception Infinity
exception Not_implemented

let rec ( **/) q n = if n = 0 then Q.one
  else begin
    let q' = q **/ (n/2) in
    if n mod 2 = 0 then
      Q.mult q' q'
    else
      Q.mult q (Q.mult q' q')
  end

(* For some reason, Q.mul_2exp does not support negative i
   (undocumented). In addition, we work around a library bug.*)
let mul_2exp x i =
  let x_2i = if i < 0 then
      Q.div_2exp x (-i)
    else Q.mult_2exp x i
  in
  (* Necessary because of bug in Zarith *)
  Q.mult x_2i Q.one

(* Similar problem with div_2exp *)
let div_2exp x i =
  mul_2exp x (-i)



(* Integer part of binary logarithm for NON-ZERO POSITIVE number *)
let log_2 m =
  let e = ref 0 in
  let rec log_aux m =
    if Q.compare_to_0 m >= 0 then
      begin
        if Q.compare m (Q.from_int 2) < 0 then
	  !e
        else
          (e := !e + 1 ;  log_aux (div_2exp m 1) )
      end
    else
      (e := !e - 1 ; log_aux (mul_2exp m 1) )
  in
  log_aux m

(* Takes m and returns (n,k) where m = n*2^k and k is the largest such
   integer. *)
let bin_fact m =
  let rec bin_fact_aux m e =
    if Z.equal (Z.rem m (Z.from_int 2)) Z.one then (m,e)
    else bin_fact_aux (Z.div m (Z.from_int 2)) (Z.add e Z.one)
  in
  if Z.equal m Z.zero then (m,Z.zero)
  else bin_fact_aux m Z.zero


(* Takes m and returns the number of bits required to
   code the mantissa of m *)
(* Raises Infinity if the number is not diadic *)
let mantissa r =
  let num, den   = Q.num r, Q.den r in
  let num', den' = fst (bin_fact num), fst (bin_fact den) in
  let q, r = Z.div_rem num' den' in
  if Z.equal r Z.zero then log_2 (Q.from_z q) else raise Infinity

(* Takes a rationnal between 0 and 1/2 and a big int k and
   rounds k to the
   nearest big_int according to q and mode *)
let rnd_big_int (mode: Term.t) q k =
  if Term.equal mode Builtin.FPA._NearestTiesToEven__rounding_mode then
    if Q.compare q (Q.from_float 0.5) > 0 then Z.add k Z.one
    else  if Q.compare q (Q.from_float 0.5) < 0 then k
    else begin
      if Z.equal (Z.rem k (Z.from_int 2)) Z.zero then
        k
      else
        Z.add k Z.one
    end
    else
      let () = fprintf fmt "%a@." Term.print mode in
      assert false

(* Given a precision, a minimal exponant, a rounding mode and a rationnal,
   returns the pair mantissa * exponent obtained by the rounding of that
   rationnal.
   Warning: will overflow for exponants or mantissa sizes larger than
   max_int (unlikely to happen). *)
(* exp = 149 for 32 bits *)
let to_mantissa_exp prec(*23/32b*) exp (*149/32b*) (mode : Term.t) x =
  if Q.equal x Q.zero then Z.zero, 1
  else
    let prec = int_of_string (Q.to_string prec) in
    (* let prec = 24 in *)
    let exp  = int_of_string (Q.to_string exp) in
    let x'   = Q.abs x in
    let e    = (log_2 x') + 1 in
    let mx = max (e - prec) (-exp) in
    (*Format.fprintf fmt  "max = %d@." mx;*)
    let y    = div_2exp x' e in
    (* rnd_big_int takes a positive Q.t between 0 and 1/2 *)
    let mode =
      if Term.equal mode Builtin.FPA._Down__rounding_mode then
        if Q.sign x < 0 then Builtin.FPA._Up__rounding_mode else mode
      else
        if Term.equal mode Builtin.FPA._Nu__rounding_mode then
          if Q.sign x < 0 then Builtin.FPA._Nd__rounding_mode else mode
        else mode
    in
    (*
      -precondition: 1/2 <= n < 1
      -postcondition: if n =bin 0.b1b2b3...
      then result = b1*2^p + b2*2^(p-1) + ... + bp*1 + rounding_bit
    *)
    let rec float_comp p n f_n =
      if p = 0 then rnd_big_int mode n f_n else
	begin
          let h = Z.shift_left f_n 1 in
          let l = Q.truncate (mul_2exp n 1) in
          let f_n = Z.add h l in
          let n = Q.sub (mul_2exp n 1) (Q.from_z l) in
          float_comp (p - 1) n f_n
	end
    in
    let k = float_comp prec y Z.zero in
    if e > (- exp) then
      (Z.mult (Z.from_int (Q.sign x)) k , e - prec)
    else (Z.zero, 1)



(* Compute the floating-point approximation of a number *)
let float_of_const prec exp mode x =
  let m, e = to_mantissa_exp prec exp mode x in
  let res = mul_2exp (Q.from_z m) e in
  (*fprintf fmt "float_of_const: %a, %a, %a : %a@."
    Q.print prec Q.print exp Q.print x Q.print res;*)
  res

(* TODO: make sure this works *)
let to_float (x : Q.t) : float =
  let m, e =
    to_mantissa_exp (Q.from_int 23) (Q.from_int 104)
      Builtin.FPA._NearestTiesToEven__rounding_mode x in
  (Z.to_float m) *. (2. ** (float e))

let fix_of_const prec mode x = raise Not_implemented

(*compute_error x is a bound on the distance between x
  and the floating point representation of x
  given a bound on x*)
(*We compute the rounding error as follows:
  if the number is normalized (log_2 x - precision) >= min_exp
  then it is equal to 2^(log_2 x - prec) + prec_err where prec_err
  is zero if rounding to nearest and 1 otherwise.

  If the number is denormalized, the error is 2^min_exp + prec_err

  Note that min_exp = - exp in the above explanation
*)
let compute_error prec exp (mode: Term.t) x =
  let min_exp = int_of_string (Q.to_string (Q.minus exp)) in
  let x = Q.abs x in
  let exponent_bnd =
    if Q.equal x Q.zero
    then min_exp
    else log_2 x in
  let prec = int_of_string (Q.to_string prec) in
  let rnd_err =
    if Term.equal mode Builtin.FPA._NearestTiesToEven__rounding_mode then 0
    else 1
  in
  let exponent_error = (max (exponent_bnd - prec) min_exp) + rnd_err in
  let e = mul_2exp Q.one exponent_error in
  e


(* precondition: n is non-negative*)
let maybe_sqrt_z n =
  let sqrt, rem = Z.sqrt_rem n in
  if Z.equal rem Z.zero then Some sqrt
  else None


let maybe_sqrt r =
  if Q.compare_to_0 r < 0 then None
  else
    let num, den = Q.num r, Q.den r in
    match maybe_sqrt_z num, maybe_sqrt_z den with
      | _, None -> None
      | None, _ -> None
      | Some p, Some q -> Some (Q.from_zz p q)

